/*	Functions for handling an (optionally) alphabetically sorted List
	Manager list.
	
	The event functions in this file will: handle the arrow
	keys; scroll to a specific entry when the user enters some
	characters; resize the list and scroll bars; handle mouse clicks
	in the list; handle activate, deactivate, suspend, and resume
	events.

	The list's 'refCon' field is currently (93/11/20) used for some
	additional data.
	
	Revision history:
	
	94/01/01 aih
	- added support for page up, page down, home, and end keys
	
	93/12/08 aih
	- frame and user data are stored in list's refCon field
	
	93/11/24 aih
	- added list flags functions
	
	93/11/20 aih
	- refCon field is used to save status of input focus
	
	93/11/17 aih
	- modified resize function to minimize drawing drawn on resizing list

	93/11/03
	- updated list validation
	
	93/10/19 aih
	- the list is created to fit within the view rectangle, taking into
	account the scroll bars and the double border drawn around lists
	to indicate the input focus
	
	93/10/16 aih
	- added functions to support use of lists in windows other than those
	created directly by the libraries (and thus lacking the extra information
	those windows contain)
	- fixed some errors in handling scroll bars
	- case is ignored when searching for typed characters
	
	93/03/16 AIH
	- Added mouseDown handler
	
	93/03/12 AIH
	- Changed name of library from "ListBoxLib" to "ListLib"
	
	92/05/24 AIH
	- Added support for left and right arrow keys
	
	92/02/26 AIH
	- Simplified event handling
	
	91/06/01 AIH
	- Added functions to calculate the width and height of a list
	
	91/05/24 AIH
	- Uses IUMagString instead of memcmp
	- Added ability to set a list definition function without using a
	separate code resource
	- Added function to return the port containing the list
	
	91/05/07 AIH
	- Fixed mistake in calculating size of list
	
	91/04/12 AIH
	- Added functions to access the fields of a list, instead of directly
	dereferencing the list
	
	91/04/03 AIH
	- Adapted for multiple event lists
	
	91/03/24 AIH
	- Current port is set before mouse region is adjusted
	
	91/03/14 AIH
	- Scroll bars are resized reparately if window has a grow icon
	
	91/03/09 Ari Halberstadt (AIH)
	- Created this library by modifying some existing code */

#include <limits.h>
#include <string.h>
#include <Packages.h>
#include "AsmLib.h"
#include "ControlLib.h"
#include "DrawLib.h"
#include "EventLib.h"
#include "KeyLib.h"
#include "ListLib.h"
#include "MathLib.h"
#include "MemoryLib.h"
#include "MenuLib.h"
#include "RectangleLib.h"
#include "ResourceLib.h"
#include "ScrollBarLib.h"
#include "WindowLib.h"

/* data stored in list's refCon field */
typedef struct {
	FrameHandle frame;
	void *userData;
} ListDataType, *ListDataPtr, **ListDataHandle;

/* true if list is valid */
Boolean ListValid(ListHandle list)
{
	Rect r, v;
	
	if (! HandleValidSize(list, sizeof(ListRec))) return(false);
	if (! WinValid((**list).port)) return(false);
	r = (**list).rView; if (! RectValid(&r)) return(false);
	r = (**list).dataBounds; if (! RectValid(&r)) return(false);
	v = (**list).visible; if (! RectValid(&v)) return(false);
	if ((**list).cellSize.h <= 0) return(false);
	if ((**list).cellSize.v <= 0) return(false);
	if ((**list).vScroll && ! CtlValid((**list).vScroll)) return(false);
	if ((**list).hScroll && ! CtlValid((**list).hScroll)) return(false);
	if (! HandleValid((**list).cells) || ! *(**list).cells) return(false);
	if (HandleSize((**list).cells) > SHRT_MAX) return(false);
	if ((**list).maxIndex < 0) return(false);
	return(true);
}

/* return window containing list */
WindowPtr ListWindow(ListHandle list)
{
	require(ListValid(list));
	return((**list).port);
}

/* get lists's private data */
static ListDataHandle GetListData(ListHandle list)
{
	ListDataHandle listData;
	
	require(ListValid(list));
	listData = (ListDataHandle) (**list).refCon;
	ensure(HandleValidSize(listData, sizeof(ListDataType)));
	return(listData);
}

/* set lists's private data */
static void SetListData(ListHandle list, ListDataHandle listData)
{	
	require(ListValid(list));
	require(HandleValidSize(listData, sizeof(ListDataType)));
	(**list).refCon = (long) listData;
	ensure(GetListData(list) == listData);
}

/* return list's user data */
void *ListData(ListHandle list)
{
	return((**GetListData(list)).userData);
}

/* set list's user data */
void ListDataSet(ListHandle list, void *data)
{
	(**GetListData(list)).userData = data;
	ensure(ListData(list) == data);
}

/* return list's frame */
FrameHandle ListFrame(ListHandle list)
{
	return((**GetListData(list)).frame);
}

/* return list's enclosing rectangle (excluding scroll bars) */
void ListRect(ListHandle list, Rect *r)
{
	require(ListValid(list));
	*r = (**list).rView;
	ensure(RectValid(r));
}

/* return list's dataBounds rectangle */
void ListDataBounds(ListHandle list, Rect *r)
{
	require(ListValid(list));
	*r = (**list).dataBounds;
	ensure(RectValid(r));
}

/* returns list's vertical scroll bar */
ControlHandle ListVScroll(ListHandle list)
{
	require(ListValid(list));
	return((**list).vScroll);
}

/* returns list's horizontal scroll bar */
ControlHandle ListHScroll(ListHandle list)
{
	require(ListValid(list));
	return((**list).hScroll);
}

/* call the list definition function */
void ListDefinitionCall(ListHandle list, short lMessage, Boolean lSelect,
	const Rect *lRect, Cell lCell, short lDataOffset, short lDataLen,
	Handle ldef)
{
	GrafPtr port = NULL;
	SignedByte state = 0;
	
	require(! ldef || HandleValid(ldef));
	if (ldef && *ldef) {
		GetPort(&port);
		SetPort((**list).port);
		state = HandleLock(ldef);
		((ListDefinitionType) *ldef)(lMessage, lSelect, lRect, lCell,
			lDataOffset, lDataLen, list);		
		HandleRestore(ldef, state);
		SetPort(port);
	}
}

/* Default list definition function. Currently, this just calls the
	standard list definition function, but in the future we could
	add other features to this function (such as drawing an icon
	to the left of the text, like in directory dialog boxes). */
pascal void ListDefinitionDefault(short lMessage, Boolean lSelect,
	const Rect *lRect, Cell lCell, short lDataOffset, short lDataLen,
	ListHandle lHandle)
{
	ListDefinitionCall(lHandle, lMessage, lSelect, lRect, lCell,
		lDataOffset, lDataLen, GetResource('LDEF', 0));
}

/* return a handle to the current LDEF */
Handle ListDefinition(ListHandle list)
{
	require(ListValid(list));
	return((**list).listDefProc);
}

/* Set the list's definition function, or use default definition ('LDEF' 0)
	if 'defproc' is NULL. If you call this function after creating a list,
	then you must also call it with a NULL defproc before calling
	ListEnd or LDispose. */
void ListDefinitionSet(ListHandle list, ListDefinitionType defproc)
{
	struct glue {		/* glue for list definition function */
		short jmp;		/* jump instruction */
		void *addr;		/* address of list definition function */
	};
	volatile Handle newdef = NULL;		/* list's new definition function */
	volatile Handle olddef = NULL;		/* list's old definition function */
	volatile Boolean closemsg = false;	/* for catch handler */
	volatile Boolean initmsg = false;	/* for catch handler */
	Cell dummy = { 0, 0 };					/* for calling the definition function */
	
	require(ListValid(list));
	TRY {
		/* if the list already has a handle to a definition function which was
			created using this function (i.e., it's not a handle to a resource)
			then send it an lCloseMsg and dispose of the glue handle */
		olddef = (**list).listDefProc;
		if (olddef) {
			ListDefinitionCall(list, lCloseMsg, false, NULL, dummy, 0, 0, olddef);
			closemsg = true;
		}
		(**list).listDefProc = NULL;
		
		/* install the new definition function */
		if (defproc) {
			/* create a glue handle to jump to the new defproc */
			newdef = HandleBegin(sizeof(struct glue));
			(**(struct glue **) newdef).jmp = ASM_JMP;
			(**(struct glue **) newdef).addr = defproc;
		}
		else
			newdef = ResGet('LDEF', 0);
		(**list).listDefProc = newdef;
	
		/* send the definition function an lInitMsg */
		if (defproc) {
			defproc(lInitMsg, false, NULL, dummy, 0, 0, list);
			initmsg = true;
		}
		HandleEnd(olddef);
		olddef = NULL;
	} CATCH {
		if (initmsg)
			defproc(lCloseMsg, false, NULL, dummy, 0, 0, list);
		(**list).listDefProc = olddef;
		if (closemsg)
			ListDefinitionCall(list, lInitMsg, false, NULL, dummy, 0, 0, olddef);
		HandleEnd(newdef);
	} ENDTRY;
	ensure(HandleValidSize(newdef, sizeof(struct glue)));
	ensure(ListDefinition(list) == newdef);
}

/* deselect all but the specified cell */
void ListSelectOne(ListHandle list, Cell select)
{
	Cell current;
	
	require(ListValid(list));
	current.h = current.v = 0;
	while (LGetSelect(true, &current, list)) {
		if (current.h != select.h || current.v != select.v)
			LSetSelect(false, current, list);
		else if (! LNextCell(true, true, &current, list))
			break;
	}
	if (! LGetSelect(false, &select, list))
		LSetSelect(true, select, list);
}

/* use RelString to compare two strings */
static short CompareString(const char *a, const char *b, short alen, short blen)
{
	asm {
		move.l a, a0
		move.l b, a1
		move.w alen, d0
		lsl.l #8, d0
		lsl.l #8, d0
		move.w blen, d0
		_RelString
	}
}

/* Comparison function for LSearch, used to find an item given the first
	few rapidly typed keystrokes. (aPtr is the contents of the cell, and
	bPtr is the string to find.) Capitalization is ignored and only the
	first 255 characters are used for the comparison. */
static pascal short InsertProc(Ptr a, Ptr b, short alen, short blen)
{
	Boolean found = false;

	if (alen == blen)
		found = (CompareString(a, b, alen, blen) == sortsEqual);
	else if (alen > blen)
		found = (CompareString(a, b, blen, blen) == sortsEqual);
	if (! found)
		found = (CompareString(a, b, min(alen, blen), min(alen, blen)) == sortsAfter);
	return(! found);
}

/* Search list for buffer and set the 'insert' parameter to the
	cell before which the data should be inserted. The list is assumed
	to be sorted using IUMagString. True is returned if the cell
	before which to insert the buffer was found, otherwise false
	is returned and 'insert' is set to the bottom right of dataBounds. */
Boolean ListSearchInsert(ListHandle list, Ptr buf, short len, Cell *insert)
{
	Boolean found = false;
	
	require(ListValid(list));
	insert->h = insert->v = 0;
	found = LSearch(buf, len, InsertProc, insert, list);
	if (! found) {
		insert->h = (**list).dataBounds.right;
		insert->v = (**list).dataBounds.bottom;
	}
	return(found);
}

/* Handle any other key by accumulating rapidly typed characters into a small
	buffer and selecting the item in the list which most closely matches
	the characters in the buffer. It is assumed that the list is in ASCII
	sorted order. */
static void KeyOther(ListHandle list, EventRecord *event)
{
	static CStr31		buf;		/* characters typed */
	static short		len;		/* length of buffer */
	static TicksType	previous;/* time when last character was typed */
	static ListHandle listbuf;	/* list for which buffer is being used */
	Boolean	found;	/* true if buffer was found in list */
	Cell	c;				/* selected cell */
	char	key;			/* key hit */

	require(ListValid(list));
	
	key = event->message;
	
	/* clear buffer if we've changed lists */
	if (list != listbuf) {
		previous = 0;
		len = 0;
	}
	listbuf = list;
	
	/* accumulate rapidly typed characters into a buffer and search
		for first occurrence of a cell that starts with those characters */
	if (! previous)
		previous = event->when;
	if (event->when - previous > GetDblTime())
		len = 0;
	if (len < sizeof(CStr31)) {

		/* add character to (or remove character from) buffer */
		if (len > 0 && key == backspaceKey)
			len--;
		else
			buf[len++] = key;
		
		/* search for cell starting with buffer and select it */
		c.h = c.v = 0;
		found = ListSearchInsert(list, buf, len, &c);
		if (! found) {
			--c.h;
			--c.v;
		}
		ListSelectOne(list, c);
		LAutoScroll(list);
	}
	previous = TickCount();
}

/* handle a right or down arrow key */
static void KeyArrowDownOrRight(ListHandle list, char key)
{
	Cell	c;				/* selected cell */
	Cell	next;			/* for finding next cell to select */
	Rect	cellRect;	/* current cell's rectangle */
	Rect	viewRect;	/* view rectangle */
	Rect	dataBounds;	/* list's bounds */
	short	horizontal;	/* 1 (true) if moving horizontally, else 0 (false) */
	short	vertical;	/* 1 (true) if moving vertically, else 0 (false) */
	
	require(ListValid(list));
	require(key == arrowRightKey || key == arrowDownKey);
	
	horizontal = (key == arrowRightKey);
	vertical = (key == arrowDownKey);
	
	/* find first selected cell */
	c.h = c.v = 0;
	if (LGetSelect(true, &c, list)) {
	
		/* find last selected cell */
		next = c;
		while (LNextCell(horizontal, vertical, &next, list)) {
			if (LGetSelect(true, &next, list))
				c = next;
			else
				break;
		}
		
		/* select next cell (if any) */
		dataBounds = (**list).dataBounds;
		c.h = min(c.h + horizontal, max(dataBounds.right - 1, 0));
		c.v = min(c.v + vertical, max(dataBounds.bottom - 1, 0));
		ListSelectOne(list, c);
			
		/* scroll list so cell is visible */
		viewRect = (**list).rView;
		LRect(&cellRect, c, list);
		if (vertical) {
			if (cellRect.bottom == viewRect.bottom)
				LScroll(0, 1, list);
			else if (cellRect.bottom > viewRect.bottom || cellRect.top < viewRect.top)
				LAutoScroll(list);
		}
		else if (horizontal) {
			if (cellRect.right == viewRect.right)
				LScroll(1, 0, list);
			else if (cellRect.right > viewRect.right || cellRect.left < viewRect.left)
				LAutoScroll(list);
		}
	}
	else {
		/* nothing's selected, so jump to start of list */
		c.h = c.v = 0;
		LSetSelect(true, c, list);
		LAutoScroll(list);
	}
}

/* handle an up arrow or left */
static void KeyArrowUpOrLeft(ListHandle list, char key)
{
	Cell c; /* currently selected cell and cell to select */

	require(ListValid(list));
	require(key == arrowLeftKey || key == arrowUpKey);

	/* find first selected cell */
	c.h = c.v = 0;
	if (LGetSelect(true, &c, list)) {
		/* select previous cell */
		if (key == arrowLeftKey) c.h = max(c.h - 1, 0);
		else if (key == arrowUpKey) c.v = max(c.v - 1, 0);
		ListSelectOne(list, c);
	}
	else {
		/* nothing's selected so jump to end of list */
		c.h = (**list).dataBounds.left;
		c.v = (**list).dataBounds.bottom;
		LSetSelect(true, c, list);
	}
	LAutoScroll(list);
}

/* handle a virtual key code, return true if handled */
static Boolean KeyVirtualCode(ListHandle list, char virtual)
{
	short dv, dh;
	Rect visible, bounds;
	Boolean result = true;
	
	require(ListValid(list));
	visible = (**list).visible;
	bounds = (**list).dataBounds;
	dv = dh = 0;
	switch(virtual) {
	case pageUpVKey:
		dv = -RectHeight(&visible);
		break;
	case pageDownVKey:
		dv = RectHeight(&visible);
		break;
	case homeVKey:
		dh = bounds.left - visible.left;
		dv = bounds.top - visible.top;
		break;
	case endVKey:
		dh = bounds.right - visible.right;
		dv = bounds.bottom - visible.bottom;
		break;
	default:
		result = false;
		break;
	}
	LScroll(dh, dv, list);
	return(result);
}

/* adjust the edit menu */
void ListAdjustMenu(ListHandle list)
{
	require(ListValid(list));
	MenuCmdEnable(CMD_EDIT);
	if ((**list).selFlags != lOnlyOne)
		MenuCmdEnable(CMD_SELECT_ALL);
}

/* handle a menu command */
Boolean ListMenu(ListHandle list, const MenuPickType *pick)
{
	Boolean handled = false;
	Rect dataBounds;
	Cell c;
	
	require(ListValid(list));
	if (pick->cmd == CMD_SELECT_ALL) {
		/* select all the cells in the list */
		dataBounds = (**list).dataBounds;
		for (c.h = dataBounds.left; c.h < dataBounds.right; c.h++) {
			for (c.v = dataBounds.top; c.v < dataBounds.bottom; c.v++)
				LSetSelect(true, c, list);
		}
		handled = true;
	}
	return(handled);
}

/* true if the point is within the list */
Boolean ListWithin(ListHandle list, Point where)
{
	return(FrameWithin(ListFrame(list), where));
}

/* handle key down events */
void ListKeyDown(ListHandle list, EventRecord *event)
{
	char key = event->message;
	
	if (! KeyVirtualCode(list, key)) {//debug_note//KeyVirtual(event))) {
		switch (key) {
		case arrowDownKey:
		case arrowRightKey:
			KeyArrowDownOrRight(list, key);
			break;
		case arrowUpKey:
		case arrowLeftKey:
			KeyArrowUpOrLeft(list, key);
			break;
		default:
			if ((event->modifiers & cmdKey) == 0 && key >= ' ')
				KeyOther(list, event);
			break;
		}
	}
}

/* update the list */
void ListUpdate(ListHandle list)
{
	LUpdate(ListWindow(list)->visRgn, list);
}

/* set input focus */
void ListFocus(ListHandle list, Boolean focus)
{
	FrameFocus(ListFrame(list), focus);
}

/* activate or deactivate the list */
void ListActivate(ListHandle list, Boolean activate)
{
	LActivate(activate, list);
	FrameActivate(ListFrame(list), activate);
}

/* set the preferred width and height for the window */
void ListZoom(ListHandle list, short *width, short *height)
{
	Rect dataBounds;
	
	require(ListValid(list));
	dataBounds = (**list).dataBounds;
	*width = RectWidth(&dataBounds) * (**list).cellSize.h;
	*height = RectHeight(&dataBounds) * (**list).cellSize.v;
	FrameZoom(ListFrame(list), width, height);
}

/* resize the list by dh and dv pixels */
void ListResize(ListHandle list, short dh, short dv)
{
	Rect view;
	
	require(ListValid(list));
	view = (**list).rView;
	if ((FrameFlags(ListFrame(list)) & FRAME_ANCHOR_RIGHT) == 0) dh = 0;
	if ((FrameFlags(ListFrame(list)) & FRAME_ANCHOR_BOTTOM) == 0) dv = 0;
	LSize(RectWidth(&view) + dh, RectHeight(&view) + dv, list);
}

/* handle a mouseDown */
void ListMouseDown(ListHandle list, EventRecord *event)
{
	Point where;

	require(ListValid(list));
	where = event->where;
	GlobalToPort(&where, (**list).port);
	LClick(where, event->modifiers, list);
}

/* Create a new list. This is similar to LNew, but the sizes of the
	scroll bars are subtracted from the view rect, and if the
	frame flag is true then the view rect is also inset by 3
	pixels to allow for a thick double frame around the list. */
ListHandle ListBegin(const Rect *viewRect, const Rect *bounds, Point size,
	short proc, WindowPtr window, Boolean draw, Boolean grow,
	Boolean horiz, Boolean vert, Boolean focus)
{
	volatile ListHandle list = NULL;
	ListDataHandle listData = NULL;
	FrameHandle frame = NULL;
	Rect content;
	
	TRY {
		/* calculate content rectangle */
		content = *viewRect;
		if (focus) InsetRect(&content, FRAME_FOCUS_SIZE, FRAME_FOCUS_SIZE);
		if (horiz) content.bottom -= SBAR_WIDTH - 1;
		if (vert) content.right -= SBAR_WIDTH - 1;
		InsetRect(&content, FRAME_WIDTH, FRAME_WIDTH);
		check(RectValid(&content));
		
		/* create list */
		list = LNew(&content, bounds, size, 0, window, draw, grow, horiz, vert);
		FailNIL(list);
		
		/* create private data and frame containing list */
		listData = HandleBeginClear(sizeof(ListDataType));
		(**list).refCon = (long) listData;
		frame = FrameBegin(window, viewRect, 0);
		(**listData).frame = frame;
		if (horiz) FrameFlagsSet(frame, FrameFlags(frame) | FRAME_HSCROLL);
		if (vert) FrameFlagsSet(frame, FrameFlags(frame) | FRAME_VSCROLL);
		if (focus) FrameFlagsSet(frame, FrameFlags(frame) | FRAME_LIST);

		/* use our default list definition function */
		ListDefinitionSet(list, ListDefinitionDefault);

		/* register list with window */
		WinRegister(ListWindow(list), list, ListEventTable());
		WinTabRegister(ListWindow(list), list);

	} CATCH {
		ListEnd(list);
	} ENDTRY;
	ensure(ListValid(list));
	return(list);
}

/* dispose of the list */
void ListEnd(ListHandle list)
{
	if (list) {
		if ((**list).refCon) {
			FrameEnd(ListFrame(list));
			HandleEnd(GetListData(list));
		}
		ListDefinitionSet(list, NULL);
		WinTabUnregister(ListWindow(list), list);
		WinUnregister(ListWindow(list), list);
		LDispose(list);
		list = NULL;
	}
	ensure(! ListValid(list));
}
