/* 93/12/22 aih - created */

#include "ArrayListLib.h"
#include "MemoryLib.h"

/* true if a valid list */
Boolean ArrayListValid(ArrayListHandle list)
{
	if (! HandleValidSize(list, sizeof(ArrayListType))) return(false);
	if (HandleSize((**list).array) != (**list).nelem * (**list).elemsz) return(false);
	return(true);	
}

/* true if a valid index into the array */
Boolean ArrayListValidIndex(ArrayListHandle list, size_t index)
{
	require(ArrayListValid(list));
	return(0 <= index && index < (**list).nelem);
}

/* create a new list */
ArrayListHandle ArrayListBegin(size_t elemsz)
{
	volatile ArrayListHandle list = NULL;
	Handle array = NULL;
	
	TRY {
		list = HandleBeginClear(sizeof(ArrayListType));
		array = HandleBegin(0);
		(**list).array = array;
		(**list).elemsz = elemsz;
	} CATCH {
		ArrayListEnd(list);
	} ENDTRY;
	ensure(ArrayListValid(list));
	return(list);
}

/* dispose of the list */
ArrayListHandle ArrayListEnd(ArrayListHandle list)
{
	require(! list || ArrayListValid(list));
	if (list) {
		HandleEnd((**list).array);
		HandleEnd(list);
		list = NULL;
	}
	ensure(! ArrayListValid(list));
}

/* return number of items in array */
size_t ArrayListCount(ArrayListHandle list)
{
	require(ArrayListValid(list));
	return((**list).nelem);
}

/* return size of items in array */
size_t ArrayListItemSize(ArrayListHandle list)
{
	require(ArrayListValid(list));
	return((**list).elemsz);
}

/* insert space for an item before the indexed item */
void ArrayListInsert(ArrayListHandle list, size_t index)
{
	require(ArrayListValid(list));
	require(0 <= index && index <= (**list).nelem);
	HandleSizeSet((**list).array, ((**list).nelem + 1) * (**list).elemsz);
	if (index < (**list).nelem) {
		BlockMove(*(**list).array + index * (**list).elemsz,
				 	 *(**list).array + (index + 1) * (**list).elemsz,
					 ((**list).nelem - index) * (**list).elemsz);
	}
	(**list).nelem++;
	ensure(ArrayListValidIndex(list, index));
}

/* remove the indexed item */
void ArrayListDelete(ArrayListHandle list, size_t index)
{
	require(ArrayListValid(list));
	require(ArrayListValidIndex(list, index));
	if (index < (**list).nelem - 1) {
		BlockMove(*(**list).array + (index + 1) * (**list).elemsz,
					 *(**list).array + index * (**list).elemsz,
					 ((**list).nelem - index - 1) * (**list).elemsz);
	}
	HandleSizeSet((**list).array, --(**list).nelem * (**list).elemsz);
	ensure(ArrayListValid(list));
}

/* set the value of the indexed item */
void ArrayListSet(ArrayListHandle list, size_t index, void *data)
{
	require(ArrayListValid(list));
	require(ArrayListValidIndex(list, index));
	BlockMove(data, *(**list).array + index * (**list).elemsz, (**list).elemsz);
	ensure(ArrayListValid(list));
}

/* get the value of the indexed item */
void ArrayListGet(ArrayListHandle list, size_t index, void *data)
{
	require(ArrayListValid(list));
	require(ArrayListValidIndex(list, index));
	BlockMove(*(**list).array + index * (**list).elemsz, data, (**list).elemsz);
	ensure(ArrayListValid(list));
}

/* return the handle to the array */
void *ArrayListGetHandle(ArrayListHandle list)
{
	require(ArrayListValid(list));
	return((**list).array);
}

/* detach the array from the list (call ArrayListGetHandle first) */
void ArrayListDetachHandle(ArrayListHandle list)
{
	require(ArrayListValid(list));
	(**list).array = HandleBegin(0);
	(**list).nelem = 0;
	ensure(ArrayListCount(list) == 0);
}

/* attach the handle to the array */
void ArrayListAttachHandle(ArrayListHandle list, void *array)
{
	require(ArrayListValid(list));
	require(HandleValid(array));
	HandleEnd((**list).array);
	(**list).array = array;
	(**list).nelem = HandleSize(array) / (**list).elemsz;
	ensure(ArrayListValid(list));
}
