/*	Implement lexical search tree using character strings as keys.
	Based on information in Sleator and Tarjan's article on splay trees.
	
	94/01/17 aih - disposal routine wasn't traversing middle nodes
	94/01/07 aih - uses debug library for debug output
					 - uses fixed size pointer library to allocate nodes
					 - lets application allocate node's data field
	93/11/01 aih - created */

#include <string.h>
#include "DebugLib.h"
#include "PointerFixedLib.h"
#include "LexTreeLib.h"
#include "MemoryLib.h"

#ifndef NDEBUG
	#define LEXTREE_DEBUG (0)
#endif

#if LEXTREE_DEBUG

/* debug output function */
static void LexTreeDebug(const char *fmt, ...)
{
	va_list ap;
	
	va_start(ap, fmt);
	DebugVPrintf(fmt, ap);
	va_end(ap);
}

#endif /* LEXTREE_DEBUG */

/* true if node is valid */
static Boolean LexNodeValid(LexNodePtr x)
{
	return(PtrFixedValidSize(x, sizeof(LexNodeType)));
}

/* return a new node */
static LexNodePtr LexNodeBegin(char key)
{
	LexNodePtr x;
	
	require(key != 0);
	x = PtrFixedBegin(sizeof(LexNodeType));
	memclr(x, sizeof(LexNodeType));
	x->key = key;
	return(x);
}

/* dispose of the node */
static void LexNodeEnd(LexNodePtr x)
{
	require(! x || LexNodeValid(x));
	if (x) {
		PtrFixedEnd(x);
		x = NULL;
	}
	ensure(! LexNodeValid(x));
}

/* true if tree is valid */
Boolean LexTreeValid(LexTreePtr t)
{
	if (! PtrValidSize(t, sizeof(LexTreeType))) return(false);
	if (t->root && ! LexNodeValid(t->root)) return(false);
	if (t->root && t->size == 0) return(false);
	if (! t->root && t->size > 0) return(false);
	return(true);
}

/* return a new tree */
LexTreePtr LexTreeBegin(void)
{
	return(PtrBeginClear(sizeof(LexTreeType)));
}

/* recursively dispose of every node in the tree */
static void dispose(LexNodePtr x)
{
	require(LexNodeValid(x));
	if (x->left) dispose(x->left);
	if (x->middle) dispose(x->middle);
	if (x->right) dispose(x->right);
	LexNodeEnd(x);
}

/* dispose of the tree */
void LexTreeEnd(LexTreePtr t)
{
	require(! t || LexTreeValid(t));
	if (t) {
		if (t->root)
			dispose(t->root);
		PtrEnd(t);
		t = NULL;
	}
	ensure(! LexTreeValid(t));
}

/* return root of the tree */
LexNodePtr LexTreeRoot(LexTreePtr t)
{
	require(LexTreeValid(t));
	return(t->root);
}

/* return number of strings (terminal nodes) in tree */
size_t LexTreeSize(LexTreePtr t)
{
	require(LexTreeValid(t));
	return(t->size);
}

/*
	Given any valid tree rooted at x and any string s, the values of
	the three local variables when finished searching, and the actions
	to be taken for each combination, are shown in the following table.
	
	x and p can have the values:
		nil = nil pointer, term = terminal node, ! term = non-terminal node
	s can have the values:
		nil = nil pointer or empty string, !nil = non-empty length string
	
#	x      p      s      explanation              action
-	-----  -----  -----  -----------------------  ------------------------------
legal conditions:
1  nil    term   nil    found entire string      return p
2  !term  term   nil    found entire string      return p
3  term   term   nil    found entire string      return p
4  nil    !term  nil    prefix found             p->terminal = true; return p
5  term   !term  nil    prefix found             p->terminal = true; return p
6  nil    term   !nil   prefix found             add s, return new terminal
7  nil    !term  !nil   prefix found             add s, return new terminal
8  nil    nil    !nil   empty tree               add s, return new terminal
ilegal conditions:
9  nil    nil    nil    empty s, empty tree      forbidden by precondition
10 !nil   nil    nil    empty s                  forbidden by precondition
11 !nil   nil    !nil   illegal loop exit        forbidden by loop expression
*/

/* insert suffix 's' as left, middle, or right child of node 'p' (depending on
	value of 'c'), return new terminal node */
LexNodePtr LexTreeInsert(LexTreePtr t, LexNodePtr p, const char *s, char c)
{
	require(LexTreeValid(t));
	require(! p ? ! t->root : LexNodeValid(p));
	require(s != NULL);
	#if LEXTREE_DEBUG
		LexTreeDebug("  LexTreeInsert, s='%s'\n", s);
	#endif
	if (! p) {
		/* empty tree (case 8), add root, make like case 4 or 7 */
		check(! t->root && *s);
		#if LEXTREE_DEBUG
			LexTreeDebug("    %-16s %c\n", "make root", *s);
		#endif
		p = t->root = LexNodeBegin(*s++);
		c = p->key;
	}
	if (*s) {
		/* prefix found (case 6 or 7), add remainder of s, make like case 4 */
		if (c < p->key) {
			check(! p->left);
			#if LEXTREE_DEBUG
				LexTreeDebug("    %-16s *s=%c, c=%c\n", "make left", *s, c);
			#endif
			p->left = LexNodeBegin(*s++);
			p = p->left;
		}
		else if (c > p->key) {
			check(! p->right);
			#if LEXTREE_DEBUG
				LexTreeDebug("    %-16s *s=%c, c=%c\n", "make right", *s, c);
			#endif
			p->right = LexNodeBegin(*s++);
			p = p->right;
		}
		check(! p->middle);
		while (*s) {
			#if LEXTREE_DEBUG
				LexTreeDebug("    %-16s *s=%c\n", "make middle", *s);
			#endif
			p->middle = LexNodeBegin(*s++);
			p = p->middle;
		}
	}
	if (! p->terminal) {
		/* prefix found (case 4 or 5), make terminal node, case 1, 2, or 3 */
		p->terminal = true;
		t->size++;
	}
	ensure(p && p->terminal && ! *s);
	ensure(LexTreeValid(t));
	ensure(t->size > 0);
	return(p);
}

/* Find string 's' in tree 't'. If 's' isn't found add it to the tree.
	Returns terminal node. The application can store data into the
	node using the node's 'data' field. For instance, if the 'data'
	field of the returned node is null then it means no data have
	been stored into the field yet, and the application could call
	malloc to store some data into the node.  */
LexNodePtr LexTreeFind(LexTreePtr t, const char *s)
{
	LexNodePtr x = NULL;
	LexNodePtr p = NULL;
	char c = 0;
	
	require(LexTreeValid(t));
	require(s && *s); /* (cases 9 and 10) */
	#if LEXTREE_DEBUG
		LexTreeDebug("LexTreeFind, s='%s'\n", s);
	#endif
	/* search for string in tree */
	x = t->root;
	while (x && *s) {
		p = x;
		c = *s;
		if (*s < x->key) {
			#if LEXTREE_DEBUG
				LexTreeDebug("  %-16s (%c < %c)\n", "go left", *s, x->key);
			#endif
			x = x->left;
		}
		else if (*s > x->key) {
			#if LEXTREE_DEBUG
				LexTreeDebug("  %-16s (%c > %c)\n", "go right", *s, x->key);
			#endif
			x = x->right;
		}
		else {
			#if LEXTREE_DEBUG
				LexTreeDebug("  %-16s (%c == %c)\n", "go middle", *s, x->key);
			#endif
			x = x->middle;
			s++;
		}
	}
	check(! (x && *s)); /* (case 11) */
	return(LexTreeInsert(t, p, s, c));
}

/* Traverse tree in in-order. 'X' should be the node from which to start
	traversing the tree. 'Prefix' is the prefix string built so far and
	must be at least large enough to contain the longest string inserted
	into the tree. 'Length' is the length of the prefix string.
	'Visit' is a function called whenever a terminal node is encountered.
	When 'visit' is called the first 'length' bytes of the 'prefix'
	parameter contain the string corresponding to the terminal node.
	'Data' is a pointer to application-supplied data to be passed to the
	'visit' function. */
void LexTreeInOrder(LexNodePtr node, char *prefix, size_t length,
	LexTreeVisitType visit, void *data)
{
	require(LexNodeValid(node));
	if (node->left)
		LexTreeInOrder(node->left, prefix, length, visit, data);
	if (node->middle || node->terminal) {
		prefix[length] = node->key;
		if (node->terminal) {
			prefix[length + 1] = 0;
			visit(node, prefix, length + 1, data);
		}
		if (node->middle)
			LexTreeInOrder(node->middle, prefix, length + 1, visit, data);
	}
	if (node->right)
		LexTreeInOrder(node->right, prefix, length, visit, data);
}
