/* Generic support for standard File menu commands.
	94/01/20 aih - added autosave flag
	94/01/14 aih - created */

#include <stdio.h>
#include <string.h>	
#include "ApplicationLib.h"
#include "ArchiveLib.h"
#include "DialogLib.h"
#include "DrawLib.h"
#include "DocumentLib.h"
#include "EventLib.h"
#include "FileDialogLib.h"
#include "FinderLib.h"
#include "HighLevelEventLib.h"
#include "MemoryLib.h"
#include "NotifyLib.h"
#include "ResourceConstantsLib.h"
#include "ResourceLib.h"
#include "ResourceTypeLib.h"
#include "ScreenLib.h"
#include "WindowLib.h"
#include "WindowMenuLib.h"

static short gDocNTypes;			/* number of file types that can be opened */
static const OSType *gDocTypes;	/* list of file types that can be opened */
static const DocumentTableType *gDocTable;	/* table of handlers */

/*----------------------------------------------------------------------------*/
/* validation */
/*----------------------------------------------------------------------------*/

/* true if a valid document */
Boolean DocValid(DocumentHandle doc)
{
	if (! HandleValidSize(doc, sizeof(DocumentType))) return(false);
	if (! HandleValidSize((**doc).print, sizeof(TPrint))) return(false);
	if (! PtrValidSize((**doc).fp, sizeof(FileType))) return(false);
	return(true);
}

/*----------------------------------------------------------------------------*/
/* allocation */
/*----------------------------------------------------------------------------*/

/* create a new document to contain the specified file */
DocumentHandle DocBegin(FileType *fp)
{
	volatile DocumentHandle doc = NULL;	/* new document */
	volatile Boolean propened = false;	/* true if opened printer */
	THPrint print = NULL;					/* new print record */
	FileType *tmpfp = NULL;					/* new file record */
	WindowPtr window = NULL;				/* window to contain document */
	CStr255 title;								/* window's title */
	CStr31 fmt;									/* format for default title */
	static int ndoc;							/* number of documents created */
	
	require(! fp || FileValid(fp));
	TRY {
		DrawCursor(watchCursor);
		
		/* get default document title */
		ResStr31(RLS_FILE, RLS_FILE_UNTITLED, fmt);
		sprintf(title, fmt, ++ndoc);
		FailInfoSet(RLS_ERR_CREATE, title, NULL);

		/* allocate document */
		doc = HandleBeginClear(sizeof(DocumentType));
		(**doc).table = gDocTable;
		tmpfp = PtrBeginClear(sizeof(FileType));
		(**doc).fp = tmpfp;

		/* create print record */
		print = HandleBeginClear(sizeof(TPrint));
		(**doc).print = print;
		PrOpen();
		if (PrError() != fnfErr) {
			FailPrError();
			propened = true;
			PrintDefault(print);
			FailPrError();
			PrClose();
			propened = false;
		}

		/* setup window */
		window = WinGet(RLW_DOCUMENT);
		(**doc).window = window;
		WinKindSet(window, WIN_KIND_DOCUMENT);
		WinDataSet(window, doc);
		WinTitleSet(window, title);
		WinMenuAdd(window);
		WinRegister(window, doc, DocEventTable());
		
		/* finish setting up document */
		DocCreatorSet(doc, AppCreator());
		
		/* set document so application's initialization function can access it */
		if (fp) {
			FileClone(fp, (**doc).fp);
			(**doc).hasfile = true;
		}
		
		/* call application's initialization function */
		if ((**doc).table && (**doc).table->begin)
			(**doc).table->begin(doc);
		
		if (fp) {
			/* clear document now that we've finished
				initializing and then read the file */
			(**doc).hasfile = false;
			DocRead(doc, fp);
		}
		
		FailInfoClear();
	} CATCH {
		if (propened)
			PrClose();
		DocEnd(doc);
	} ENDTRY;
	ensure(DocValid(doc));
	return(doc);
}

/* dispose of the document */
void DocEnd(DocumentHandle doc)
{
	require(! doc || DocValid(doc));
	if (doc) {
		if ((**doc).table && (**doc).table->end)
			(**doc).table->end(doc);
		if (DocWindow(doc)) {
			WinUnregister(DocWindow(doc), doc);
			WinMenuRemove(DocWindow(doc));
			WinEnd(DocWindow(doc));
		}
		PtrEnd((**doc).fp);
		HandleEnd((**doc).print);
		HandleEnd(doc);
		doc = NULL;
	}
	ensure(! DocValid(doc));
}

/*----------------------------------------------------------------------------*/
/* active document */
/*----------------------------------------------------------------------------*/

DocumentHandle DocActive(void)
{
	DocumentHandle doc = NULL;
	
	if (FocusWindow() && WinKind(FocusWindow()) == WIN_KIND_DOCUMENT)
		doc = WinData(FocusWindow());
	ensure(! doc || DocValid(doc));
	return(doc);
}

/*----------------------------------------------------------------------------*/
/* attributes */
/*----------------------------------------------------------------------------*/

void *DocData(DocumentHandle doc)
{
	require(DocValid(doc));
	return((**doc).data);
}

void DocDataSet(DocumentHandle doc, void *data)
{
	require(DocValid(doc));
	(**doc).data = data;
	ensure(DocData(doc) == data);
}

WindowPtr DocWindow(DocumentHandle doc)
{
	require(DocValid(doc));
	return((**doc).window);
}

void DocName(DocumentHandle doc, FileNameType name)
{
	if (DocHasFile(doc))
		strcpy(name, FileName(DocFile(doc)));
	else {
		CStr255 title;
		WinTitle(DocWindow(doc), title);
		strncpy(name, title, sizeof(FileNameType));
		title[sizeof(FileNameType)-1] = 0;
	}
}

Boolean DocDirty(DocumentHandle doc)
{
	Boolean dirty;
	
	require(DocValid(doc));
	dirty = (**doc).dirty;
	if (! dirty && (**doc).table && (**doc).table->dirty)
		dirty = (**doc).table->dirty(doc);
	return(dirty);
}

void DocDirtySet(DocumentHandle doc, Boolean dirty)
{
	require(DocValid(doc));
	(**doc).dirty = dirty;
	ensure(DocDirty(doc) == dirty);
}

Boolean DocHasFile(DocumentHandle doc)
{
	require(DocValid(doc));
	return((**doc).hasfile);
}

FileType *DocFile(DocumentHandle doc)
{
	require(DocHasFile(doc));
	return((**doc).fp);
}

void DocFileSet(DocumentHandle doc, FileType *fp)
{
	require(DocValid(doc));
	FileClone(fp, (**doc).fp);
	WinMenuRemove(DocWindow(doc));
	WinTitleSet(DocWindow(doc), FileName(fp));
	WinMenuAdd(DocWindow(doc));
	(**doc).hasfile = true;
	ensure(DocHasFile(doc) && FilesAreSame(DocFile(doc), fp));
}

OSType DocCreator(DocumentHandle doc)
{
	require(DocValid(doc));
	return((**doc).creator);
}

void DocCreatorSet(DocumentHandle doc, OSType creator)
{
	require(DocValid(doc));
	(**doc).creator = creator;
	ensure(DocCreator(doc) == creator);
}

OSType DocType(DocumentHandle doc)
{
	require(DocValid(doc));
	return((**doc).type);
}

void DocTypeSet(DocumentHandle doc, OSType type)
{
	require(DocValid(doc));
	(**doc).type = type;
	ensure(DocType(doc) == type);
}

THPrint DocPrintRecord(DocumentHandle doc)
{
	require(DocValid(doc));
	return((**doc).print);
}

Boolean DocAutoSave(DocumentHandle doc)
{
	require(DocValid(doc));
	return((**doc).autosave);
}

void DocAutoSaveSet(DocumentHandle doc, Boolean autosave)
{
	require(DocValid(doc));
	(**doc).autosave = autosave;
	ensure(DocAutoSave(doc) == autosave);
}

const DocumentTableType *DocTable(DocumentHandle doc)
{
	require(DocValid(doc));
	return((**doc).table);
}

void DocTableSet(DocumentHandle doc, const DocumentTableType *table)
{
	require(DocValid(doc));
	(**doc).table = table;
	ensure(DocTable(doc) == table);
}
	
/*----------------------------------------------------------------------------*/
/* file IO */
/*----------------------------------------------------------------------------*/

/* read the document from the file */
void DocRead(DocumentHandle doc, FileType *fp)
{
	volatile FileRefType ref = FILE_CLOSED;	/* reference number of resource file */
	volatile Boolean propened = false;	/* true if opened print manager */
	Boolean untitled = false;				/* true if should use an untitled window */
	FInfo info;									/* for getting file's Finder flags */
	
	require(DocValid(doc));
	TRY {
	
		DrawCursor(watchCursor);
		FailInfoSet(RLS_ERR_READ, FileName(fp), NULL);

		/* reset document */
		PrintDefault((**doc).print);

		/* use an untitled window if the file is a stationary document */
		FileInfoGet(fp, &info);
		if ((info.fdFlags & fIsStationery) != 0)
			untitled = true;
		
		/* open file */
		FileOpen(fp, fsRdPerm);
		if (ResFileExists(fp))
			ref = ResFileOpen(fp, fsRdPerm);

		if (ref != FILE_CLOSED) {
			/* read printer record */
			if (ResExists1(RES_PRNT_TYPE, DOC_PRNT)) {
				size_t size = sizeof(TPrint);
				long version = DOC_VERSION;
				ArchiveReadResHandle(RES_PRNT_TYPE, DOC_PRNT,
						DocPrintRecord(doc), &size, &version);
				PrOpen();
				if (PrError() != fnfErr) {
					FailPrError();
					propened = true;
					(void) PrValidate(DocPrintRecord(doc));
					FailPrError();
					PrClose();
					propened = false;
				}
			}
		}
		
		/* read file */
		if ((**doc).table && (**doc).table->read)
			(**doc).table->read(doc, fp, &untitled);
		
		if (ref != FILE_CLOSED) {
			/* restore saved window position (after reading document, in
				case window was zoomed) */
			if (ResExists1(RES_WPOS_TYPE, DOC_WPOS))
				WinZoomRead(DocWindow(doc), DOC_WPOS);
			WinZoomRemember(DocWindow(doc));
		}
		
		/* close file */
		ResFileClose(ref);
		ref = FILE_CLOSED;
		FileClose(fp);

		/* set document's file */
		if (fp != (**doc).fp && ! untitled)
			DocFileSet(doc, fp);

		DocDirtySet(doc, false);
		FailInfoClear();
	} CATCH {
		if (propened)
			PrClose();
		ResFileClose(ref);
		FileClose(fp);
	} ENDTRY;
	ensure(! DocDirty(doc));
}

/* write the document to the file using a safe-save strategy */
void DocWrite(DocumentHandle doc, FileType *docfp)
{
	FileType tmpfile;									/* temporary file for save-save */
	Boolean canfail;									/* saved state of canfail flag */
	volatile FileType *savefp = docfp;			/* file to write to */
	volatile Boolean safesave = false;			/* true if using safe-save */
	volatile FileRefType ref = FILE_CLOSED;	/* resource file */
						
	require(DocValid(doc));
	TRY {

		canfail = MemCanFail(false);
		DrawCursor(watchCursor);
		FailInfoSet(RLS_ERR_WRITE, FileName(docfp), NULL);

		/* If the file exists, use a safe-save strategy: save the
			document to a temporary file, then replace the existing file
			with the temporary file. This avoids corrupting the existing
			file if there's an error. */
		if (FileExists(savefp)) {
			savefp = &tmpfile;
			FileClone(docfp, savefp);
			FileSetTemp(savefp);
			safesave = true;
		}
		check(! FileExists(savefp));
		
		/* create the file */
		FileCreate(savefp, DocCreator(doc), DocType(doc));
		ResFileCreate(savefp);

		if (safesave && ResFileExists(docfp)) {
			/* Preserve the original file's resources. Any resources which
				conflict with our own resources will simply be overwritten
				when we save our file. This allows us to preserve any special
				resources like projector, version, and custom icon resources.*/
			FileCopyRes(docfp, savefp);
		}
	
		/* open file */
		FileVerifySet(savefp, true);
		FileOpen(savefp, fsWrPerm);
		ref = ResFileOpen(savefp, fsWrPerm);

		/* write document */
		ArchiveWriteResHandle(RES_PRNT_TYPE, DOC_PRNT,
				DocPrintRecord(doc), sizeof(TPrint), DOC_VERSION);
		WinZoomWrite(DocWindow(doc), DOC_WPOS);
		FinderWriteApplicationName();
		if ((**doc).table && (**doc).table->write)
			(**doc).table->write(doc, savefp);
		
		/* close file */
		ResFileClose(ref);
		ref = FILE_CLOSED;
		FileClose(savefp);
		
		if (safesave) {
			CInfoPBRec cat;
			/* swap files and set creator, type, and modification date */
			FileExchange(savefp, docfp);
			FileCatalog(docfp, &cat);
			cat.hFileInfo.ioFlFndrInfo.fdCreator = DocCreator(doc);
			cat.hFileInfo.ioFlFndrInfo.fdType = DocType(doc);
			GetDateTime(&cat.hFileInfo.ioFlMdDat);
			FileCatalogSet(docfp, &cat);			
		}
		
		/* successfully wrote file */
		DocDirtySet(doc, false);
		FailInfoClear();
		MemCanFail(canfail);
	} CATCH {
		ResFileClose(ref);
		FileClose(savefp);
		if (safesave && FileExists(savefp))
			FileDelete(savefp);
	} ENDTRY;
	ensure(FileExists(docfp));
	ensure(! DocDirty(doc));
}

/*----------------------------------------------------------------------------*/
/* closing multiple documents */
/*----------------------------------------------------------------------------*/

/* close all unmodified windows, from back window to front window */
static void DocCloseUnmodified(WindowPtr window, MemoryWarningType warning)
{
	if (window) {
		DocCloseUnmodified(WinNext(window), warning);
		if (MemWarning() >= warning) {
			if (WinKind(window) == WIN_KIND_DOCUMENT && ! DocDirty(WinData(window)))
				DocClose(WinData(window), 0);
		}
	}
}

/* close all modified documents, from front window to back window */
static void DocCloseModified(short prompt, MemoryWarningType warning)
{
	WindowPtr window = WinFirst(WIN_LAYER_DOCUMENT);
	WindowPtr next = NULL;
	
	while (window && MemWarning() >= warning) {
		next = WinNext(window);
		if (WinKind(window) == WIN_KIND_DOCUMENT) {
			if (DocDirty(WinData(window)))
				DocClose(WinData(window), prompt);
		}
		window = next;
	}
}

/* First close all unmodified documents and then close all modified documents.
	The 'prompt' parameter should be the index to a 'STR#' resource to
	pass to DocClose which will be displayed in a standard save alert.
	If 'warning' is not MEM_WARNING_NONE, then windows are closed until
	the memory warning condition has become less severe (i.e., the value
	returned by MemWarning is less than the value in 'warning). */
static void DocCloseAll(short prompt, MemoryWarningType warning)
{
	DocCloseUnmodified(WinFirst(WIN_LAYER_DOCUMENT), warning);
	DocCloseModified(prompt, warning);
}

/*----------------------------------------------------------------------------*/
/* File menu operations */
/*----------------------------------------------------------------------------*/

/* create a new document */
void DocNew(void)
{
	WinShow(DocWindow(DocBegin(NULL)));
}

/* select a file and open it */
void DocOpen(void)
{
	FileType file;
	
	FileSFGet(&file, gDocNTypes, gDocTypes);
	WinShow(DocWindow(DocBegin(&file)));
}

/* Close the document. If the document has been modified then a standard
	save dialog is displayed. If the user chooses to save the document then it's
	saved. The window is then closed unless the user clicked the cancel button.
	The 'prompt' parameter is an index to a string resource to display in the
	save changes alert if the document is dirty. */
void DocClose(DocumentHandle doc, short prompt)
{
	FileNameType name;		/* name of document */
	FileNameType appname;	/* name of application */
	short item;					/* item hit in alert */
	Boolean canfail;			/* saved state of canfail flag */
	Boolean saved = false;	/* true if document was saved */
	
	canfail = MemCanFail(false);
	if (DocDirty(doc)) {
		if (DocAutoSave(doc)) {
			/* automatically save document */
			DocSave(doc);
			saved = true;
		}
		else {
			/* ask user, using a standard save changes alert, if the document
				should be saved before it is closed */
			DocName(doc, name);
			WinSelect(DocWindow(doc));
			ScreenSet(ScreenContainingWindow(DocWindow(doc)));
			ResStrLen(RLS_FILE, RLS_FILE_APPLICATION, appname, sizeof(FileNameType));
			item = NotifyCustom(RLA_SAVE, RLS_ALERT, prompt,
						cautionIcon, NULL, NULL, appname, name);
			if (item == cancel)
				FailOSErr(userCanceledErr);
			if (item == ok) {
				DocSave(doc);
				saved = true;
			}
		}
	}
	if (! saved && DocHasFile(doc)) {
		/* save position of document's window since file hasn't been saved,
			but the document's window may have been moved */
		WinZoomSave(DocWindow(doc), DocFile(doc), DOC_WPOS);
	}
	DocEnd(doc);
	MemCanFail(canfail);
}

/* save the document */
void DocSave(DocumentHandle doc)
{
	if (DocHasFile(doc))
		DocWrite(doc, (**doc).fp);
	else
		DocSaveAs(doc);
}

/* select a file and save the document to the file */
void DocSaveAs(DocumentHandle doc)
{
	FileNameType name;
	CStr255 prompt;
	FileType file;
	Boolean canfail;

	canfail = MemCanFail(false);
	DocName(doc, name);
	ResStr(RLS_SF, RLS_SF_SAVE, prompt);
	FileSFPut(&file, prompt, name);
	DocWrite(doc, &file);
	DocFileSet(doc, &file);
	MemCanFail(canfail);
}

/* Revert the document to its last saved version. Confirms with user before
	reverting a modified document. */
void DocRevert(DocumentHandle doc)
{
	short item;					/* item hit in alert */
	FileNameType name;		/* name of document */
	Boolean revert = true;	/* true if document should be reverted */

	require(DocHasFile(doc));
	if (DocDirty(doc)) {
		DocName(doc, name);
		DlgModalOk(RLS_BUTTON, RLS_BUTTON_REVERT);
		item = NotifyCustom(RLA_CANCEL_OK, RLS_ALERT,
					RLS_ALERT_REVERT, cautionIcon, NULL, NULL, name);
		DlgModalOk(0, 0);
		revert = (ok == item);
	}
	if (revert)
		DocRead(doc, DocFile(doc));
}

/* display a page setup dialog to configure the document's printer record */
void DocPageSetup(DocumentHandle doc)
{
	FileNameType name;

	TRY {
		DocName(doc, name);
		FailInfoSet(RLS_ERR_PAGE_SETUP, name, NULL);
		PrOpen();
		FailPrError();
		EventInteractWithUser();
		DrawCursor(arrowCursor);
		if (PrStlDialog(DocPrintRecord(doc)))
			DocDirtySet(doc, true);
		FailPrError();
		FailInfoClear();
	} CLEANUP {
		PrClose();
	} ENDTRY;
}

/* print the document */
void DocPrint(DocumentHandle doc)
{
	DocPrintWithRecord(doc, NULL);
}

/* close all documents before quitting */
void DocQuit(void)
{
	Boolean canfail;
	
	canfail = MemCanFail(false);
	DocCloseAll(RLS_ALERT_QUTTING, MEM_WARNING_NONE);
	MemCanFail(canfail);
}

/*----------------------------------------------------------------------------*/
/* menu handling */
/*----------------------------------------------------------------------------*/

/* adjust menu items */
void DocAdjustMenu(void)
{
	DocumentHandle doc;
	
	doc = DocActive();
	if (! MemWarning()) MenuCmdEnable(CMD_NEW);
	if (! MemWarning()) MenuCmdEnable(CMD_OPEN);
	if (doc) {
		if (DocDirty(doc) || ! DocHasFile(doc)) MenuCmdEnable(CMD_SAVE);
		if (! MemWarning() && DocHasFile(doc) && DocDirty(doc)) MenuCmdEnable(CMD_REVERT);
		if (! MemWarning()) MenuCmdEnable(CMD_PAGE_SETUP);
		if (! MemWarning()) MenuCmdEnable(CMD_PRINT);
		MenuCmdEnable(CMD_SAVE_AS);
		MenuCmdEnable(CMD_CLOSE);
	}
}

/* handle a menu command */
Boolean DocMenu(const MenuPickType *pick)
{
	Boolean handled = false;
	DocumentHandle doc = NULL;
	FileType	fp;
	
	doc = DocActive();
	if (doc) {
		handled = true;
		switch(pick->cmd) {
		case CMD_SAVE:			DocSave(doc); break;
		case CMD_SAVE_AS:		DocSaveAs(doc); break;
		case CMD_REVERT:		DocRevert(doc); break;
		case CMD_CLOSE:		DocClose(doc, RLS_ALERT_CLOSING); break;
		case CMD_PAGE_SETUP:	DocPageSetup(doc); break;
		case CMD_PRINT:		DocPrint(doc); break;
		default:
			handled = false;
			break;
		}
	}
	if (! handled) {
		handled = true;
		switch(pick->cmd) {
		case CMD_NEW:	DocNew(); break;
		case CMD_OPEN:	DocOpen(); break;
		default:
			handled = false;
			break;
		}
	}
	return(handled);
}

/* close documents until there's enough memory */
void DocMemoryLow(void)
{
	DocCloseAll(RLS_ALERT_MEMORY_CLOSING, MEM_WARNING_CRITICAL);
}

/*----------------------------------------------------------------------------*/
/* handler table */
/*----------------------------------------------------------------------------*/

void DocTableRegister(const DocumentTableType *table, short ntypes,
	const OSType *types)
{
	require(ntypes == DOC_ALL_TYPES || ntypes > 0);
	require(ntypes > 0 ? types != NULL : types == NULL);
	gDocTable = table;
	gDocNTypes = ntypes;
	gDocTypes = types;
}

/*----------------------------------------------------------------------------*/
/* high-level event utilities */
/*----------------------------------------------------------------------------*/

/* return of a copy of the printer record */
static THPrint CopyPrintRecord(THPrint src)
{
	volatile THPrint dst = NULL;
	volatile Boolean opened = false;

	require(HandleValidSize(src, sizeof(TPrint)));
	TRY {
		dst = HandleBeginClear(sizeof(TPrint));
		PrOpen();
		FailPrError();
		opened = true;
		PrJobMerge(src, dst);
		FailPrError();
		PrClose();
	} CATCH {
		HandleEnd(dst);
		if (opened)
			PrClose();
	} ENDTRY;
	ensure(HandleValidSize(dst, sizeof(TPrint)));
	return(dst);
}

/* try to start interacting with the user and return true if can interact */
static Boolean CanInteract(void)
{
	volatile Boolean caninteract = true;

	TRY {
		if (caninteract)
			EventInteractWithUser();
	} CATCH {
		if (caninteract) {
			caninteract = false;
			RETRY;
		}
	} ENDTRY;
	return(caninteract);
}

/* true if the file is of a type that can be opened by us */
static Boolean CanOpen(FileType *fp)
{
	OSType type;
	short i;
	
	type = FileTypeGet(fp);
	for (i = 0; i < gDocNTypes; i++) {
		if (gDocTypes[i] == type)
			break;
	}
	return(gDocNTypes == DOC_ALL_TYPES || i < gDocNTypes);
}

/*----------------------------------------------------------------------------*/
/* high-level event interface */
/*----------------------------------------------------------------------------*/

static void DocHLEOpenOne(FileType *fp, void *data)
{
	if (CanOpen(fp))
		WinShow(DocWindow(DocBegin(fp)));
}

static void DocHLEPrintOne(FileType *fp, void *data)
{
	volatile DocumentHandle doc = NULL;
	static THPrint print = NULL;
	
	/* The printer record is initialized to NULL so that a job dialog will
		be displayed for the first file. The settings made for the first
		file will be used for all subsequent files, as specified in IM-II.
		The printer record is disposed of and reset to NULL when all the files
		have been processed. However, if an apple event is being processed
		and no interaction with the user is possible, then the document's
		current printer record is used. */
	TRY {
		if (! fp) {
			HandleEnd(print);
			print = NULL;
		}
		else if (CanOpen(fp)) {
			doc = DocBegin(fp);
			if (CanInteract()) {
				/* print the document using the common printer record, or
					display a page setup dialog if this is the first document */
				DocPrintWithRecord(doc, print);
				/* use a common printer record after the first document is printed */
				if (! print)
					print = CopyPrintRecord(DocPrintRecord(doc));
			}
			else /* print the document using its own printer record */
				DocPrintWithRecord(doc, DocPrintRecord(doc));
		}
	} CLEANUP {
		DocEnd(doc);
	} ENDTRY;
}

static void DocHLEOpen(AEDescList *list, long n, void *data)
{
	FileType file, *fp = &file;
	long i;
	
	for (i = 1; i <= n; i++) {
		AEGetNthFile(list, i, fp);
		DocHLEOpenOne(fp, data);
	}
}

static void DocHLEPrint(AEDescList *list, long n, void *data)
{
	FileType file;
	long i;
	
	for (i = 1; i <= n; i++) {
		AEGetNthFile(list, i, &file);
		DocHLEPrintOne(&file, data);
	}
	DocHLEPrintOne(NULL, data);
}

static void DocHLEOpenApplication(void *data)
{
	DocNew();
}

static void DocHLEQuit(void *data)
{
	DocQuit();
}

/* install the high-level event handlers */
void DocHLEInstall(void)
{
	HLEventInstall(HL_OAPP, DocHLEOpenApplication, NULL);
	HLEventInstall(HL_QUIT, DocHLEQuit, NULL);
	HLEventInstall(HL_OPEN, DocHLEOpen, NULL);
	HLEventInstall(HL_OPEN_ONE, DocHLEOpenOne, NULL);
	HLEventInstall(HL_PRINT, DocHLEPrint, NULL);
	HLEventInstall(HL_PRINT_ONE, DocHLEPrintOne, NULL);
}
