/*	Generic document printing loop.
	
	program_note: my libraries expect the current resource file to be the
	application's resource file, yet the print driver's resource file will be
	searched before the application's or document's resource files...argh
	
	94/01/16 aih
	- split printing loop and text printing code into separate files
	
	93/10/21 aih
	- simplified progress dialog a bit
	
	93/03/23 AIH
	- Split some of the code into a separate file
	
	92/02/29 AIH
	- To simplify things a picture of the document window is not made,
	which means the status window should be a regular modal window,
	not a movable modal window, since the text wouldn't get updated
	properly if it were movable (since update events wouldn't be
	processed)
	
	92/02/20 AIH
	- Updating libraries
	
	91/06/13 AIH
	- A few minor changes to some comments
	
	91/05/21 AIH
	- Improved printing by calculating page heights more accurately and taking
	into account more degenerate cases (ie, a line crossing more than one
	page)
	- Added test preprocessor macro
	- Improved printing using method suggested in IM-II, p156.
	
	91/05/18 AIH
	- The Apple Event manager is called before displaying a printer job
	dialog
	
	91/05/17 AIH
	- The clip region is set wide open before saving the picture of the
	document
	
	91/05/15 AIH
	- Some more fine-tuning
	
	91/05/10-11 Ari Halberstadt (AIH)
	- Adapted for Text Document library
	
	Adapted from my printing code for TextWindow, which was adapted from
	Kevin Calhoun's TextWindows printing code, which he adapted from
	IM-II, pp. 155. */

#include <string.h>
#include <stdio.h>
#include <limits.h>
#include "ApplicationPreferencesLib.h"
#include "DocumentLib.h"
#include "DrawLib.h"
#include "EventLib.h"
#include "ProgressLib.h"
#include "MathLib.h"
#include "MemoryLib.h"
#include "RectangleLib.h"
#include "ResourceLib.h"
#include "ResourceConstantsLib.h"
#include "WindowLib.h"

/* If this is set to true then printing doesn't go to the printer, instead the
	current window is used as the printer. You will see each page displayed
	in the window, which can save a lot of paper and time. */
Boolean gDocPrintTest = false;

/*----------------------------------------------------------------------------*/
/* idle dialog */
/*----------------------------------------------------------------------------*/

/* information needed for the idle loop */
typedef struct {
	DocumentHandle doc;			/* document being printed */
	ProgressHandle progress;	/* progress dialog */
	FileRefType		ref;			/* application's resource file */
	TPrStatus		spool;		/* status of spooled job */
	THPrint			print;		/* print record */
	short				status;		/* status of print job */
	long				npages;		/* number of pages printed */
} IdleType;

static IdleType gIdle;			/* for access from idle procedure */

/* update idle dialog's prompt string */
static void IdleStatus(IdleType *idle, short status)
{
	CStr31 fmt;
	FileNameType name;
	
	idle->status = status;
	DocName(idle->doc, name);
	ResStr31(RLS_PRINTING, status, fmt);
	ProgressPrompt(idle->progress, fmt, name);
}

/* run the idle dialog */
static void IdleRun(IdleType *idle)
{
	FileRefType	ref = FILE_CLOSED;	/* saved resource file reference number */
	GrafPtr port = NULL;					/* saved port */
	volatile Boolean canceled = false;/* true if canceled */
	
	TRY {
		if (! canceled) {
			
			/* use application's resource file, not print driver's file */
			ref = CurResFile();
			UseResFile(idle->ref);
			
			/* save port so we won't inadvertantly change
				it from the printing port */
			GetPort(&port);
			
			/* update page count if spooling */
			if (idle->status == RLS_PRINTING_SPOOLING) {
				 idle->npages = idle->spool.iTotPages + idle->spool.iCurPage +
									 (idle->spool.iCurCopy - 1) *
									 idle->spool.iTotPages / idle->spool.iTotCopies;
			}
			
			/* handle events and update progress indicator */
			ProgressRun(idle->progress, idle->npages);
		}
	} CLEANUP {
		if (port) {
			/* restore environment */
			SetPort(port);
			UseResFile(ref);
		}
	} CATCH {
		if (! canceled && FailReason() == userCanceledErr) {
			IdleStatus(idle, RLS_PRINTING_STOPPING);
			PrSetError(iPrAbort);
			canceled = true;
			RETRY;
		}
	} ENDTRY;
}

/* idle routine called by printing manager */
static pascal void IdleProc(void)
{
	IdleRun(&gIdle);
}

/* set the maximum value for the idle dialog */
static void IdleMaxSet(IdleType *idle, THPrint print,
	short maxpages, short npages, short ncopies)
{
	/* calculate approximate number of pages to print */
	npages = min(maxpages * ncopies, npages * ncopies);
	if ((**print).prJob.bJDocLoop == bSpoolLoop)
		npages *= 2;
	ProgressReset(idle->progress, npages);
}

/* setup the idle dialog */
static void IdleBegin(IdleType *idle, DocumentHandle doc, THPrint print)
{
	memclr(idle, sizeof(IdleType));
	idle->doc = doc;
	idle->print = print;
	
	/* initialize progress dialog */
	idle->progress = ProgressBegin(LONG_MAX, NULL, NULL, 0, 0);
	ProgressOpenDialog(idle->progress, true);
	IdleStatus(idle, RLS_PRINTING_PRINTING);
	
	/* make dialog visible immediately, without the usual delay used
		in the progress dialog */
	WinShow((**idle->progress).dlg);
	
	/* install the idle function so it will be called in the print loop */
	(**print).prJob.pIdleProc = IdleProc;
	
	/* remember the current resource file (which should be the application's
		resource file) */
	idle->ref = CurResFile();
}

/* end the idle loop */
static void IdleEnd(IdleType *idle)
{
	if (idle->print)
		(**idle->print).prJob.pIdleProc = (ProcPtr) NULL;
	if (idle->progress)
		ProgressUpdate(idle->progress);
	ProgressEnd(idle->progress);
	memclr(idle, sizeof(IdleType));
}

/*----------------------------------------------------------------------------*/
/* printing loop */
/*----------------------------------------------------------------------------*/

/* print next page */
static Boolean PrintPage(DocumentHandle doc, TPPrPort port, void *data)
{
	Boolean done = true;
	
	TRY {
		if (! gDocPrintTest)
			PrOpenPage(port, NULL);
		FailPrError();
		if ((**doc).table && (**doc).table->print.page)
			done = (**doc).table->print.page(doc, data);
	} CLEANUP {
		if (! gDocPrintTest)
			PrClosePage(port);
		FailPrError();
	} ENDTRY;
	return(done);
}

/* print a single copy of the document */
static void Print1Copy(DocumentHandle doc, IdleType *idle, THPrint print,
	const Rect *pageRect, short ncopies, Boolean firstCopy,
	short first, short last)
{
	volatile TPPrPort port = NULL;	/* port to draw into */
	volatile void *data = NULL;		/* application defined data */
	Boolean done = false;				/* true if finished printing */
	short npages;							/* number of pages to print */
	
	TRY {
		if (! gDocPrintTest)
			port = PrOpenDoc(print, NULL, NULL);
		FailPrError();
		if ((**doc).table && (**doc).table->print.begin) {
			data = (**doc).table->print.begin(doc, print, port, pageRect,
				first, last);
		}
		/* set maximum value for idle dialog */
		npages = last - first + 1;
		if (firstCopy && (**doc).table && (**doc).table->print.npages) {
			short maxpages = (**doc).table->print.npages(doc, data);
			IdleMaxSet(idle, print, maxpages, npages, ncopies);
		}
		while (npages-- > 0 && ! done) {
			IdleRun(idle);
			done = PrintPage(doc, port, data);
			idle->npages++;
			IdleRun(idle);
		}
	} CLEANUP {
		if ((**doc).table && (**doc).table->print.end)
			(**doc).table->print.end(doc, data);
		if (port && ! gDocPrintTest)
			PrCloseDoc(port);
		FailPrError();
	} ENDTRY;
}

/* print the specified number of copies */
static void PrintNCopies(DocumentHandle doc, IdleType *idle, THPrint print,
	const Rect *pageRect, short ncopies)
{
	short first;	/* first page to print */
	short last;		/* last page to print */
	short copy;		/* copy being printer */
	long height;	/* height of document */
	
	/* get first and last pages */
	first = (**print).prJob.iFstPage;
	last = (**print).prJob.iLstPage;
	(**print).prJob.iFstPage = 1;
	(**print).prJob.iLstPage = SHRT_MAX;
	
	/* print all the copies */
	for (copy = 0; copy < ncopies; copy++)
		Print1Copy(doc, idle, print, pageRect, ncopies, copy == 0, first, last);
		
	/* spool the resulting file */
	if ((**print).prJob.bJDocLoop == bSpoolLoop) {
		IdleStatus(idle, RLS_PRINTING_SPOOLING);
		if (! gDocPrintTest)
			PrPicFile(print, NULL, NULL, NULL, &idle->spool);
		FailPrError();
	}
}

/* Print the document using the specified printer record. If the 'print'
	parameter is NULL then a job dialog is displayed and the document
	is printed using its own printer record. */
void DocPrintWithRecord(DocumentHandle doc, THPrint print)
{
	short	ncopies;			/* number of copies to print */
	Rect pageRect;			/* rectangle to print in */
	FileNameType name;	/* name of document */
	
	require(DocValid(doc));
	TRY {
		
		DocName(doc, name);
		FailInfoSet(RLS_ERR_PRINT, name, NULL);

		PrOpen();
		FailPrError();
		
		/* display job dialog */
		if (! print) {
			print = DocPrintRecord(doc);
			EventInteractWithUser();
			DrawCursor(arrowCursor);
			if (! PrJobDialog(print))
				FailOSErr(iPrAbort);
		}

		DrawCursor(watchCursor);

		/* Call PrValidate so the printing manager will get the name of our
			document, which is assumed to be in the frontmost window. See TN #149. */
		(void) PrValidate(print);
		FailPrError();
		
		/* setup the idle dialog */
		IdleBegin(&gIdle, doc, print);
		
		/* calculate page rectangle */
		pageRect = (**print).prInfo.rPage;
		pageRect.top += AppPrefs()->doc.margin.top * (**print).prInfo.iVRes;
		pageRect.left += AppPrefs()->doc.margin.left * (**print).prInfo.iHRes;
		pageRect.bottom -= AppPrefs()->doc.margin.bottom * (**print).prInfo.iVRes;
		pageRect.right -= AppPrefs()->doc.margin.right * (**print).prInfo.iHRes;

		/* print the requested number of copies */
		if ((**print).prJob.bJDocLoop == bSpoolLoop)
			ncopies = 1;
		else
			ncopies = (**print).prJob.iCopies;
		PrintNCopies(doc, &gIdle, print, &pageRect, ncopies);
		
		FailInfoClear();

	} CLEANUP {
		IdleEnd(&gIdle);
		PrClose();
	} ENDTRY;
	ensure(DocValid(doc));
}
