/*	Functions for displaying error messages in an alert.
	
	Revision History:
	
	93/10/22 aih
	- decided english doesn't support simplicity and wrote complex error message
	function with even more complex comment. i'm sure the comment will be out
	of date in a week, but for now it's a nice example of literate programming...

	93/10/15 aih
	- simplified error displays, removed unneeded functions
	
	93/03/14 AIH
	- Moved error string back into this file
	
	93/03/10 AIH
	- Removed ErrOk macros and a couple of other functions
	- userCanceledErr are not displayed in an error dialog
	
	91/05/09 AIH
	- The fake alert library is used if the regular alert couldn't be
	displayed
	
	91/04/07 AIH
	- Recursive error alerts are not possible, and are prevented using a static
	flag. Otherwise, the attempt to display the error message could itself
	encounter an error, which would in turn cause another error to be
	displayed, which would then encounter an error...
	
	91/01/21 AIH
	- Got rid of some functions I never use, added aliases for ErrOk
	functions, simplified some of the code
	
	91/01/20 Ari Halberstadt
	- Moved ErrorString back into a separate file
	
	91/01/13 Ari Halberstadt
	- Moved strings into resources
	
	91/01/05 Ari Halberstadt
	- Inserted this standard header in all files */

#include <string.h>
#include <stdio.h>
#include "pstr.h"
#include "AlertLib.h"
#include "AlertFakeLib.h"
#include "DialogLib.h"
#include "ErrorLib.h"
#include "ResourceLib.h"
#include "ResourceConstantsLib.h"
#include "StringLib.h"

/* Return error string. This won't call any functions in other
	segments, so it can be used even in case of a segment loader
	error. */
const char *ErrorString(OSErr err)
{
	static Str255 str;
	Handle rsrc = NULL;
	long len = 0;
	
	*str = 0;
	rsrc = GetResource('Estr', err);
	if (! rsrc)
		rsrc = GetResource('Estr', noErr);
	if (rsrc) {
		len = SizeResource(rsrc);
		if (len > sizeof(CStr255))
			len = sizeof(CStr255);
		BlockMove(*rsrc, str, len);
	}
	return(p2cstr(str));
}

/*
	An error message has five components:
	
		1. the action which led to the error (e.g., opening a file)
		2. the object to which the failed operation was applied (e.g.,
		"MyFile")
		3. an explanation of the error (e.g., "The file couldn't be
		opened because it contains illegal data")
		4. a string describing an operating system error (e.g.,
		"Resource not found")
		5. and an operating system error number (e.g., "-108").
		
	The syntax of an error message is (assuming the current [93/10/22]
	versions of the resources and code):
	
	<alert_string> =
		[<action>][<object>][<explanation>][<error>]
	
	<action> =
		  "Could not complete the operation on "
		| "Could not complete the operation."
		| "Could not " <string>
	
	<object> =
		  <empty>
		| '' <string> ''
	
	<explanation> =
		  <empty>
		| <string> '.'
	
	<error> =
		  <empty>
		| <string> '[' <number> ']'
					
	ErrorDisplay(oserr, action, object, explanation)
	
	Call:		ErrorDisplay(0, 0, NULL, NULL)
	Output:	"Could not complete the operation."
	
	Call:		ErrorDisplay(-192, 0, NULL, NULL)
	Output:	"Could not complete the operation. A resource was not found [-192]."
	
	Call:		ErrorDisplay(0, RLS_ERR_REINVENT_THE_WHEEL, NULL, NULL)
	Output:	"Could not reinvent the wheel."

	Call:		ErrorDisplay(0, RLS_ERR_OPEN, "MyFile", NULL)
	Output:	"Could not open MyFile."
	
	Call:		ErrorDisplay(-49, RLS_ERR_WRITE, "MyFile", NULL)
	Output:	"Could not write MyFile. The file is already open for writing [-49]."
	
	Call:		ErrorDisplay(-108, 0, "MyFile", NULL)
	Output:	"Could not complete the operation on MyFile. Out of memory [-108]."
	
	Call:		ErrorDisplay(0, 0, "Important Data", "The data in the file are damaged")
	Output:	"Could not complete the operation on Important Data. The data in
				the file are damaged."
	
	Call:		ErrorDisplay(-43, RLS_ERR_OPEN, "Term Paper", "The document no
				longer exists because you just deleted it")
	Output:	"Could not open Term Paper. The document no longer exists because
				you just deleted it. The file wasn't found [-43]."
*/
void ErrorDisplay(OSErr err, short action,
	const char *object, const char *explanation)
{
	CStr63 fmtstr;	/* passed to sprintf to format error message */
	CStr31 actstr;	/* string describing attempted action */
	CStr255 tmpstr;/* temporary formatted string */
	CStr255 alertstr;/* formatted string to display in alert */
	volatile Boolean retry = false; /* true if exception occurred */
		
	TRY {
		if (retry)
			AlertFakeFailed(err, c2pstrcpy((StringPtr) alertstr, explanation));
		else {
			*alertstr = 0;
			
			/* format action */
			*actstr = 0;
			if (action) {
				ResStrLen(RLS_ERR, action, actstr, sizeof(actstr));
				ResStrLen(RLS_ERRFMT, RLS_ERRFMT_ACTION, fmtstr, sizeof(fmtstr));
			}
			else if (object && *object)
				ResStrLen(RLS_ERRFMT, RLS_ERRFMT_OPERATION_ON, fmtstr, sizeof(fmtstr));
			else
				ResStrLen(RLS_ERRFMT, RLS_ERRFMT_OPERATION, fmtstr, sizeof(fmtstr));
			sprintf(tmpstr, fmtstr, actstr);
			check(strlen(alertstr) + strlen(tmpstr) < sizeof(alertstr));
			strcat(alertstr, tmpstr);
			
			/* format object */
			if (object && *object) {
				ResStrLen(RLS_ERRFMT, RLS_ERRFMT_OBJECT, fmtstr, sizeof(fmtstr));
				sprintf(tmpstr, fmtstr, object);
				check(strlen(alertstr) + strlen(tmpstr) < sizeof(alertstr));
				strcat(alertstr, tmpstr);
			}
						
			/* format explanation */
			if (explanation && *explanation) {
				ResStrLen(RLS_ERRFMT, RLS_ERRFMT_EXPLANATION, fmtstr, sizeof(fmtstr));
				sprintf(tmpstr, fmtstr, explanation);
				check(strlen(alertstr) + strlen(tmpstr) < sizeof(alertstr));
				strcat(alertstr, tmpstr);
			}

			/* format error number */
			if (err) {
				ResStrLen(RLS_ERRFMT, RLS_ERRFMT_OSERR, fmtstr, sizeof(fmtstr));
				sprintf(tmpstr, fmtstr, ErrorString(err), err);
				check(strlen(alertstr) + strlen(tmpstr) < sizeof(alertstr));
				strcat(alertstr, tmpstr);
			}

			(void) AlertStop(RLA_OK, alertstr);
		}
	} CATCH {
		if (! retry) {
			retry = true;
			RETRY;
		}
	} ENDTRY;
}
