/* display a list of icons in a floating window.
	93/01/04 aih changed name from ToolBar to IconList;
					 removed resize task; use standard scroll bar width
	93/12/15 aih added window id param to begin proc
	93/12/13 aih cells are selected with an outline rather than inversion
	93/12/07 aih created */
	
#include "DrawLib.h"
#include "IconListLib.h"
#include "ListLib.h"
#include "MemoryLib.h"
#include "RectangleLib.h"
#include "ResourceLib.h"
#include "ScrollBarLib.h"
#include "WindowLib.h"

#define ICON_SIZE				(32)	/* size of an icon */
#define ICON_MARGIN			(6)	/* margin to edge of cell */
#define SMALL_ICON_SIZE		(16)	/* size of a small icon */
#define SMALL_ICON_MARGIN	(4)	/* margin to edge of cell */

/* data stored in a cell */
typedef struct {
	short id;
} IconListDataType;

/* Draw the indexed icon with its top left corner at the specified point.
	Size is the icon's dimensions in pixels. If 'invert' is
	true then the icon is also inverted. */
static void DrawGenericIcon(Handle hicon, Point pt, short index, short size,
	Boolean iconMask, Boolean invert)
{
	SignedByte	state = 0;		/* saved state of icon handle */
	BitMap		iconBits;		/* bit map of icon */
	BitMap		maskBits;		/* bit map of icon's mask */
	Rect			srcRect;			/* source rectangle for copy */
	Rect			dstRect;			/* destination rectangle for copy */
	GrafPtr		port = NULL;	/* port to draw to */
	char			mask[128];		/* storage for mask if need to calculate it */
	char			*icon;			/* dereferenced icon */
	short			width, height;	/* width and height of icon */
	short			nbytes;			/* number of bytes in an icon */
	
	require(size >= CHAR_BIT * 2);
	require(size % CHAR_BIT == 0);
	require(size % 2 * CHAR_BIT == 0);
	require(HandleSize(hicon) >= index * (width * height / CHAR_BIT));
	state = HandleLock(hicon);
	GetPort(&port);
	width = height = size;
	nbytes = width * height / CHAR_BIT;
	icon = *hicon + index * nbytes;
	SetRect(&dstRect, pt.h, pt.v, pt.h + width, pt.v + height);
	SetRect(&srcRect, 0, 0, width, height);
	maskBits.rowBytes = iconBits.rowBytes = width / CHAR_BIT;
	maskBits.bounds = iconBits.bounds = srcRect;
	iconBits.baseAddr = icon;
	if (iconMask && HandleSize(hicon) >= 2 * nbytes) {
		check(index == 0);
		maskBits.baseAddr = icon + nbytes;
	}
	else {
		maskBits.baseAddr = mask;
		CalcMask(icon, mask, iconBits.rowBytes, maskBits.rowBytes, height,
			width / (CHAR_BIT * 2));
	}
	CopyBits(&maskBits, &port->portBits, &srcRect, &dstRect, srcBic, NULL);
	CopyBits(&iconBits, &port->portBits, &srcRect, &dstRect, srcOr, NULL);
	if (invert) {
		SetHiliteMode();
		CopyBits(&maskBits, &port->portBits, &srcRect, &dstRect, srcXor, NULL);
	}
	HandleRestore(hicon, state);
}

/* list definition function for drawing lists of icons */
static pascal void ldef(short msg, Boolean select, const Rect *lrect, Cell cell,
	short offset, short length, ListHandle list)
{
	short id = 0;
	Handle icon = NULL;
	Rect r = *lrect;
	IconListDataType data;
	IconListHandle bar = (IconListHandle) (**list).userHandle;
	
	require(IconListValid(bar));
	switch (msg) {
	case lDrawMsg:
	case lHiliteMsg:
		check(length == sizeof(IconListDataType));
		data = *(IconListDataType *) (*(**list).cells + offset);
		InsetRect(&r, ((**bar).cellSize - (**bar).iconSize) / 2,
						  ((**bar).cellSize - (**bar).iconSize) / 2);
		DrawGenericIcon(ResGet((**bar).type, data.id), topLeft(r), 0,
			(**bar).iconSize, (**bar).iconMask, false);
		InsetRect(&r, -3, -3);
		BorderDraw(&r, 2, 1, select);
		break;
	}
}

/* true if a valid icon list */
Boolean IconListValid(IconListHandle bar)
{
	if (! HandleValidSize(bar, sizeof(IconListType))) return(false);
	if ((**bar).cellSize < (**bar).iconSize) return(false);
	return(true);
}

/* create an empty list */
static void IconListCreateList(IconListHandle bar)
{
	Rect view, bounds = { 0, 0, 1, 0 };
	Cell size;
	ListHandle list = NULL;
	
	size.h = size.v = (**bar).cellSize;
	WinPortRect((**bar).window, &view);
	WinSize((**bar).window, RectWidth(&view), SBAR_WIDTH + (**bar).cellSize);
	WinPortRect((**bar).window, &view);
	InsetRect(&view, -1, -1);
	list = ListBegin(&view, &bounds, size, 0, (**bar).window,
				false, true, true, false, false);
	(**bar).list = list;
	check((**list).userHandle == NULL);
	(**list).userHandle = (Handle) bar;
	ListDefinitionSet((**bar).list, ldef);
	ListActivate((**bar).list, true);
}

/* initialize the list */
static void IconListReadList(IconListHandle bar)
{
	short i = 0;				/* index to resources */
	short n = 0;				/* number of resources */
	Handle icon = NULL;		/* handle to icon resource */
	short id = 0;				/* id of resource */
	OSType type = 0;			/* type of resource (ignored) */
	Str255 name;				/* name of resource (ignored) */
	Cell cell;					/* index to cell in list */
	IconListDataType data;	/* data installed in list */

	LDoDraw(false, (**bar).list);
	n = CountResources((**bar).type);
	LAddColumn(n, 0, (**bar).list);
	cell.h = cell.v = 0;
	for (i = 1; i <= n; i++) {
		icon = ResGetIndexed((**bar).type, i);
		GetResInfo(icon, &id, &type, name);
		FailResError();
		data.id = id;
		cell.h = i - 1;
		LSetCell(&data, sizeof(IconListDataType), cell, (**bar).list);
	}
	LDoDraw(true, (**bar).list);
}

/* create a new icon list */
IconListHandle IconListBegin(OSType type, short id)
{
	volatile IconListHandle bar = NULL;
	WindowPtr window = NULL;
	
	TRY {
		bar = HandleBeginClear(sizeof(IconListType));
		window = WinGet(id);
		(**bar).window = window;
		(**bar).type = type;
		switch (type) {
		case ICON_LIST_ICON_LIST:
			(**bar).iconMask = true;
			/* no break */
		case ICON_LIST_ICON:
			(**bar).iconSize = ICON_SIZE;
			(**bar).cellSize = ICON_SIZE + 6;
			break;
		case ICON_LIST_SMALL_ICON:
			(**bar).iconSize = SMALL_ICON_SIZE;
			(**bar).cellSize = SMALL_ICON_SIZE + 2 * SMALL_ICON_MARGIN;
			(**bar).iconMask = true;
			break;
		default:
			check(false);
		}
		IconListCreateList(bar);
		IconListReadList(bar);
		WinRegister(window, bar, IconListEventTable());
		WinRegisterID(window, bar, ICON_LIST_ID);
		WinShow(window);
	} CATCH {
		IconListEnd(bar);
	} ENDTRY;
	ensure(IconListValid(bar));
}

/* dispose of the icon list */
void IconListEnd(IconListHandle bar)
{
	if (bar) {
		WinUnregister((**bar).window, bar);
		ListEnd((**bar).list);
		WinEnd((**bar).window);
		HandleEnd(bar);
	}
}

/* true if point is in icon list */
Boolean IconListWithin(IconListHandle bar, Point where)
{
	require(IconListValid(bar));
	return(ListWithin((**bar).list, where));
}

/* handle a click */
Boolean IconListMouseDown(IconListHandle bar, EventRecord *event)
{
	Cell cell;					/* cell index */
	Rect visible;				/* visible cells */
	Point where;				/* mouse point in local coordinates */
	IconListDataType data;	/* data of clicked cell */
	short len;					/* length of data */
	short offset;				/* offset into icon's mask */
	Rect r;						/* cell's rectangle */
	Handle icon;				/* handle to icon */
	Boolean hit = false;		/* true if hit a cell */
	char mask[128];			/* icon's mask */
	
	require(IconListValid(bar));
	where = event->where;
	GlobalToPort(&where, (**bar).window);
	LClick(where, event->modifiers, (**bar).list);
	return(true);
#ifdef NOT_READY
	if (ListHScroll(list) && TestControl(ListHScroll(list), where)) {
		LClick(where, event->modifiers, list);
		return(true);
	}
	if (ListVScroll(list) && TestControl(ListVScroll(list), where)) {
		LClick(where, event->modifiers, list);
		return(true);
	}
	visible = (**list).visible;
	for (cell.h = visible.left; cell.h < visible.right; cell.h++) {
		for (cell.v = visible.top; cell.v < visible.bottom; cell.v++) {
			LRect(&r, cell, list);
			if (PtInRect(where, &r)) {
				offset = (where.h - r.left) * (where.v - r.top);
				check(0 <= offset && offset < 128 * CHAR_BIT);
				len = sizeof(IconListDataType);
				LGetCell(&data, &len, cell, list);
				check(len == sizeof(IconListDataType));
				icon = ResGet(ICON_LIST_ICON_LIST, data.icon);
				/* CalcMask won't move or purge memory */
				CalcMask(*icon, mask, 4, 4, 32, 2);
				hit = BitTst(mask, offset);
				if (hit)
					LClick(where, event->modifiers, list);
				return(hit);
			}
		}
	}
	return(false);
#endif /* NOT_READY */
}

/* set minimum and maximum size */
void IconListGrow(IconListHandle bar, Rect *size)
{
	require(IconListValid(bar));
	size->left = (**bar).cellSize;
	size->top = size->bottom = SBAR_WIDTH + (**bar).cellSize + 1;
	ensure(IconListValid(bar));
}

/* close icon lists if memory is low */
void IconListMemoryLow(void)
{
	WindowPtr window = WinFirst(WIN_LAYER_FLOAT);
		
	while (window && MemWarning() >= MEM_WARNING_VERY_LOW) {
		WindowPtr next = WinNext(window);
		IconListEnd(WinRegisteredID(window, ICON_LIST_ID));
		window = next;
	}
}
