/* Interface to the Memory Manager.

	Revision History:
	
	94/01/10 aih
	- improved some pre- and postconditions
	- changed code in all files so that validation of handles would also validate
	that the handle wasn't purged; this will allow eventual modification of
	this library to support purged handles
	
	94/01/08 aih
	- if memory can't be allocated in the application heap then temporary
	memory is used if it's available
	
	94/01/04 aih
	- added region allocation functions
	- this file must be in the main segment, which keeps the grow zone
	function locked and makes sure that functions like HandleNoPurge
	won't accidentally purge a resource due to segment loading
	- segments are unloaded when memory can't be allocated
	
	94/01/02 aih
	- added functions for allocating a cleared handle or pointer
	
	93/11/05 aih
	- during debugging, after reducing the size of a handle or a pointer,
	unused trailing bytes are filled with garbage, and after increasing
	the size of a handle or pointer the new bytes are filled with garbage
	
	93/10/30 aih
	- profiling is turned off in the grow zone function
	
	93/10/17 aih
	- added function to resize pointers
	
	93/03/25 AIH
	- The segment containing the grow zone function is locked
	
	93/03/12 AIH
	- Removed memory limit/unlimit functions
	- Added function to indicate that memory allocation must not fail

	93/03/06 AIH
	- Modified memory warning functions
	
	92/03/04 AIH
	- Changed names of function prefixes from "MemH" to "Handle"
	and from "MemPtr" to "Ptr"
	
	92/02/26 AIH
	- Added function to return a number indicating how low memory is
	
	91/07/28 AIH
	- Fixed bug in function for disposing of a handle which could crash
	if the handle had been purged or if memset could move memory.

	91/06/01 AIH
	- Added function to duplicate a handle
	
	91/05/05 AIH
	- At most half of the free memory is ever used for the memory reserves,
	so that they don't get stuck like a non-relocatable block
	
	91/04/22 AIH
	- StripAddress is called when calculating the size of the stack and heap
	- Made memory validation better
	
	91/04/19 AIH
	- A function is used to determine if the Macintosh is in 32-bit mode
	
	91/04/09 AIH
	- Added function for setting a handle's and a pointer's size
	
	91/03/23 AIH
	- Wrote new memory allocation functions
	
	91/03/20 AIH
	- Added functions to allocate memory after setting a memory limit
	
	91/03/17 AIH
	- Memory validation functions are only compiled if we're not debugging,
	otherwise macros which always return true are used
	- Memory reserves and grow zone functions are only compiled into an
	application, not into code resources
	- Added limit function for memory reserves
	
	91/03/15 AIH
	- Added check that handle is in the current zone to HandleCheck
	
	91/01/31 AIH
	- Moved functions for getting stack and heap sizes from MacLib into
	this library
	
	91/01/21 AIH
	- Added system zone checks to HandleCheck
	
	91/01/05 Ari Halberstadt (AIH)
	- Inserted this standard header in all files */

#include <limits.h>
#include <string.h>
#include "ByteLib.h"
#include "GlobalLib.h"
#include "LLHandleLib.h"
#include "LowMemLib.h"
#include "MacLib.h"
#include "MemoryLib.h"
#include "SegmentLib.h"

/*---------------------------------------------------------------------------*/
/* memory reserves */
/*---------------------------------------------------------------------------*/

struct {
	Handle reserve;	/* handle to reserve memory */
	size_t reservesz;	/* size of memory reserves */
	size_t cushionsz;	/* size of memory cushion */
	Boolean canfail;	/* true if a memory request can fail */
} gMemory;

/* return the size of the memory reserves */
static size_t MemReservesSize(void)
{
	return(gMemory.reserve ? GetHandleSize(gMemory.reserve) : 0);
}

/* true if the memory reserves are not full */
static Boolean MemReservesLow(void)
{
	return(MemReservesSize() < gMemory.reservesz);
}

/* release some space from the reserved memory */
static pascal long MemGrowZone(Size needed)
{
	SignedByte state;	/* saved state of svhnd */
	Handle svhnd;		/* handle not to move */
	long result;		/* number of bytes freed */
	
	/* setup global variable base register and lock grow zone save handle */
	GlobalSetup();
	result = 0;
	svhnd = GZSaveHnd();
	if (svhnd) {
		state = HGetState(svhnd);
		HLock(svhnd);
	}
	
	/* release the reserve memory handle */
	if (gMemory.reserve && gMemory.reserve != svhnd) {
		result = GetHandleSize(gMemory.reserve);
		EmptyHandle(gMemory.reserve);
	}
	
	/* restore environment */
	if (svhnd)
		HSetState(svhnd, state);
	GlobalRestore();
	return(result);
}

/* replenish the reserve memory; call this function periodically,
	once every time through the main event loop, or on every null
	event */
static void MemReservesReplenish(void)
{
	if (gMemory.reserve && MemReservesLow()) {
		SetGrowZone(NULL);
		ReallocateHandle(gMemory.reserve, gMemory.reservesz);
		SetGrowZone((ProcPtr) MemGrowZone);
	}
}

/* Initialize memory reserves by setting aside 'reserve' bytes and
	installing a grow zone function. */
void MemInit(size_t cushion, size_t reserve)
{
	GlobalRemember(); /* for grow zone function */
	gMemory.reserve = NewHandle(0); /* can't use HandleBegin */
	if (gMemory.reserve) {
		gMemory.cushionsz = cushion;
		gMemory.reservesz = reserve;
		gMemory.canfail = true;
		SetGrowZone((ProcPtr) MemGrowZone);
		MemReservesReplenish();
	}
}

/* Return indication of whether there's a memory shortage and, if so,
	how severe the shortage is. When recovering from a memory shortage
	by releasing unneeded memory, call this function periodically to see
	if memory is still low. Each time this function is called it
	attempts to replenish the memory reserves before checking for
	a low memory condition. */
MemoryWarningType MemWarning(void)
{
	MemReservesReplenish();
	if (MemReservesSize() == 0) return(MEM_WARNING_CRITICAL);
	if (MemReservesLow()) return(MEM_WARNING_VERY_LOW);
	if (MemContiguousSize() < gMemory.cushionsz) return(MEM_WARNING_LOW);
	return(MEM_WARNING_NONE);
}

/* True if 'n' bytes are available. Doesn't unload segments or consider the
	memory cushion. Primarily written for use from within the LoadSeg patch. */
Boolean MemAvailableNoCushion(size_t n)
{
	return(! gMemory.canfail || ! MemReservesLow());
}

/* true if 'n' bytes are available (based on "Inside Macintosh: Memory", p1-47) */
Boolean MemAvailable(size_t n)
{
	Boolean available = false;
	
	//debug_note
	//{ static long count; long grow;
	//	if (count++ % 5) {
	//		SegmentsUnload();
	//		MaxMem(&grow);
	//	}
	//}
	
	if (! gMemory.canfail)
		available = true;
	else {
		if (MemReservesLow() || n + gMemory.cushionsz > MemContiguousSize()) {
			/* memory is low, but we may be able to free some
				memory by unloading unneeded segments */
			//SegmentsUnload(); // debug_note: this occasionally destroys return addresses on the stack
			MemReservesReplenish();
		}
		if (! MemReservesLow() && n + gMemory.cushionsz <= MemContiguousSize())
			available = true;
	}
	return(available);
}

/* fail if memory isn't available */
void MemCheck(size_t n)
{
	if (! MemAvailable(n))
		FailOSErr(memFullErr);
}

/* set the canfail flag; return old value */
Boolean MemCanFail(Boolean canfail)
{
	Boolean old = gMemory.canfail;
	gMemory.canfail = canfail;
	return(old);
}

/*---------------------------------------------------------------------------*/
/* validating memory */
/*---------------------------------------------------------------------------*/

#define isodd(n) (((long) (n) & 1) != 0)

/* true if a valid relocatable or nonrelocatable block */
static Boolean BlockValid(void *block, Boolean relocatable)
{
	if (! block) return(false);
	if (isodd(block)) return(false);
	if (relocatable) {
		Ptr p = *(Handle) block;
		if (! p) return(false);
		if (isodd(p)) return(false);
	}
	return(true);
}

/* true if a valid pointer to any type of data, not just a pointer allocated
	with the memory manager */
Boolean MemValid(const void *p)
{
	return(p != NULL && ((long) p & 1) == 0);
}

/*---------------------------------------------------------------------------*/
/* operations on handlers */
/*---------------------------------------------------------------------------*/

/* true if the handle is valid */
Boolean HandleValid(void *h)
{
	return(BlockValid(h, true));
}

/* true if the handle is valid and is at least n bytes */
Boolean HandleValidSize(void *h, size_t n)
{
	return(HandleValid(h) && GetHandleSize(h) >= n);
}

/* return the master pointer flags for the handle */
SignedByte HandleState(void *h)
{
	SignedByte state;
	
	require(HandleValid(h));
	state = HGetState(h);
	FailMemError();
	return(state);
}

/* true if the handle is a handle to a resource */
Boolean HandleIsResource(void *h)
{
	require(HandleValid(h));
	return(HandleState(h) & (1<<5));
}

/* lock the handle; the handle should be unlocked by calling HandleRestore */
SignedByte HandleLock(void *h)
{
	SignedByte state;
	
	require(HandleValid(h));
	state = HandleState(h);
	HLock(h);
	FailMemError();
	return(state);
}

/* same as HandleLock, but calls MoveHHi before locking the handle */
SignedByte HandleLockHi(void *h)
{
	require(HandleValid(h));
	MoveHHi(h);
	return(HandleLock(h));
}

/* make the handle unpurgeable and return its state before HNoPurge was called */
SignedByte HandleNoPurge(void *h)
{
	SignedByte state;
	
	require(HandleValid(h));
	state = HandleState(h);
	HNoPurge(h);
	FailMemError();
	return(state);
}

/* true if the handle has been purged */
static Boolean HandlePurged(void *h)
{
	return(! *(Handle) h);
}

/* restore the state of the handle */
void HandleRestore(void *h, SignedByte state)
{
	require(HandleValid(h));
	HSetState(h, state);
	FailMemError();
}

/* unlock the handle */
void HandleUnlock(void *h)
{
	require(HandleValid(h));
	HUnlock(h);
	FailMemError();
}

/* make the handle purgeable */
void HandlePurge(void *h)
{
	require(HandleValid(h));
	HPurge(h);
	FailMemError();
}

/* Allocate a new handle. The handle may be allocated from temporary memory,
	so you should not lock the handle across calls to GetNextEvent or
	WaitNextEvent (IM Memory, p2-10) */
static void *HandleNew(size_t n)
{
	OSErr err;
	Handle h;
	
	if (MemAvailable(n + MEM_HDRSZ))
		h = NewHandle(n);
	else if (MacHasTempMem()) {
		h = TempNewHandle(n, &err);
		FailOSErr(err);
	}
	else
		FailOSErr(memFullErr);
	if (RELEASE_LEVEL == RELEASE_DEV) {
		(void) HandleLock(h);
		memset(*h, MEM_JUNK, n);
		HandleUnlock(h);
	}
	ensure(HandleSize(h) == n);
	return(h);
}

/* allocate a new handle */
void *HandleBegin(size_t n)
{
	return(HandleNew(n));
}

/* allocate a new handle filled with nulls */
void *HandleBeginClear(size_t n)
{
	Handle h;
	
	h = HandleNew(n);
	(void) HandleLock(h);
	memclr(*h, n);
	HandleUnlock(h);
	return(h);
}

/* a protected version of GetHandleSize */
size_t HandleSize(void *h)
{
	size_t size;
	
	require(HandleValid(h));
	size = GetHandleSize(h);
	FailMemError();
	ensure(size >= 0);
	return(size);
}

/* a protected version of SetHandleSize */
void HandleSizeSet(void *h, size_t n)
{
	volatile SignedByte state = 0;
	size_t oldsize = 0;
	
	require(HandleValid(h));
	TRY {
		state = HandleNoPurge(h);
		oldsize = HandleSize(h);
		if (n > oldsize)
			MemCheck(n - oldsize);
		if (RELEASE_LEVEL == RELEASE_DEV && n < oldsize) {
			/* fill discarded bytes with junk */
			(void) HandleLock(h);
			memset(*(Handle) h + n, MEM_JUNK, oldsize - n);
			HandleUnlock(h);
		}
		SetHandleSize(h, n);
		FailMemError();
		if (RELEASE_LEVEL == RELEASE_DEV && n > oldsize) {
			/* fill new bytes with junk */
			(void) HandleLock(h);
			memset(*(Handle) h + oldsize, MEM_JUNK, n - oldsize);
			HandleUnlock(h);
		}
	} CLEANUP {
		HandleRestore(h, state);
	} ENDTRY;
	ensure(HandleSize(h) == n);
}

/* return a copy of the source handle */
void *HandleCopy(void *src)
{
	size_t n = 0;
	Handle dst = NULL;
	SignedByte state = 0;
	
	require(HandleValid(src));
	state = HandleNoPurge(src);
	n = HandleSize(src);
	dst = HandleBegin(n);
	BlockMove(*(Handle) src, *dst, n);
	HandleRestore(src, state);
	return(dst);
}
 
/* Release the memory used by the handle. Also works with handles to
	resources, which is a real boon when recovering from a failed
	AddResource call. */
void HandleEnd(void *h)
{
	require(! h || HandleValid(h));
	if (h) {
		if (HandleIsResource(h))
			ReleaseResource(h);
		else {
			if (RELEASE_LEVEL == RELEASE_DEV) {
				(void) HandleLock(h);
				memset(*(Handle) h, MEM_JUNK, HandleSize(h));
			}
			DisposHandle(h);
		}
		h = NULL;
	}
	ensure(! HandleValid(h));
}

/*---------------------------------------------------------------------------*/
/* operations on pointers */
/*---------------------------------------------------------------------------*/

/* true if the pointer is valid */
Boolean PtrValid(void *p)
{
	return(BlockValid(p, false));
}

/* true if the pointer is valid and is at least n bytes */
Boolean PtrValidSize(void *p, size_t n)
{
	return(PtrValid(p) && GetPtrSize(p) >= n);
}

/* allocate a new pointer */
void *PtrBegin(size_t n)
{
	Ptr p;
	
	MemCheck(n + MEM_HDRSZ);
	p = NewPtr(n);
	FailMemError();
	if (RELEASE_LEVEL == RELEASE_DEV)
		memset(p, MEM_JUNK, n);
	ensure(PtrSize(p) == n);
	return(p);
}

/* allocate a new pointer filled with nulls */
void *PtrBeginClear(size_t n)
{
	return(memclr(PtrBegin(n), n));
}

/* return size of pointer */
size_t PtrSize(void *p)
{
	size_t size;
	
	require(PtrValid(p));
	size = GetPtrSize(p);
	FailMemError();
	ensure(size >= 0);
	return(size);
}

/* resize the pointer */
void PtrSizeSet(void *p, size_t n)
{
	volatile SignedByte state = 0;
	size_t oldsize = PtrSize(p);
	
	if (n > oldsize)
		MemCheck(n - PtrSize(p));
	if (RELEASE_LEVEL == RELEASE_DEV && n < oldsize)
		memset((Ptr) p + n, MEM_JUNK, oldsize - n);
	SetPtrSize(p, n);
	FailMemError();
	if (RELEASE_LEVEL == RELEASE_DEV && n > oldsize)
		memset((Ptr) p + oldsize, MEM_JUNK, n - oldsize);
	ensure(PtrSize(p) == n);
}

/* return a handle to a copy of the pointer */
void *PtrToHandle(void *p)
{
	Handle h = NULL;
	
	require(PtrValid(p));
	MemCheck(PtrSize(p) + MEM_HDRSZ);
	PtrToHand(p, &h, PtrSize(p));
	FailMemError();
	ensure(HandleSize(h) == PtrSize(p));
	return(h);
}

/* dispose of the pointer */
void PtrEnd(void *p)
{
	require(! p || PtrValid(p));
	if (p) {
		if (RELEASE_LEVEL == RELEASE_DEV)
			memset(p, MEM_JUNK, GetPtrSize(p));
		DisposPtr(p);
		p = NULL;
	}
	ensure(! PtrValid(p));
}

/*---------------------------------------------------------------------------*/
/* operations on regions */
/*---------------------------------------------------------------------------*/

/* list of recently created regions */
static RgnHandle gRegionList;

/* true if a valid region */
Boolean ValidateRgn(RgnHandle rgn)
{
	if (! HandleValidSize(rgn, sizeof(Region))) return(false);
	if (HandleSize(rgn) != (**rgn).rgnSize) return(false);
	if ((**rgn).rgnBBox.top > (**rgn).rgnBBox.bottom) return(false);
	if ((**rgn).rgnBBox.left > (**rgn).rgnBBox.right) return(false);
	return(true);
}

/* return a new empty region */
RgnHandle BeginRgn(void)
{
	Rect empty = { 0, 0, 0, 0 };
	RgnHandle rgn;
	
	if (gRegionList) {
		rgn = gRegionList;
		gRegionList = LLHDelete(gRegionList, gRegionList);
		/* can't use SetEmptyRgn, so set fields directly */
		(**rgn).rgnSize = HandleSize(rgn);
		(**rgn).rgnBBox = empty;
	}
	else {
		rgn = NewRgn();
		FailNIL(rgn);
	}
	ensure(ValidateRgn(rgn) && EmptyRgn(rgn));
	return(rgn);
}

/* End use of the region. The region isn't actually disposed of, it's
	just set to the empty region and added to our list of regions. This
	makes region allocation very quick, which is useful since a small
	number of regions are typically allocated for short periods of
	time in drawing intensive routines. This is much better than
	keeping a bunch of global utility handles around, with no guarantee
	that they aren't already in use. */
void EndRgn(RgnHandle rgn)
{
	require(! rgn || ValidateRgn(rgn));
	if (rgn) {
		SetEmptyRgn(rgn);
		gRegionList = LLHInsert(gRegionList, rgn);
	}
}

/*---------------------------------------------------------------------------*/
/* heap and stack information */
/*---------------------------------------------------------------------------*/

/* return size of application's stack */
size_t MemStackSize(void)
{
	return(StripAddress(GetCurStackBase()) - StripAddress(GetApplLimit()));
}

/* return size of application's heap */
size_t MemHeapSize(void)
{
	return(StripAddress(GetHeapEnd()) - StripAddress((void *) ApplicZone()));
}

/* maximum contiguous block that could be allocated */
size_t MemContiguousSize(void)
{
	long total, contig;
	PurgeSpace(&total, &contig);
	return(contig);
}
