/*
	93/12/26 aih
	- split out of DebugLib.c
	- made window smaller; option click a view to disable it */

#include <stdio.h>
#include <string.h>
#include "DialogLib.h"
#include "DrawLib.h"
#include "MemoryLib.h"
#include "MemoryWindowLib.h"
#include "PreferencesLib.h"
#include "ResourceConstantsLib.h"
#include "SegmentLib.h"

/* font characteristics */
#define MEMORY_VIEW_FONT	(monaco)
#define MEMORY_VIEW_SIZE	(9)
#define MEMORY_VIEW_FACE	(0)
#define MEMORY_WINDOW_FONT	(monaco)
#define MEMORY_WINDOW_SIZE	(9)
#define MEMORY_WINDOW_FACE	(0)

/* items in memory dialog */
enum {
	iTitleText = 1,
	iMemoryBox1,
	iMemoryBox2,
	iMemoryDiff,
	iMemoryCompactButton,
	iMemoryPurgeButton,
	iMemoryLast
};

MemoryWindowType gMemoryWindow;

/*----------------------------------------------------------------------------*/
/* memory info */
/*----------------------------------------------------------------------------*/

/* get information about current memory usage */
static void MemoryInfo(MemoryInfoType *info)
{
	memclr(info, sizeof(MemoryInfoType));
	info->stacksize = MemStackSize();
	info->stackfree = StackSpace();
	info->stackused = info->stacksize - info->stackfree;
	info->heapsize = MemHeapSize();
	info->heapfree = FreeMem();
	info->heapused = info->heapsize - info->heapfree;
	PurgeSpace(&info->purgeable, &info->maxcontig);
	info->purgeable -= info->heapfree;
}

/* calculate the difference between two memory info structures */
static void MemoryInfoDiff(const MemoryInfoType *info1,
	const MemoryInfoType *info2, MemoryInfoType *diff)
{
	diff->stacksize = info1->stacksize - info2->stacksize;
	diff->stackfree = info1->stackfree - info2->stackfree;
	diff->stackused = info1->stackused - info2->stackused;
	diff->heapsize = info1->heapsize - info2->heapsize;
	diff->heapfree = info1->heapfree - info2->heapfree;
	diff->heapused = info1->heapused - info2->heapused;
	diff->purgeable = info1->purgeable - info2->purgeable;
	diff->maxcontig = info1->maxcontig - info2->maxcontig;
}

/*----------------------------------------------------------------------------*/
/* memory view */
/*----------------------------------------------------------------------------*/

static void MemoryViewDrawHook(OffscreenHandle offscreen, void *data);

/* create a new memory view in the specified area of the port */
static void MemoryViewBegin(MemoryViewType *view, GrafPtr port, const Rect *bounds)
{
	memclr(view, sizeof(MemoryViewType));
	view->offscreen = OffscreenBegin(port, MemoryViewDrawHook, view);
	OffscreenBoundsSet(view->offscreen, bounds);
}

/* dispose of a memory view area */
static void MemoryViewEnd(MemoryViewType *view)
{
	OffscreenEnd(view->offscreen);
	memclr(view, sizeof(MemoryViewType));
}

/* draw the frame around the view--black if enabled, gray if disabled */
static void MemoryViewDrawFrame(MemoryViewType *view)
{
	Rect bounds;
	GrafPtr port = NULL;
	PenState pen;

	GetPort(&port);
	SetPort(OffscreenPort(view->offscreen));
	OffscreenBounds(view->offscreen, &bounds);
	GetPenState(&pen);
	if (! view->enabled)
		PenPat(gray);
	FrameRect(&bounds);
	SetPenState(&pen);
	SetPort(port);
}

/* draw the memory info to the view's port, without setting up the
	offscreen bitmap, so this can be used either to draw into
	the offscreen bitmap (if OffscreenBeginDrawing has been called)
	or directly to the screen */
static void MemoryViewDrawDirect(MemoryViewType *view)
{
	Rect bounds;
	CStr255 str;
	TextState state;
	GrafPtr port = NULL;
	
	MemoryViewDrawFrame(view);
	GetPort(&port);
	SetPort(OffscreenPort(view->offscreen));
	OffscreenBounds(view->offscreen, &bounds);
	sprintf(str, "%8ld\r%8ld\r%8ld\r%8ld\r%8ld\r",
		view->info.heapsize, view->info.heapused,
		view->info.heapfree, view->info.purgeable,
		view->info.maxcontig);
	GetTextState(&state);
	TextFont(MEMORY_VIEW_FONT);
	TextSize(MEMORY_VIEW_SIZE);
	TextFace(MEMORY_VIEW_FACE);
	InsetRect(&bounds, 1, 1);
	EraseRect(&bounds);
	TextBox(str, strlen(str), &bounds, teJustLeft);
	SetTextState(&state);
	SetPort(port);
}

static void MemoryViewDrawHook(OffscreenHandle offscreen, void *data)
{
	MemoryViewDrawDirect(data);
}

/* enable the view--a disabled view has a gray background, and calls
	to MemoryViewRefresh don't update the view's appearance */
static void MemoryViewEnable(MemoryViewType *view, Boolean enable)
{
	if (enable != view->enabled) {
		view->enabled = enable;
		OffscreenChange(view->offscreen);
		OffscreenDraw(view->offscreen);
	}
}

/* true if the view is enabled */
static Boolean MemoryViewEnabled(MemoryViewType *view)
{
	return(view->enabled);
}

/* update the display to reflect the memory info */
static void MemoryViewRefresh(MemoryViewType *view, const MemoryInfoType *info)
{	
	if (view->enabled) {
		if (memcmp(&view->info, info, sizeof(MemoryInfoType)) != 0) {
			view->info = *info;
			OffscreenChange(view->offscreen);
			OffscreenDraw(view->offscreen);
		}
	}
}

/* true if point is in view */
static Boolean MemoryViewWithin(MemoryViewType *view, Point where)
{
	Rect bounds;
	
	OffscreenBounds(view->offscreen, &bounds);
	return(PtInRect(where, &bounds));
}

/* find memory view containing point */
static MemoryViewType *MemoryViewFind(MemoryViewType *views, short nviews,
	Point where)
{
	MemoryViewType *found = NULL;
	short i;
	
	for (i = 0; ! found && i < nviews; i++) {
		if (MemoryViewWithin(&views[i], where))
			found = &views[i];
	}
	return(found);
}

/*----------------------------------------------------------------------------*/
/* memory window */
/*----------------------------------------------------------------------------*/

/* redraw the memory views */
void MemoryWindowUpdate(MemoryWindowType *memory)
{
	short i;
	
	for (i = 0; i < memory->nviews; i++)
		OffscreenDraw(memory->views[i].offscreen);
}

/* update the data displayed in the memory views to correspond to current
	memory usage */
void MemoryWindowRefresh(MemoryWindowType *memory)
{
	MemoryInfoType info;
	MemoryInfoType diff;
	
	MemoryInfo(&info);
	MemoryViewRefresh(&memory->views[0], &info);
	MemoryViewRefresh(&memory->views[1], &info);
	MemoryInfoDiff(&memory->views[0].info, &memory->views[1].info, &diff);
	MemoryViewRefresh(&memory->views[2], &diff);
}

/* task installed to refresh the memory views after every event */
static void MemoryWindowTask(TaskHandle task, void *data)
{
	MemoryWindowRefresh(data);
}

/* create a window to display information about memory usage */
void MemoryWindowBegin(MemoryWindowType *memory)
{
	GrafPtr port;
	const short nviews = 3;
	Rect boxes[3];
	short i;
	
	TRY {
		if (! memory->dlg) {
			memclr(memory, sizeof(MemoryWindowType));
			memory->dlg = DlgBegin(RLD_DEBUG_MEMORY);
			memory->task = EventPostTaskInsert(MemoryWindowTask, memory);
			DlgBox(memory->dlg, iMemoryBox1, &boxes[0]);
			DlgBox(memory->dlg, iMemoryBox2, &boxes[1]);
			DlgBox(memory->dlg, iMemoryDiff, &boxes[2]);
			for (i = 0; i < nviews; i++) {
				MemoryViewBegin(&memory->views[i], memory->dlg, &boxes[i]);
				MemoryViewEnable(&memory->views[i], true);
			}
			memory->nviews = nviews;
			MemoryWindowRefresh(memory);
			GetPort(&port);
			SetPort(memory->dlg);
			TextFont(MEMORY_WINDOW_FONT);
			TextSize(MEMORY_WINDOW_SIZE);
			TextFace(MEMORY_WINDOW_FACE);
			SetPort(port);
			WinRegister(memory->dlg, memory, MemoryWindowEventTable());
			WinZoomRestore(memory->dlg, PrefsFile(), PREFS_MEMORY_WINDOW_POSITION);
			WinShow(memory->dlg);
		}
		WinSelect(memory->dlg);
	} CATCH {
		MemoryWindowEnd(memory);
	} ENDTRY;
}

/* dispose of the memory display window */
void MemoryWindowEnd(MemoryWindowType *memory)
{
	short i;
	
	if (memory->dlg) {
		WinZoomSave(memory->dlg, PrefsFile(), PREFS_MEMORY_WINDOW_POSITION);
		for (i = 0; i < memory->nviews; i++)
			MemoryViewEnd(&memory->views[i]);
		EventPostTaskDelete(memory->task);
		WinUnregister(memory->dlg, memory);
		DlgEnd(memory->dlg);
		memclr(memory, sizeof(MemoryWindowType));
	}
}

/* true if point is in one of the memory views */
Boolean MemoryWindowWithin(MemoryWindowType *memory, Point where)
{
	return(MemoryViewFind(memory->views, memory->nviews, where) != NULL);
}

/* handle a click in one of the memory views */
Boolean MemoryWindowMouseDown(MemoryWindowType *memory, EventRecord *event)
{
	MemoryViewType *view;
	Point where;

	where = event->where;
	GlobalToPort(&where, memory->dlg);
	view = MemoryViewFind(memory->views, memory->nviews, where);
	if (view)
		MemoryViewEnable(view, ! MemoryViewEnabled(view));
	return(view != NULL);
}

/* handle a click in one of the memory display window buttons */
void MemoryWindowClicked(MemoryWindowType *memory, long clickedID)
{
	long grow;
	short i;
	
	switch (clickedID) {
	case iMemoryCompactButton:
	case iMemoryPurgeButton:
		SegmentsUnload();
		for (i = 0; i < memory->nviews; i++)
			MemoryViewEnable(&memory->views[i], true);
		if (clickedID == iMemoryCompactButton)
			MaxMem(&grow);
		else
			PurgeMem(LONG_MAX);
		MemoryWindowRefresh(memory);
		MemoryViewEnable(&memory->views[1], false);
		break;
	}
}

void MemoryWindowAdjustMenu(void)
{
	if (GetMHandle(RLM_DEBUG)) {
		MenuCmdEnable(CMD_DEBUG_MEMORY);
		if (gMemoryWindow.dlg && WinFirstVisible(WIN_LAYER_FLOAT) == gMemoryWindow.dlg)
			MenuCmdCheck(CMD_DEBUG_MEMORY, true);
	}
}

Boolean MemoryWindowMenu(const MenuPickType *pick)
{
	Boolean handled = false;
	
	if (pick->cmd == CMD_DEBUG_MEMORY) {
		MemoryWindowBegin(&gMemoryWindow);
		handled = true;
	}
	return(handled);
}
