/* Copyright (c) 1992 Ari I. Halberstadt */

/* Heap sort. This version could be improved by Using pointers instead
   of array indexes, thus reducing the number of integer multiplications.
   This is tricky due to pointer arithmetic and the need for compatability
   with ANSI C.
  
   Based on information in "Algorithms", by Robert Sedgewick,
   2nd ed., Addison-Wesley, 1988. */

#include <string.h>
#include "sortlib.h"
#include "sort_private.h"

static void hpsort1(void *base, size_t n, size_t size,
		    int (*compar)(const void *, const void *), void *v);
static void downheap1(void *base, size_t n, size_t size,
		      int (*compar)(const void *, const void *),
		      void *v, size_t i);

static void hpsort2(void *base, size_t nmemb, size_t size,
		    int (*compar)(const void *, const void *));
static void downheap2(void *base, size_t n, size_t size,
		      int (*compar)(const void *, const void *), size_t i);

/* Call hpsort just like you would the standard library routine qsort.
   First, hpsort tries to allocate a temporary variable; if this
   is succesful, then it uses the function hpsort1, which performs
   about 1/3 of the data moves of hpsort2. Otherwise, it uses hpsort2,
   which uses the swap function. */ 
void hpsort(void *base, size_t n, size_t size,
	    int (*compar)(const void *, const void *))
{
  void *v;

  if (size <= 1)
    return;
  
  v = malloc(size);
  if (v) {
    hpsort1(base, n, size, compar, v);
    free(v);
  }
  else
    hpsort2(base, n, size, compar);
}

/*----------------------------------------------------------------*/

/* Sort the array using heap sort. There are more comments for hpsort2. */
static void hpsort1(void *base, size_t n, size_t size,
		    int (*compar)(const void *, const void *), void *v)
{
  size_t i;
  
  i = n / 2;
  while (i > 0)
    downheap1(base, n, size, compar, v, --i);
  while (n > 1) {
    SWAP(get(0), get(--n), size);
    downheap1(base, n, size, compar, v, 0);
  }
}

/* Move the i'th item down the heap until heap order is restored. Use
   The temporary variable 'v' to stash a copy of the i'th item, so that
   we avoid making unnecessary data swaps. */
static void downheap1(void *base, size_t n, size_t size,
		      int (*compar)(const void *, const void *), void *v, size_t i)
{
  size_t j;
  void *pj;
  
  COPY(v, get(i), size);
  for (; i < n / 2; i = j) {
    j = i + i + 1;
    pj = get(j);
    if (j + 1 < n && compar(pj, succ(pj)) < 0)  {
      j += 1;
      pj = succ(pj);
    }
    if (compar(v, pj) >= 0)
      break;
    COPY(get(i), pj, size);
  }
  COPY(get(i), v, size);
}

/*----------------------------------------------------------------*/

/* Sort the array using heap sort. We repair the heap using swaps,
   which does more data moving than hpsort1, but will work even if
   we can't allocate any memory for a temporary variable. */
static void hpsort2(void *base, size_t n, size_t size,
		    int (*compar)(const void *, const void *))
{
  size_t i;

  /* At this point the array can be in any random order. */

  /* The first stage of heap sort is to construct the initial
     heap. This results in n/2 calls to downheap, with each
     call causing the heap to grow by one. */
  i = n / 2;
  while (i-- > 0)
    downheap2(base, n, size, compar, i);
    
  /* At this point the entire array is a valid heap. */
  
  /* The next stage is to swap the largest item, which is at
     the top of the heap, with the item which is at the end
     of the array. This ensures that the largest item is
     in its correct position. Then, repair the heap
     which has now shrunk by one (since the last item is
     already in place). This results in n-1 calls to downheap,
     with each call causing the heap to shrink by one. */
  while (n > 1) {
    SWAP(get(0), get(--n), size);
    downheap2(base, n, size, compar, 0);
  }
  
  /* At this point the array is sorted, after having made
     (n + n/2 - 1) calls to downheap. */
}

/* Move the i'th item down the heap until heap order is restored. */
static void downheap2(void *base, size_t n, size_t size,
		      int (*compar)(const void *, const void *), size_t i)
{
  size_t j;	  /* Index to child of i */
  void *pi, *pj;  /* Pointers to i and j. This is just hand-done common
  		     subexpression elimination, which wouldn't be
  		     necessary if all compilers used half-decent
  		     optimization algorithms. */
  
  for (; i < n / 2; i = j) {
  
    j = i + i + 1; /* calculate index to left child of i'th element */
    
    /* choose larger of left and right children of i */
    pj = get(j);
    if (j+1 < n && compar(pj, succ(pj)) < 0) {
      j += 1;
      pj = succ(pj);
    }
    
    /* check for violation of heap order */
    if (compar(pi=get(i), pj) >= 0)
      break; /* no violation: heap is ok */
    
    /* swap i'th element with child */
    SWAP(pi, pj, size);
  }
}
