/* Copyright (c) 1992 Ari I. Halberstadt */

/* Insertion sort. Based on information in "Algorithms", by Robert Sedgewick,
   2nd ed., Addison-Wesley, 1988. */

/* 94/01/01 aih - Fixed error if 'n' param to _insort was greater than 'nmemb'
   93/11/03 aih - Added move of smallest item to head of array to speed up
		  inner loop. Idea from qsort in GNU C library, v1.06,
		  written by Douglas C. Schmidt (schmidt@ics.uci.edu).
   93/10/31 aih - rechecked code and did some additional analysis
   90/??/?? aih - wrote this */
  
#include <assert.h>
#include <string.h>
#include "sortlib.h"
#include "sort_private.h"

/* Move the 'size' bytes after 'i' before 'j'.
   For instance,
     given:   size = 4, j = 0, i = 16
     address: 01234567890123456789
     before:  abcdefghijklmnopqrst
     after:   qrstabcdefghijklmnop
   This code has the advantage of not requiring any temporary
   storage except for a single byte. The alternative is
   to first copy the last i..i+size bytes to a temporary
   location, moving the bytes j..i to j+size..i+size, then
   replacing bytes j..j+size with the temporary copy. */
static void insert(char *i, register char *j, register size_t size)
{
  register char *p;
  register char c;
  register size_t n = size;

  /* this table can be used as a guide for deciding which variables to
     put in registers.
       variable	# of times accessed
       n	n
       i	n
       j	n * (i - j) / size + n
       p	n * (i - j) / size + 2n
       size	n * (i - j) / size
       c	2n */
  /* assert(j+size <= i); */
  while (n--) {
    c = *(p = i++);
    while (p > j) {
      *p = *(p-size);
      p -= size;
    }
    *p = c;
    ++j;
  }
}

/* This is just like insort, but the smallest item in the first 'n' elements
   is first moved the start of the array, which eliminates the test
   for j>base in the inner loop. The smallest item in the array must be
   within the first n elements, or the sort will fail (and probably crash).
   This function was written for use by qksort, which guarantees
   that each element of the array is within 'n' items of its final
   position. */
void _insort(void *base, size_t nmemb, register size_t size,
  register int (*compar)(const void *, const void *), size_t n)
{
  register char *i, *j, *last;
  
  /* find smallest item in first n items */
  if (n > nmemb)
    n = nmemb;
  j = base;
  last = (char *) base + size * n;
  for (i = (char *) base + size; i < last; i += size)
    if (compar(j, i) > 0)
    	j = i;
 
  /* move smallest item to start */
  if (j > (char *) base)
    insert(j, base, size);

  /* assert: item [0] <= [1..nmemb] */
  
  /* sort the array using insertion sort */
  last = (char *) base + size * nmemb;
  for (i = (char *) base + size; i < last; i += size) {
    /* assert: items 0..i-1 are in sorted order */
    for (j = i; compar(j - size, i) > 0; j -= size)
      ;
     if (j < i)
      insert(i, j, size);
    /* assert: items 0..i are in sorted order */
  }
}

/* call insort just like you would the standard library function qsort */
void insort(void *base, size_t nmemb, register size_t size,
  register int (*compar)(const void *, const void *))
{
  register char *i, *j, *last;
  
  last = (char *) base + size * nmemb;
  for (i = (char *) base + size; i < last; i += size) {
    /* assert: items 0..i-1 are in sorted order */
    for (j = i; (void *) j > base && compar(j - size, i) > 0; j -= size)
      ;
     if (j < i)
      insert(i, j, size);
    /* assert: items 0..i are in sorted order */
  }
}
