/* Copyright (c) 1992 Ari I. Halberstadt */

/* Merge sort. Based on information in "Algorithms", by Robert
   Sedgewick, 2nd ed., Addison-Wesley, 1988. */

#include <assert.h>
#include <string.h>
#include "sortlib.h"
#include "sort_private.h"

#define SMALL (1) /* Size of small subfiles to sort using insertion sort
		     before merging the entire file. This didn't actually
		     save much, even when I coded the insertion sort
		     in-line to avoid function calls. I'm leaving in the
		     code just in case I missed something trivial. SMALL
		     would normally be around 10 or 16. */

#define MIN(a, b)	((a) < (b) ? (a) : (b))
#define mget(a, i)	((void*)((char*)a + size * (i)))

static void _mgsort(void *base, size_t nmemb, size_t size,
		    int (*compar)(const void *, const void *), void *aux);

static void merge(char *a, size_t na, char *b, size_t nb, char *aux,
		  size_t size, int (*compar)(const void *, const void *));

/* You can call mgsort just as you would the standard qsort. This
   implementation does a bottom up merge sort on the array,
   using O(n*lg(n)) comparisons, O(n*lg(n)) copy operations,
   and an auxiliary array of size n, or size*nmemb bytes.
   If the auxiliary array could not be allocated, then we
   call heap sort instead. While heap sort will be slower,
   its performance, even in the worst case, is still
   O(n*lg(n)). */
void mgsort(void *base, size_t nmemb, size_t size,
	    int (*compar)(const void *, const void *))
{
  void *aux = malloc(nmemb * size);
  if (aux) {
    _mgsort(base, nmemb, size, compar, aux);
    free(aux);
  }
  else
    hpsort(base, nmemb, size, compar);
}

/* Sort the array pointed to by 'base' using the auxiliary array
   of the same size. This function could be improved by doing
   an insertion sort on small subfiles. */
static void _mgsort(void *base, size_t n, size_t size,
		    int (*compar)(const void *, const void *), void *aux)
{
  size_t sz;	/* Size of current subfile. We merge subfiles
		   of size 1, 2, 4 ... n. */
  size_t i;	/* Offset of first subfile to merge. */
  void *src;	/* Instead of only merging from the original array
		   into the auxiliary array, we alternate merging:
		   once we merge from the source into the auxiliary,
		   and then from the auxiliary into the source. This
		   eliminates an extra copy of the data from the inner
		   loop. */
  void *tmp;	/* We use this to swap the source and auxiliary pointers. */
  

  #if SMALL > 1
    /* First, do an insertion sort on small subfiles. */
    for (i = 0; i < n; i += SMALL) {
      sz = (i+SMALL < n ? SMALL : n-i);
      insort(mget(base, i), sz, size, compar);
    }
  #endif

  /* Merge succesively larger subfiles. */
  src = base;
  for (sz = SMALL; sz < n; sz *= 2) {
    
    /* Merge all adjacent subfiles of size 'sz'. */
    for (i = 0; i < n; i += 2*sz) {
      size_t offset = i+sz < n ? i+sz : 0;
      merge(mget(src, i), MIN(sz, n-i),
	    mget(src, offset), offset ? MIN(sz, n-i-sz) : 0,
	    mget(aux, i), size, compar);
    }
    
    /* Swap the source array with the auxiliary array. This eliminates
       the need to copy the merged array from the auxiliary array
       back into the source array. Instead, we alternate: once we
       merge the source into the auxiliary, then we merge the
       auxiliary into the source. */
    tmp = src; src = aux; aux = tmp;
    
  }
  
  /* If there were an odd number of subfiles then the sorted array
     is currently in the auxiliary array. All that remains is to
     copy back into the original array. */
  if (src != base)
    memcpy(base, src, size * n);

}

/* Merge the first 'na' items of 'a' with the first 'nb' items of 'b'
   and place the result in 'aux'. */
static void merge(char *a, size_t na, char *b, size_t nb, char *aux,
		  size_t size, int (*compar)(const void *, const void *))
{
  void *end_a, *end_b;
  size_t i;
  
  end_a = mget(a, na);
  end_b = mget(b, nb);
  while ((void*)a < end_a && (void*)b < end_b) {
    i = size;
    if (compar((void*)a, (void*)b) <= 0)
      while (i-- != 0)
	*aux++ = *a++;
    else
      while (i-- != 0)
	*aux++ = *b++;
  }
  while ((void*)a < end_a) *aux++ = *a++;
  while ((void*)b < end_b) *aux++ = *b++;
}
