/* Copyright (c) 1992 Ari I. Halberstadt */

/* This implementation of the quick sort algorithm incorporates
   several improvements over a simple recursive version. These
   improvements should reduce execution time by around 20%-30%.
   Call qksort in precisely the same way as the standard qsort.
   It is identical to qsort except that it should be faster
   than many implementations of qsort. Also, qksort is written
   in ANSI C, and so should be portable to almost any ANSI
   compiler.
  
   Based on information in "Algorithms", by Robert Sedgewick,
   2nd ed., Addison-Wesley, 1988, p. 116-126.
  
   List of improvements to naive quick sort algorithm:
  
   - Tail recursion eliminated.
   
   - A private stack is used instead of recursive calls.
  
   - The smaller subfile is sorted first (larger is placed on stack),
   reducing the size of the stack.
  
   - The last two improvements ensure that the stack will
   never be larger than O(lg(n)), so we need very little extra
   memory to execute the sort (naive quick sort could
   use stack space O(n^2) in the worst case).
  
   - Insertion sort on small subfiles (this alone speeds
   execution by around 20%).
  
   - Median of three selection of the partition item. This makes
   worst case of O(n^2) very unlikely.
   
   - Use of pointers, instead of array indexes. This eliminates
   redundant multiplications to get an item. On the Macintosh
   Plus this just about halved execution time.
   
   - Partitioning is done within the same function, instead of
   calling a separate partition function, thus reducing the number
   of function calls.
   
   - Careful coding of inner loops.
   
   DISCUSSION OF FURTHER OPTIMIZATIONS
   
   The greatest amount of time is spent calling the comparison
   function, which is executed on average about 1.38*n*lg(n) times.
   Since the comparison function accounts for 70%-90% of the
   execution time, there isn't much point in trying to reduce the
   amount of time spent in the quick sort function. For instance,
   by writing the comparison function as a macro, execution time was
   reduced by around 50%. Within the quick sort function, about
   75% of the time is spent on the partitioning stage, so this
   would be the part to optimize if additional gains were required.
   
   One should concentrate on avoiding detrioration to the worst
   case of O(n^2), and on reducing the complexity of the algorithm
   (so that less instructions are required). It is also possible
   to reduce the number of comparisons for very large files, so that
   it approaches n*lg(n), but this seems fairly complicated (see
   Knuth for a mention of this method).
   
   In summary, it appears that this implemetation is near the
   limits of efficiency possible with the quick sort algorithm
   and the constraints of compatability with the standard library's
   implementation. The user's effort would best be spent on
   optimizing the comparison function and on finding a more
   efficient sorting algorithm for a specific problem.

   93/10/31 aih - changed so SMALL ranges from 1 up, rather than 2 up */

#include <assert.h>
#include <stddef.h>
#include <limits.h>
#include "sortlib.h"
#include "sort_private.h"

#define SMALL	(10) /* Size of smallest subfile sorted by quick sort. We sort
			subfiles smaller than this by executing a single
			insertion sort on the entire array after the
			quick sort has finished. The value of SMALL should
			be between 5 and 25. If SMALL is 1 then the insertion
			sort has no effect, since the array will have been
			sorted entirely by quick sort. SMALL must not be
			smaller than 1. */

#define MAXSTK	(sizeof(size_t) * CHAR_BIT)
		     /* Size of quick sort's stack. The stack must be
			at least lg(n), where n is the maximum number of
			items to sort. This value should be equal to the
			number of bits in the data type size_t. */

/* macros for handling the pushdown stack */
#define push(stk, n)	    ((*(stk)++ = (n)))
#define pop(stk)	    (*--(stk))
#define push2(stk, a, b)    ((void) (push((stk), (a)), push((stk), (b))))
#define pop2(stk, a, b)	    ((void) (((a) = pop(stk)), ((b) = pop(stk))))
#define stkempty(base, stk) ((base) >= (stk))

/* execute the quick sort algorithm */
void qksort(void *base, size_t nmemb, register size_t size,
	    register int (*compar)(const void *, const void *))
{
  register char *middle;/* The middle item of the array. During partitioning
  			   we compare left and right to this item. */
  register char *left;	/* This starts at the left end of the subfile, and
  			   during partioning advances to the right. When
  			   left crosses right we've finished partitioning. */
  register char *right; /* This starts at the right end of the subfile,
  			   and during partitioning advances to the left.
  			   When we've finished partitioning, it will be
  			   equal to the item to split the array on. This
  			   item is sometimes called the "pivot" item. */
  char *lo;	        /* First item in the current subfile. */
  char *hi;		/* Last item in the current subfile. */
  char **top;		/* Top of stack of subfiles. */
  char *stack[MAXSTK];  /* Storage for stack of subfiles. */
  ptrdiff_t small;	/* Size, in bytes, of small subfiles to skip. */
  
  assert(SMALL > 0);
  
  lo = get(0);
  hi = get(nmemb);
  top = stack;
  push2(top, lo, hi);
  small = SMALL * size;
  
  do {
    
    /* Skip subfiles smaller than some limit. When we're "done", we'll
       run a single insertion sort on the entire array. */
    if (hi - lo <= small) {
      pop2(top, hi, lo);
      continue;
    }
    
    /* Exchange the median of the left, right, and middle items with
       the middle item. The purpose of this is to avoid worst case
       performance, which could occur if the array was already sorted.
       This code is somewhat ugly due to the use of macros. */
    middle = lo + size * ((hi - lo) / (size<<1)); /* not same as lo+(hi-lo)/2 */
    #define last pred(hi)
    if (compar(lo, middle) <= 0) {
      if (compar(middle, last) <= 0)
	;
      else if (compar(lo, last) < 0) {
	SWAP(last, middle, size);
      }
      else {
	SWAP(lo, middle, size);
      }
    }
    else if (compar(lo, last) < 0) {
      SWAP(lo, middle, size);
    }
    else if (compar(last, middle) > 0) {
      SWAP(last, middle, size);
    }
    	
    /* Partition the array. This is the most time consuming
       part of the algorithm, since, as you can see, we're
       dealing with two while loops embedded in a for loop
       which itself is embedded in a do...while loop. Anything
       which can reduce the number of instructions in the
       inner loops is desirable. */
    left = pred(lo); right = hi;
    for (;;) {
      if (left < middle)
	while (compar(left = succ(left), middle) < 0)
	  ;
      if (right > middle)
	while (compar(right = pred(right), middle) > 0)
	  ;
      if (left >= right)
	break;
      SWAP(left, right, size);
      if (left == middle)
	middle = right;
      else if (right == middle)
	middle = left;
    }
    
    /* assert([lo...middle-1] <= middle <= [middle+1...hi]); */
    
    /* Put larger subfile on stack and sort smaller subfile first.
       Using a stack instead of recursive calls leads to
       some improvement on most machines. By sorting
       the smaller subfile first, we ensure that the stack
       will never be larger than lg(n). We also only place
       subfiles larger than SMALL items onto the stack. */
    if (middle - lo > hi - middle - size) {
      if (middle - lo > small)
	push2(top, lo, middle);	        /* qksort(lo, middle) */
      lo = succ(middle);		/* qksort(middle+1, hi) */
    }
    else {
      if (hi - middle - size > small)
	push2(top, succ(middle), hi);	/* qksort(middle+1, hi) */
      hi = middle;		        /* qksort(lo, middle) */
    }
    
  } while (! stkempty(stack, top));
  
  /* All items in the array are within SMALL items of their final position,
     so insertion sort is the best way to complete the sort. */
  _insort(base, nmemb, size, compar, SMALL);
}
