/*	Functions which are installed as call-backs for DocumentLib.c, and
	which support an editable text document in a window.

	Revisions:
	
	94/01/20 aih
	- updated a comment
	
	94/01/16 aih
	- most of the functionality is now in DocumentLib.c, so this file is
	much simpler
	
	94/01/06 aih
	- document creation function accepts preferences parameter
	
	93/12/24 aih
	- application preferences are used to set document defaults
	
	93/11/30 aih
	- moved undo code to TextScrollLib.c
	
	93/11/25 aih
	- removed wrap to view
	
	93/11/16 aih
	- fixed error trying to initialize printer record when there's no
	printer device (create new document failed with fnfErr, now we don't
	try to initialize printer record if PrOpen returns fnfErr).
	
	93/11/04 aih
	- revised to support TE32K and similar structures
	
	93/03/23 AIH
	- Moved printing functions into this file
	
	92/02/22 AIH
	- Simplified event handling, fixed some bugs, etc.
	
	91/05/20 AIH
	- Whether the text wraps to the window or to a fixed margin is now
	remembered with the document
	
	91/05/06-10 AIH
	- Added buffer to speed up keyboard input
	- Made separate functions for undo/cut/copy/paste/clear instead of
	using a switch statement
	- Made resource and constant declarations more consistent and portable
	- The file's resource fork is preserved, even if it was created
	by a different application
	- The 'EFNT' resource is used to set the font, and files created
	by THINK C, MPW, and Edit have word wrap turned off as the default
	
	91/05/05 AIH
	- Changed functions' prefixes from "U" followed by a letter,
	to "Tx" followed by several characters. For instance, "UT"
	became "TxUndo". Type names were also renamed to be more
	consistent.
	- Numerous other changes and bug fixes

	91/04/25 AIH
	- Added functions for: reading and writing files and for getting and setting
	various attributes
	
	91/04/03 AIH
	- created this library */
	
#include <string.h>
#include <limits.h>
#include <Script.h>
#include "ApplicationPreferencesLib.h"
#include "ArchiveLib.h"
#include "DocumentLib.h"
#include "MemoryLib.h"
#include "NotifyLib.h"
#include "RectangleLib.h"
#include "ResourceConstantsLib.h"
#include "ResourceLib.h"
#include "ResourceTypeLib.h"
#include "TextDocumentLib.h"
#include "TextIOLib.h"
#include "TextScrollLib.h"
#include "WindowLib.h"

#define TXDOC_TYPE		'TEXT'	/* type of files read/written */
#define TXDOC_VERSION	(3)		/* version of this library */
#define TXDOC_STYL		(128) 	/* style resource */
#define TXDOC_TDOC		(128)		/* document settings */

/* Data saved in the document's resource fork, in addition to the text which
	is stored in the data fork and the 'styl' record which is stored in the
	resource fork. */		
typedef struct {
	short		just;		/* text justification */
	short		margin;	/* wrap margin */
	Boolean	wrap;		/* true if word wrap is enabled */
} TxDocDataType;

/* return document's scroll text */
TextScrollHandle TxDocTextScroll(DocumentHandle doc)
{
	return(DocData(doc));
}

/* true if the document has been modified since it was last saved */
Boolean TxDocDirty(DocumentHandle doc)
{
	return(TxScrlDirty(TxDocTextScroll(doc)));
}

/* read the document from the file */
void TxDocRead(DocumentHandle doc, FileType *fp, Boolean *untitled)
{
	TxDocDataType	data;					/* archived data */
	size_t			dataSize = 0;		/* size of archived data */
	long				dataVersion = 0;	/* version of archived data */
	TextStyle		style;				/* for setting text style from 'EFNT' resource */
	Str255			font;					/* name of font from 'EFNT' resource */
	short				family = 0;			/* font family */
	Handle			efnt = NULL;		/* 'EFNT' resource */
	Boolean			dirty = true;		/* for setting text's dirty flag */
	TextScrollHandle scrl = TxDocTextScroll(doc);
	TextHandle		text = TxScrlText(scrl);
	
	require(FileTypeGet(fp) == TXDOC_TYPE);
	TRY {

		/* reset document */
		TxUndoReset((**scrl).undo);
		TxSelStartSet(text, 0);
		TxSelEndSet(text, TxLength(text));
		TxDelete(text);
		TxAutoRedrawSet(text, false);
		
		/* rewind to start and read text */
		FileSeek(fp, fsFromStart, 0);
		TxReadText(text, fp);
		
		/* if resource file exists then it has already been opened */	
		if (ResFileExists(fp)) {
			
			/* read style resource */
			if (ResExists1('styl', TXDOC_STYL))
				TxReadStyl(text, TXDOC_STYL);
			else {

				/* disable word wrap and set default font */
				TxWrapSet(text, false);
				style.tsSize = 9;
				style.tsFont = monaco;
				
				/* get 'EFNT' resource (documented in a TechNote) */
				if (ResExists1('EFNT', 1003)) {
					efnt = ResGet('EFNT', 1003);
					style.tsSize = *(short *) *efnt;
					*font = *(*efnt + sizeof(short));
					BlockMove(*efnt + sizeof(short), font, *font + 1);
					GetFNum(font, &family);
					style.tsFont = family;
				}						

				/* apply style */
				TxStyleSet(text, doFont | doSize, &style);
			}
			
			/* read document attributes */
			if (ResExists1(RES_TDOC_TYPE, TXDOC_TDOC)) {
				dataSize = sizeof(TxDocDataType);
				dataVersion = TXDOC_VERSION;
				ArchiveReadRes(RES_TDOC_TYPE, TXDOC_TDOC, &data,
									&dataSize, &dataVersion);
												
				/* validate the data before applying them */
				
				/* reset if wrong size or version */
				if (dataSize != sizeof(data) || dataVersion != TXDOC_VERSION)
					memclr(&data, sizeof(data));
												
				/* set justification */
				switch (data.just) {
				case teJustLeft:
				case teForceLeft:
				case teJustCenter:
				case teJustRight:
					break;
				default:
					data.just = GetSysJust();
					break;
				}
				TxJustSet(text, data.just);
				
				/* enable or disable word wrap */
				if (data.wrap != true && data.wrap != false)
					data.wrap = true;
				TxWrapSet(text, data.wrap);
				
				/* set wrap margin */
				if (data.margin < TX_MIN_MARGIN || data.margin > TX_MAX_MARGIN)
					data.margin = 0;
				TxMarginSet(text, data.margin);
			}
		}

		check(FileSize(fp) >= TxLength(text));
		if (TxLength(text) < FileSize(fp)) {
			/* warn user that document was truncated */
			NotifyNote(RLS_ALERT_TRUNCATED, FileName(fp));
			*untitled = true;
		}
		
		dirty = false;
	} CLEANUP {
		TxSelStartSet(text, 0);
		TxSelEndSet(text, 0);
		TxAutoRedrawSet(text, true);
		TxCalculate(text);
		TxInval(text);
		TxScrlChanged(scrl);
		TxScrlDirtySet(scrl, dirty);
	} ENDTRY;
}

/* write the document to the file */
void TxDocWrite(DocumentHandle doc, FileType *fp)
{
	TextScrollHandle scrl = TxDocTextScroll(doc);
	TextHandle text = TxScrlText(scrl);
	TxDocDataType data;
	
	TxWriteText(text, fp);
	TxWriteStyl(text, TXDOC_STYL);
	memclr(&data, sizeof(TxDocDataType));
	data.just = TxJust(text);
	data.wrap = TxWrap(text);
	data.margin = TxMargin(text);
	ArchiveWriteRes(RES_TDOC_TYPE, TXDOC_TDOC, &data,
						 sizeof(TxDocDataType), TXDOC_VERSION);
	TxScrlDirtySet(scrl, false);
}

/* Set default position and size for window. The window is sized to
	the size of a printed page (limited to the minimum size of the window
	and the maximum size of a window on the current screen) and is then
	positioned by staggering it on the current screen. */
static void TxDocPosition(DocumentHandle doc)
{
	short width, height;		/* width and height of window */
	short hres, vres;			/* resolution of screen */
	TPrInfo prinfo;			/* for page dimensions */
	
	ScreenRes(&hres, &vres);
	prinfo = (**DocPrintRecord(doc)).prInfo;
	if (RectValid(&prinfo.rPage) && RectArea(&prinfo.rPage) > 0) {
		/* use page size, converting from printer to screen pixels */
		width = (float) RectWidth(&prinfo.rPage) / prinfo.iHRes * hres;
		height = (float) RectHeight(&prinfo.rPage) / prinfo.iVRes * vres;
	}
	else {
		/* no printer so default to 8.5 by 11 inches */
		width = 8.5 * hres;
		height = 11 * vres;
	}
	if (TxWrap(TxScrlText(TxDocTextScroll(doc)))) {
		/* use word wrap margin for width */
		width = TxMargin(TxScrlText(TxDocTextScroll(doc))) + SBAR_WIDTH - 1 + 2 * TX_MARGIN;
	}
	WinStagger(DocWindow(doc), false, width, height);
}

/* dispose of the document */
void TxDocEnd(DocumentHandle doc)
{
	if (doc) {
		TxScrlEnd(TxDocTextScroll(doc));
		DocDataSet(doc, NULL);
	}
}

/* initialize a new document */
void TxDocBegin(DocumentHandle doc)
{
	TextScrollHandle scrl;
	TextPrefsType prefs;
	Rect bounds;
	
	TRY {

		/* Automatically switch to extended TextEdit (i.e., TE32K)
			if the file is too large to fit in a regular TextEdit
			record or if its size is close to TextEdit's limits. */
		prefs = AppPrefs()->text;
		if (DocHasFile(doc) &&
			 FileExists(DocFile(doc)) &&
			 FileSize(DocFile(doc)) >= 30000)
		{
			prefs.kind = TX_EXTENDED_KIND;
		}

		/* allocate and setup text record */
		WinPortRect(DocWindow(doc), &bounds);
		InsetRect(&bounds, -1, -1);
		scrl = TxScrlBegin(DocWindow(doc), &bounds, true, true,
			true, prefs.buffer, prefs.kind, &prefs.state);
		DocDataSet(doc, scrl);
		DocTypeSet(doc, TXDOC_TYPE);
		TxJustSet(TxScrlText(scrl), prefs.just);
		TxWrapSet(TxScrlText(scrl), prefs.wrap);
		TxMarginSet(TxScrlText(scrl), prefs.margin);

		/* set default window position */
		TxDocPosition(doc);
		
		WinFocusSet(DocWindow(doc), scrl);
	} CATCH {
		TxDocEnd(doc);
	} ENDTRY;
}

/* install the document handlers */
void TxDocTableRegister(void)
{
	static DocumentTableType table;
	static OSType types[] = { TXDOC_TYPE };
	
	table.begin = TxDocBegin;
	table.end = TxDocEnd;
	table.read = TxDocRead;
	table.write = TxDocWrite;
	table.dirty = TxDocDirty;
	table.print.begin = TxDocPrintBegin;
	table.print.end = TxDocPrintEnd;
	table.print.page = TxDocPrintPage;
	table.print.npages = TxDocPrintNPages;
	DocTableRegister(&table, 1, types);
}
