/* Print handlers for a text document.
	94/01/17 aih created by surgery on the now defunct TextPrintLib.c */

#include "MathLib.h"
#include "MemoryLib.h"
#include "RectangleLib.h"
#include "TextDocumentLib.h"
#include "TextLib.h"

/* information about the page being printed */
typedef struct {
	TextHandle text;			/* text record being printed */
	TextIntType view;			/* height of view rectangle (i.e., page rectangle) */
	TextIntType heightAll;	/* height of entire text */
	TextIntType total;		/* height of all of text printed so far */
	TextIntType remainder;	/* height remaining from last line on previous page */
	TextIntType lnum;			/* line number */
} PageDataType;

/* information about the document being printed */
typedef struct {
	PageDataType page;		/* information about current page */
	struct {
		GrafPtr port;			/* text's original graf port */
		Rect view;				/* text's original view rectangle */
		LRectType dest;		/* text's original destination rectangle */
	} save;
} PrintDataType;

/* true if the page data are valid */
static Boolean PageValid(PageDataType *pg)
{
	if (! TxValid(pg->text)) return(false);
	if (pg->view <= 0) return(false);
	if (pg->heightAll != TxHeightAll(pg->text)) return(false);
	if (pg->total < 0 || pg->total > pg->heightAll) return(false);
	if (pg->lnum < 0 || pg->lnum > TxLines(pg->text)) return(false);
	if (pg->remainder < 0 || pg->remainder > pg->heightAll) return(false);
	return(true);
}

/* initialize page data */
static void PageInit(PageDataType *pg, TextHandle text)
{
	Rect viewRect;
	
	memclr(pg, sizeof(PageDataType));
	TxView(text, &viewRect);
	pg->text = text;
	pg->view = RectHeight(&viewRect);
	pg->heightAll = TxHeightAll(text);
	ensure(PageValid(pg));
}

/* calculate height of the next page to print */
static short PageHeight(PageDataType *pg)
{
	TextIntType pageHeight;	/* height of page */
	TextIntType lineHeight;	/* height of line */
	
	require(PageValid(pg));
	
	if (TxLines(pg->text) == 0) {
		/* an empty document (though heightAll may still be greater than 0) */
		pageHeight = pg->heightAll;
	}
	else {
		
		/* use up as much as possible of any partially printed line */
		pageHeight = 0;
		if (pg->remainder) {
			pageHeight = min(pg->remainder, pg->view);
			pg->remainder -= pageHeight;
		}
	
		/* calculate number of whole lines that can entirely fit on page */
		while (pageHeight < pg->view && pg->total + pageHeight < pg->heightAll) {
			check(pg->lnum < TxLines(pg->text));
			lineHeight = TxHeight(pg->text, pg->lnum, pg->lnum);
			pageHeight += lineHeight;
			pg->lnum++;
		}
		if (pageHeight > pg->view) {
			pageHeight -= lineHeight;
			pg->lnum--;
			if (pageHeight == 0) {
				pageHeight = pg->view;
				pg->remainder = lineHeight - pageHeight;
			} 
		}
	}
	pg->total += pageHeight;
	
	ensure(PageValid(pg));
	ensure(TxLines(pg->text) == 0 ? pageHeight == pg->heightAll :
		0 < pageHeight && pageHeight <= pg->view);
	return(pageHeight);
}

/* offset to the start of the specified page */
static void PageOffset(PageDataType *pg, short pgnum)
{
	LRectType destRect;	/* destination rectangle of text */
	short	page;				/* current page number */

	require(PageValid(pg));
	/* calculate offset to page number */
	for (page = 1; page < pgnum && pg->total < pg->heightAll; page++)
		(void) PageHeight(pg);	
	/* scroll destination rectangle to page */
	TxDestination(pg->text, &destRect);
	LRectOffset(&destRect, 0, -pg->total);
	TxDestinationSet(pg->text, &destRect);
	ensure(PageValid(pg));
}

/* print next page of document */
Boolean TxDocPrintPage(DocumentHandle doc, void *data)
{
	PrintDataType *prdata = data;
	Rect viewRect;			/* view rectangle of text */
	LRectType destRect;	/* destination rectangle of text */
	short	height = 0;		/* height of this page */

	/* calculate the height of this page */
	height = PageHeight(&prdata->page);

	/* clip view rectangle to size of page */
	TxView(prdata->page.text, &viewRect);
	viewRect.bottom = viewRect.top + height;
	TxViewSet(prdata->page.text, &viewRect);
	
	/* draw page into the printing port */
	TxUpdate(prdata->page.text);
	
	/* scroll destination rectangle down to next page */
	TxDestination(prdata->page.text, &destRect);
	LRectOffset(&destRect, 0, -height);
	TxDestinationSet(prdata->page.text, &destRect);

	return(prdata->page.total >= prdata->page.heightAll);
}

/* return number of pages in document */
short TxDocPrintNPages(DocumentHandle doc, void *data)
{
	PrintDataType *prdata = data;
	float npages;
	Rect view;
	
	TxView(prdata->page.text, &view);
	npages = (float) TxHeightAll(prdata->page.text) / RectHeight(&view);
	check(0 <= npages && npages <= SHRT_MAX);
	if (npages > (short) npages)
		npages = (short) npages + 1;
	return(npages);
}

/* begin printing document */
void *TxDocPrintBegin(DocumentHandle doc, THPrint print,
	TPPrPort port, const Rect *pageRect,
	short first, short last)
{
	volatile PrintDataType *prdata = NULL;
	LRectType lpageRect;
	TextHandle text;

	TRY {
		prdata = PtrBegin(sizeof(PrintDataType));
		text = TxScrlText(TxDocTextScroll(doc));

		/* save text's parameters */
		TxView(text, &prdata->save.view);
		TxDestination(text, &prdata->save.dest);
		prdata->save.port = TxPort(text);
		
		if (! gDocPrintTest) {
			/* setup text so it draws into printer's port */
			LRectFromRect(&lpageRect, pageRect);
			TxViewSet(text, pageRect);
			TxDestinationSet(text, &lpageRect);
			TxPortSet(text, (GrafPtr) port);
		}
		
		/* initialize to first page */
		PageInit(&prdata->page, text);
		PageOffset(&prdata->page, first);
		
	} CATCH {
		PtrEnd(prdata);
	} ENDTRY;	
	return(prdata);
}

/* end printing document */
void TxDocPrintEnd(DocumentHandle doc, void *data)
{
	PrintDataType *prdata;
	TextHandle text;
	
	if (data) {
		prdata = data;
		text = TxScrlText(TxDocTextScroll(doc));
		TxPortSet(text, prdata->save.port);
		TxViewSet(text, &prdata->save.view);
		TxDestinationSet(text, &prdata->save.dest);
		if (gDocPrintTest)
			TxInval(text); /* force an update event */
		PtrEnd(prdata);
	}
}
