/*	Library of functions to do things to TextEdit records which Apple didn't
	provide functions for. Included are functions for accessing and setting
	the elements of a TextEdit record. Using these functions and those
	provided by Apple, you should not have to directly access
	any of the TextEdit record's fields. Also, several functions
	which only operated on styled text records can now be called
	for regular text records as well. For instance, instead of
	using GetStylScrap and SetStylScrap you can use TEStyleScrap
	and TEStyleScrapSet. And, the text edit scrap is always converted to
	the desk scrap, whether you're using an old or a new style edit record.
	
	Revision History:

	94/01/02 aih
	- if pasted characters would exceed maximum length of text then the pasted
	text is truncated
	
	93/12/29 aih
	- added timeout to width calculation functions
	
	93/12/23 aih
	- added functions for calculating text width
	
	93/12/19 aih
	- made many functions define's
	- removed most pre- and post- conditions since they're handled in TextLib.c
	
	93/12/15 aih
	- added work around for TEContinuousStyle bug
	- added functions to get and set style scrap for old and new style records
		
	93/11/13 aih
	- tevalid doesn't try to repair the record, it's just too hard to do so
	
	93/10/25 aih
	- The TEDo... functions now require that the length of the text not
	be increased beyond TE_MAX_LENGTH
	
	93/10/20 aih
	- TEValid will attempt to repair a damaged TEHandle
	
	93/03/19 AIH
	- Added equivalents to TECut, TECopy, TEPaste, TEStylePaste,
	TEDelete, TEKey, and TESetText which check for memory and text length
	before executing
	
	93/03/06 AIH
	- Fixed typo which resulted in comparison of a function's address
	instead of its return value
	
	92/03/04 AIH
	- Added a few functions to access offsets to lines, select entire lines,
	and return the text of a specific line
	
	92/02/20 AIH
	- Added TEActive function
	- Added functions to determine if a selection's style was continuous
	
	91/11/14 AIH
	- Uses the abs() function instead of an ABS() macro
	
	91/05/20 AIH
	- The wrap margin is either the width of the view rectangle or SHRT_MAX,
	never 0
	
	91/05/09 AIH
	- Strengthened pre and post conditions

	91/05/06 AIH
	- Added function to return text justification
	- Added functions for getting and setting the wrap margin
	
	91/04/25 AIH
	- Removed 'actsize' parameter from text reading functions, since the caller
	can use GetEOF to determine the size of the file
	
	91/04/09 Ari Halberstadt (AIH)
	- Created this file by modifying some old functions from TextWindows and
	adding other functions. */

#include <limits.h>
#include "DrawLib.h"
#include "MathLib.h"
#include "MemoryLib.h"
#include "RectangleLib.h"
#include "ScrapLib.h"
#include "TextEditLib.h"

static Boolean TEOldStyle(TEHandle te)
{
	return((**te).txSize != -1);
}

void TEDest(LRectType *ldest, TEHandle te)
{
	Rect dest;
	
	dest = (**te).destRect;
	LRectFromRect(ldest, &dest);
}

void TEDestSet(const LRectType *ldest, TEHandle te)
{
	Rect dest;

	LRectToRect(ldest, &dest);
	(**te).destRect = dest;
}

/* get style information */
void TEStyleContinuous(short *mode, TextStyle *style, TEHandle te)
{
	short height, ascent;
	
	if (TEOldStyle(te))
		TEStyle(style, &height, &ascent, te);
	else
		TEContinuousStyle(mode, style, te);
}

/* get the style of the text at the specified offset */
void TEStyleOffset(short offset, TextStyle *style, short *height,
	short *ascent, TEHandle te)
{
	TEGetStyle(offset, style, height, ascent, te);
}

/* Get the style of the first character of the current selection. Like
	TEGetStyle, but works with null style records. In other words, this
	will return the character style of the next character to be typed. */
void TEStyle(TextStyle *style, short *height, short *ascent, TEHandle te)
{
	StScrpHandle nullScrap = NULL;
	
	/* get null scrap handle */
	if ((**te).selStart == (**te).selEnd && GetStylHandle(te)) {
		if ((**GetStylHandle(te)).nullStyle)
			nullScrap = (**(**GetStylHandle(te)).nullStyle).nullScrap;
	}
	if (nullScrap && (**nullScrap).scrpNStyles > 0) {
		/* use info from null style record */
		style->tsFont = (**nullScrap).scrpStyleTab[0].scrpFont;
		style->tsFace = (**nullScrap).scrpStyleTab[0].scrpFace;
		style->tsSize = (**nullScrap).scrpStyleTab[0].scrpSize;
		style->tsColor = (**nullScrap).scrpStyleTab[0].scrpColor;
		*height = (**nullScrap).scrpStyleTab[0].scrpHeight;
		*ascent = (**nullScrap).scrpStyleTab[0].scrpAscent;
	}
	else if ((**te).selStart == (**te).selEnd && (**te).selStart > 0)
		TEStyleOffset((**te).selStart - 1, style, height, ascent, te);
	else
		TEStyleOffset((**te).selStart, style, height, ascent, te);
}

/* set the style of the text; works with old and new style edit records */
void TEStyleSet(short mode, const TextStyle *style, Boolean redraw, TEHandle te)
{
	Rect view;
	FontInfo info;
	GrafPtr port = NULL;
	TextState svtext;
	
	if (TEOldStyle(te)) {
		GetPort(&port);
		SetPort((**te).inPort);
		if (mode & doFont)
			(**te).txFont = style->tsFont;
		if (mode & doFace) {
			if (mode & doToggle)
				(**te).txFace ^= style->tsFace;
			else
				(**te).txFace = style->tsFace;
		}
		if (mode & addSize) {
			(**te).txSize += style->tsSize;
			if ((**te).txSize < 1)
				(**te).txSize = 1;
		}
		else if (mode & doSize)
			(**te).txSize = style->tsSize;
		if ((mode & doSize) || (mode & addSize) || (mode & doFont)) {
			GetTextState(&svtext);
			TextSize((**te).txSize);
			TextFont((**te).txFont);
			TextFace((**te).txFace);
			GetFontInfo(&info);
			SetTextState(&svtext);
			(**te).lineHeight = info.ascent + info.descent + info.leading;
			(**te).fontAscent = info.ascent;
		}
		if (redraw) {
			TECalText(te);
			TEView(&view, te);
			EraseRect(&view);
			TEUpdate(&view, te);
		}
		SetPort(port);
	}
	else
		TESetStyle(mode, style, redraw, te);
}

/* return text's style scrap; works for old and new edit records and for
	empty selections */
StScrpHandle TEStyleScrap(TEHandle te)
{
	TextStyle style;
	short height, ascent;
	StScrpHandle scrp = NULL;
	
	if (TEOldStyle(te) || (**te).selStart == (**te).selEnd) {
		TEStyle(&style, &height, &ascent, te);
		scrp = HandleBeginClear(sizeof(short) + sizeof(ScrpSTElement));
		(**scrp).scrpNStyles = 1;
		(**scrp).scrpStyleTab[0].scrpStartChar = (**te).selStart;
		(**scrp).scrpStyleTab[0].scrpHeight = height;
		(**scrp).scrpStyleTab[0].scrpAscent = ascent;
		(**scrp).scrpStyleTab[0].scrpFont = style.tsFont;
		(**scrp).scrpStyleTab[0].scrpFace = style.tsFace;
		(**scrp).scrpStyleTab[0].scrpSize = style.tsSize;
	}
	else {
		scrp = GetStylScrap(te);
		FailNIL(scrp);
	}
	ensure(HandleValidSize(scrp, sizeof(short) + sizeof(ScrpSTElement)));
	return(scrp);
}

/* Set text's style scrap. Works for old and new edit records,
	except for old style records applies the first scrap table
	entry to the entire text. Also works on empty selections. */
void TEStyleScrapSet(StScrpHandle scrp, Boolean redraw, TEHandle te)
{
	TextStyle style;
	
	if (TEOldStyle(te) || (**te).selStart == (**te).selEnd) {
		if (scrp && (**scrp).scrpNStyles > 0) {
			style.tsFont = (**scrp).scrpStyleTab[0].scrpFont;
			style.tsFace = (**scrp).scrpStyleTab[0].scrpFace;
			style.tsSize = (**scrp).scrpStyleTab[0].scrpSize;
			style.tsColor = (**scrp).scrpStyleTab[0].scrpColor;
			TEStyleSet(doAll, &style, redraw, te);
		}
	}
	else
		SetStylScrap((**te).selStart, (**te).selEnd, scrp, redraw, te);
}

/* Calculate the width of the widest line in the line range. If the
	operation takes longer than 'timeout' ticks to compute, then the
	last line reached is returned in 'finalLine', and the result of
	the function is the width of the widest line between 'firstLine'
	and 'finalLine'. Otherwise (i.e., no timeout), the result of the
	function is the width of the widest line and 'finalLine' is equal
	to 'lastLine'.
	
	The time out is provided since calculating the widths of the lines--
	especially for long styled text documents--can take quite a while, and
	doing so every time the document changes could be prohibitively slow.
	Instead, if 'finalLine' is less than 'lastLine', then after processing some
	events, the application can continue to call this function, passing
	the value of 'finalLine' as the 'firstLine' parameter, until 'finalLine'
	is equal to 'lastLine', at which point the maximum of the values returned
	by the successive calls to this function will be the width of the widest
	line in the original line range. */
static short TEWidthUnstyled(short firstLine, short lastLine, TicksType timeout,
	short *finalLine, TEHandle te)
{
	GrafPtr		svPort;					/* saved graf port */
	TextState	svTextState;			/* saved text state of port */
	SignedByte	svTextHandleState;	/* saved state of handle to text */
	Ptr			textPtr;					/* pointer to text */
	short			textWidth;				/* width of text */
	short			lineIndex;				/* current line number */
	short			lineLongest;			/* length of longest line */
	short			lineStart;				/* first character of line */
	short			lineEnd;					/* last character of line */
	long			startTicks;				/* time when we started */
	
	require(TEOldStyle(te));
	
	/* quick out */
	if ((**te).nLines == 0) return(0);
		
	/* keep track of elapsed time */
	startTicks = TickCount();
	
	/* find longest line */
	lineStart = lineEnd = lineLongest = 0;
	for (lineIndex = firstLine; lineIndex < lastLine; lineIndex++) {

		/* remember length of longest line encountered so far */
		if (lineLongest < (**te).lineStarts[lineIndex + 1] - (**te).lineStarts[lineIndex]) {
			lineStart = (**te).lineStarts[lineIndex];
			lineEnd = (**te).lineStarts[lineIndex + 1];
			lineLongest = lineEnd - lineStart;
		}

		/* stop if took too long */
		if (TickCount() - startTicks > timeout)
			break;
	}
	
	/* calculate width of longest line */
	GetPort(&svPort);
	SetPort((**te).inPort);
	GetTextState(&svTextState);
	TextFont((**te).txFont);
	TextSize((**te).txSize);
	TextFace((**te).txFace);
	svTextHandleState = HandleLock((**te).hText);
	textPtr = *(**te).hText;
	textWidth = TextWidth(textPtr, lineStart, lineEnd - lineStart);
	HandleRestore((**te).hText, svTextHandleState);
	SetTextState(&svTextState);
	SetPort(svPort);

	*finalLine = lineIndex;
	return(textWidth);
}

/* see description of TEWidthUnstyled */
static short TEWidthStyled(short firstLine, short lastLine, TicksType timeout,
	short *finalLine, TEHandle te)
{
	GrafPtr		svPort;					/* saved graf port */
	TextState	svTextState;			/* saved text state of port */
	SignedByte	svTextHandleState;	/* saved state of handle to text */
	SignedByte	svStyleHandleState;	/* saved state of handle to style record */
	SignedByte	svStyleTableState;	/* saved state of handle to style table */
	TEStylePtr	stylePtr;				/* text's style record */
	STPtr			styleTable;				/* text's style table */
	StyleRun 	*runPtr;					/* style runs array */
	short			runStart;				/* start of next style run */
	Ptr			textPtr;					/* pointer to text */
	short			textLength;				/* length of text run */
	short			textWidth;				/* width of text */
	short			lineWidth;				/* width of current line */
	short			lineIndex;				/* current line number */
	short			lineStart;				/* first character of line */
	short			lineEnd;					/* last character of line */
	long			startTicks;				/* time when we started */
	
	require(! TEOldStyle(te));
	
	/* quick out */
	if ((**te).nLines == 0) return(0);

	/* keep track of elapsed time */
	startTicks = TickCount();
	
	/* setup port */	
	GetPort(&svPort);
	SetPort((**te).inPort);
	GetTextState(&svTextState);
	
	/* lock and dereference handles */
	svTextHandleState = HandleLock((**te).hText);
	textPtr = *(**te).hText;
	svStyleHandleState = HandleLock(GetStylHandle(te));
	stylePtr = *GetStylHandle(te);
	svStyleTableState = HandleLock(stylePtr->styleTab);
	styleTable = *stylePtr->styleTab;
	
	/* initialize loop */
	textWidth = 0;
	runPtr = stylePtr->runs;
	runStart = runPtr[1].startChar;
	TextFont(styleTable[runPtr->styleIndex].stFont);
	TextFace(styleTable[runPtr->styleIndex].stFace);
	TextSize(styleTable[runPtr->styleIndex].stSize);
	
	/* calculate widths of all lines */
	for (lineIndex = firstLine; lineIndex < lastLine; lineIndex++) {
	
		/* calculate width of line */
		lineWidth = 0;
		lineStart = (**te).lineStarts[lineIndex];
		lineEnd = (**te).lineStarts[lineIndex + 1];
		while (lineStart < lineEnd) {
			if (lineStart == runStart) {
				/* use next style run (assumes run array is sorted by startChar) */
				runPtr++;
				runStart = runPtr[1].startChar;
				check(runPtr->startChar == lineStart);
				TextFont(styleTable[runPtr->styleIndex].stFont);
				TextFace(styleTable[runPtr->styleIndex].stFace);
				TextSize(styleTable[runPtr->styleIndex].stSize);
			}
			/* calculate width of style run */
			check(lineStart < runStart);
			textLength = min(lineEnd - lineStart, runStart - lineStart);
			check(textLength >= 1);
			lineWidth += TextWidth(textPtr, lineStart, textLength);
			lineStart += textLength;
		}
		textWidth = max(textWidth, lineWidth);
		
		/* stop if took too long */
		if (TickCount() - startTicks > timeout)
			break;
	}

	/* restore saved state */
	HandleRestore(RecoverHandle((Ptr) styleTable), svStyleTableState);
	HandleRestore(RecoverHandle((Ptr) stylePtr), svStyleHandleState);
	HandleRestore(RecoverHandle((Ptr) textPtr), svTextHandleState);
	SetTextState(&svTextState);
	SetPort(svPort);
	
	*finalLine = lineIndex;
	return(textWidth);
}

/* see description of TEWidthUnstyled */
short TEWidthTimeout(short start, short end,
	TicksType timeout, short *final, TEHandle te)
{
	return(TEOldStyle(te) ? TEWidthUnstyled(start, end, timeout, final, te) :
									TEWidthStyled(start, end, timeout, final, te));
}

/* calculate the width of the widest line in the line range */
short TEWidth(short start, short end, TEHandle te)
{
	return(TEWidthTimeout(start, end, LONG_MAX, 0, te));
}

void TEDoCopy(TEHandle te)
{
	/* see IM-VI, p15-36 for use of TENumStyles */
	MemCheck((**te).selEnd - (**te).selStart +
				TENumStyles((**te).selStart, (**te).selEnd, te)
					* sizeof(ScrpSTElement) + 2);
	TECopy(te);
	if (TEOldStyle(te)) {
		ZeroScrap();
		TEToScrap();
	}
}

void TEDoPaste(TEHandle te)
{
	volatile StScrpHandle style = NULL;
	volatile Handle scrap = NULL;
	volatile SignedByte state = 0;
	long length = 0;
	
	TRY {
		/* This line would be needed if we were still using TEPaste or
			TEStylPaste: if (TEOldStyle(te)) TEFromScrap(); */
			
		/* insert scrap, truncated to maximum length */
		length = ScrapLength('TEXT');
		if (length > 0) {
			TEDelete(te);
			if (length > TE_MAX_LENGTH - (**te).teLength)
				length = TE_MAX_LENGTH - (**te).teLength;
			style = (StScrpHandle) ScrapGet('styl');
			scrap = ScrapHandle();
			state = HandleLock(scrap);
			TEDoInsert(*scrap + ScrapOffset('TEXT'), length, style, te);
		}
	} CLEANUP {
		if (scrap)
			HandleRestore(scrap, state);
		HandleEnd(style);
	} ENDTRY;
}

void TEDoSetText(const void *text, long length, TEHandle te)
{
	TESelStartSet(0, te);
	TESelEndSet((**te).teLength, te);
	TEDelete(te);
	MemCheck(length);
	TESetText(text, length, te);
}

void TEDoInsert(const void *text, long length, StScrpHandle style, TEHandle te)
{
	Boolean active;
	
	MemCheck(length + (style ? HandleSize(style) : 0));
	active = ((**te).active != 0);
	if (! active)
		TEActivate(te); /* fixes bug in system 4.2 (TN#131) */
	TEStylInsert(text, length, style, te);
	if (! active)
		TEDeactivate(te); /* restore active state */
}

/* this is similar to TESetText, except the handle is used rather
	than a copy and the old text is first disposed of */
void TEReplaceText(Handle text, TEHandle te)
{
	SignedByte state = HandleLockHi(text);
	TEDoSetText(*text, HandleSize(text), te);
	HandleEnd(text);
}

void TEDoActivate(Boolean active, TEHandle te)
{
	if (active)
		TEActivate(te);
	else
		TEDeactivate(te);
}
