/*	file input/output for text
	94/01/15 aih removed a couple of unused functions for reading/writing
					 both the text and the style resource with one function call
	93/11/05 aih split from another file */

#include "MathLib.h"
#include "MemoryLib.h"
#include "ResourceLib.h"
#include "ResourceTypeLib.h"
#include "TextIOLib.h"

/* Replace the text with the text of the specified file. Only the first
	TxMaxLength() characters are read. The caller should compare the
	file's actual size with the TxLength; if they differ then the file
	was too large to be read and the user should be notified. */
void TxReadText(TextHandle text, FileType *fp)
{
	SignedByte			state = 0;		/* state of handle */
	FilePosType			count = 0;		/* number of bytes to read */
	volatile Handle	htext = NULL;	/* handle to text */
	volatile Boolean	opened = false;/* true if opened file */

	require(TxValid(text));
	TRY {
		if (fp->ref == FILE_CLOSED) {
			FileOpen(fp, fsRdPerm);
			opened = true;
		}
		count = min(FileSize(fp), TxMaxLength(text));
		htext = HandleBegin(count);
		state = HandleLockHi(htext);
		(void) FileRead(fp, count, *htext);
		HandleRestore(htext, state);
		TxReplaceText(text, htext);
		htext = NULL;
	} CLEANUP {
		if (opened) FileClose(fp);
	} CATCH {
		HandleEnd(htext);
	} ENDTRY;
	ensure(TxLength(text) == count);
}

/* read the text from a resource */
void TxReadTextRes(TextHandle text, ResType type, short id)
{
	Handle rsrc = NULL;
	
	require(TxValid(text));
	TRY {
		rsrc = ResGet(type, id);
		DetachResource(rsrc);
		TxReplaceText(text, rsrc);
		rsrc = NULL;
	} CATCH {
		HandleEnd(rsrc);
	} ENDTRY;
	ensure(TxValid(text));
}

/* Read a resource of type RES_STYL_TYPE with the specified ID from the current
	resource file and apply it to the text. */
void TxReadStyl(TextHandle text, short id)
{
	Handle rsrc = NULL;
	SignedByte state = 0;
	TextIntType start = 0;
	TextIntType end = 0;
	Boolean available = false;
	
	require(TxValid(text));
	rsrc = ResGet(RES_STYL_TYPE, id);
	state = HandleNoPurge(rsrc);
	available = MemAvailable(HandleSize(rsrc));
	if (available) {
		start = TxSelStart(text);
		end = TxSelEnd(text);
		TxSelStartSet(text, 0);
		TxSelEndSet(text, TxLength(text));
		TxStyleScrapSet(text, (StScrpHandle) rsrc);
		TxSelStartSet(text, start);
		TxSelEndSet(text, end);
	}
	HandleRestore(rsrc, state);
	if (! available)
		FailOSErr(memFullErr);
	ensure(TxValid(text));
}

/* write the text to the file */
void TxWriteText(TextHandle text, FileType *fp)
{
	volatile Boolean		opened = false;/* true if opened file */
	volatile Boolean		locked = false;/* true if locked handle to text */
	volatile SignedByte	state = 0;		/* saved state of handle to text */
	
	require(TxValid(text));
	TRY {
		if (fp->ref == FILE_CLOSED) {
			FileOpen(fp, fsWrPerm);
			opened = true;
		}
		state = HandleLockHi(TxText(text)); locked = true;
		FileWrite(fp, TxLength(text), *TxText(text));
		HandleRestore(TxText(text), state);
		FileResize(fp, TxLength(text));
	} CLEANUP {
		if (locked) HandleRestore(TxText(text), state);
		if (opened) FileClose(fp);
	} ENDTRY;
	ensure(TxValid(text));
}

/* write the text to a resource */
void TxWriteTextRes(TextHandle text, OSType type, short id)
{
	volatile Handle rsrc = NULL;

	require(TxValid(text));
	TRY {
		rsrc = HandleCopy(TxText(text));
		ResSet(rsrc, type, id);
		ResAttributeSet(rsrc, resPurgeable, true);
		ResWrite(rsrc);
	} CATCH {
		HandleEnd(rsrc);
	} ENDTRY;
	ensure(TxValid(text));
}

/* write the text's style scrap to a resource of type RES_STYL_TYPE */
void TxWriteStyl(TextHandle text, short id)
{
	volatile Handle rsrc = NULL;
	TextIntType start = 0;
	TextIntType end = 0;

	require(TxValid(text));
	TRY {
		start = TxSelStart(text);
		end = TxSelEnd(text);
		TxSelStartSet(text, 0);
		TxSelEndSet(text, TxLength(text));
		rsrc = (Handle) TxStyleScrap(text);
		TxSelStartSet(text, start);
		TxSelEndSet(text, end);
		ResSet(rsrc, RES_STYL_TYPE, id);
		ResAttributeSet(rsrc, resPurgeable, true);
		ResWrite(rsrc);
	} CATCH {
		HandleEnd(rsrc);
	} ENDTRY;
	ensure(TxValid(text));
}
