/*	Handle the font, size, and style menus.

	93/12/13 aih
	- added smaller/larger size menu commands
	- changed from TextDocumentEventLib to TextMenuLib
	
	93/11/29 aih
	- split wrap dialog into a separate file
	
	93/11/25 aih
	- removed wrap to view
	
	93/10/18 aih
	- removed unneeded function
	
	93/03/26 AIH
	- Cleaned up some of the keyboard buffering code
	
	93/03/16 AIH
	- mouseDown event is passed to TxScrlMouseDown
	
	93/03/10 AIH
	- Slightly improved error checking/reporting of edit menu commands
	
	93/03/06 AIH
	- Menu commands are disabled if memory is low	
	
	92/02/20-22 AIH
	- Simplified event handling and fixed a few bugs
	
	91/05/29 AIH
	- The size dialog doesn't increase the speed with which the thumb control
	increases or decreases size (formerly, if the button was held down for
	a few seconds the rate would double)
	
	91/05/26 AIH
	- Fixed crash when a menu couldn't be found
	
	91/05/22 AIH
	- Improved the margin dialog slightly
	
	91/05/20 AIH
	- Whether the text wraps to the window or to a fixed margin is now
	remembered with the document, so that if the window's size just happens
	to coincide with the wrap margin it doesn't automatically cause the
	program to assume that the text wraps to the window
	- The wrap margin dialog switches to "Wrap to margin" if the user
	types or clicks in the margin field
	- After the wrap margin is set the text is recalculated

	91/05/15 AIH
	- Added key event buffering to improve performance
	- Edit errors are checked
	
	91/05/07-08 AIH
	- Added up/down button to size dialog
	- Added wrap margin dialog

	91/05/05 AIH
	- Changed functions' prefixes from "U" followed by a letter,
	to "Tx" followed by several characters. For instance, "UT"
	became "TxUndo". Type names were also renamed to be more
	consistent.
	- Numerous other changes and bug fixes
	- Tinkering with system 7.0 required events

	91/04/03, part of April AIH
	- created this file, got it to work, etc. */

#include "FontLib.h"
#include "MenuLib.h"
#include "TextMenuLib.h"
#include "TextSizeDialogLib.h"
#include "TextWrapDialogLib.h"

/* subtract from number of size menu items to get menu command item */
#define iLarger	(2)
#define iSmaller	(3)

/*---------------------------------------------------------------------------*/
/* font menu */
/*---------------------------------------------------------------------------*/

static Boolean FontMenu(TextScrollHandle scrl, const MenuPickType *pick)
{
	TextStyle	style;
	Str255		name;
	short			family = 0;

	GetItem(pick->menu, pick->item, name);
	GetFNum(name, &family);
	style.tsFont = family;
	TxScrlStyleSet(scrl, doFont, &style);
	return(true);
}

/*---------------------------------------------------------------------------*/
/* size menu */
/*---------------------------------------------------------------------------*/

/* Set the size of the text; if the user selected "Other...", then
	put up the "Other Size" dialog and then change sizes. */
static Boolean SizeMenu(TextScrollHandle scrl, const MenuPickType *pick)
{
	TextStyle	style;		/* for getting/setting text size */
	Str31			name;			/* name of selected menu item */
	short			mode = 0;	/* for style information */
	long			size = 0;	/* size specified by menu item */
	short			nitems = 0;	/* number of items in menu */
	
	nitems = CountMItems(pick->menu);
	if (pick->item == nitems) {
		/* display other size dialog */
		mode = doSize + doFont + doFace;
		TxStyleContinuous(TxScrlText(scrl), &mode, &style);
		size = TxSizeDialog(mode, &style);
		mode = doSize;
	}
	else if (pick->item == nitems - iLarger) {
		/* larger */
		size = 1;
		mode = addSize;
	}
	else if (pick->item == nitems - iSmaller) {
		/* smaller */
		size = -1;
		mode = addSize;
	}
	else {
		/* convert menu item to size */
		GetItem(pick->menu, pick->item, name);
		StringToNum(name, &size);
		mode = doSize;
	}
	/* set text size */
	if (mode == addSize || (TX_MIN_FONT <= size && size <= TX_MAX_FONT)) {
		style.tsSize = size;
		TxScrlStyleSet(scrl, mode, &style);
	}
	return(true);
}

/*---------------------------------------------------------------------------*/
/* style menu */
/*---------------------------------------------------------------------------*/

static Boolean StyleMenu(TextScrollHandle scrl, const MenuPickType *pick)
{
	TextStyle style;
	
	GetItemStyle(pick->menu, pick->item, &style.tsFace);
	if (style.tsFace)
		TxScrlStyleSet(scrl, doFace + doToggle, &style);
	else {
		(*(short *) &style.tsFace) = 0; /*program_note: should this fix be in TELib.c? *//* fixes bug in system 4.1 (TN#131) */
		TxScrlStyleSet(scrl, doFace, &style);
	}
	return(true);
}

/*---------------------------------------------------------------------------*/
/* text menu */
/*---------------------------------------------------------------------------*/

/* set the text's justification */
static Boolean JustMenu(TextScrollHandle scrl, const MenuPickType *pick)
{
	short just = 0;
	Boolean handled = true;
	
	switch(pick->cmd) {
	case CMD_JUSTIFY_LEFT:		just = teForceLeft;	break;
	case CMD_JUSTIFY_CENTER:	just = teJustCenter;	break;
	case CMD_JUSTIFY_RIGHT: 	just = teJustRight;	break;
	default: handled = false; break;
	}
	if (handled && TxJust(TxScrlText(scrl)) != just) {
		TxInval(TxScrlText(scrl));
		TxJustSet(TxScrlText(scrl), just);
		TxScrlChanged(scrl);
	}
	return(handled);
}

/* handle a selection from the text menu */
static Boolean TextMenu(TextScrollHandle scrl, const MenuPickType *pick)
{
	Boolean handled = false;
	
	switch (pick->cmd) {
	case CMD_MARGIN:
		if (TxWrapDialog(TxScrlText(scrl)))
			TxScrlChanged(scrl);
		handled = true;
		break;
	}
	return(handled);
}

/*---------------------------------------------------------------------------*/
/* menu adjustment */
/*---------------------------------------------------------------------------*/

/* put a check mark by the current font */
static void FontMenuAdjust(MenuHandle menu, short mode, const TextStyle *style,
	TextHandle text)
{
	Str255 name;
	
	MenuCmdEnable(CMD_FONT);
	if ((mode & doFont) != 0) {
		GetFontName(FontID(style->tsFont), name);
		CheckItem(menu, MenuFindItem(menu, p2cstr(name)), true);
	}
}

/* outline available font sizes and mark the current size */
static void SizeMenuAdjust(MenuHandle menu, short mode, const TextStyle *style,
	TextHandle text)
{
	short item = 0;
	short nitems = CountMItems(menu);
	
	MenuCmdEnable(CMD_SIZE);
	MenuOutlineSizes(menu, FontID(style->tsFont), (mode & doFont) != 0);
	if ((mode & doSize) != 0) {
		Str31 size;
		NumToString(TxSize(text), size);
		item = MenuFindItem(menu, p2cstr(size));
	}
	CheckItem(menu, item ? item : nitems, true);
}
	
/* put checkmarks by the current styles */
static void StyleMenuAdjust(MenuHandle menu, short mode, const TextStyle *style,
	TextHandle text)
{
	short item;			/* current item */
	short nitems;		/* number of items in menu */
	Style itemStyle;	/* style of current item */
	
	MenuCmdEnable(CMD_STYLE);
	if ((mode & doFace) != 0) {
		nitems = CountMItems(menu);
		for (item = 1; item <= nitems; item++) {
			GetItemStyle(menu, item, &itemStyle);
			CheckItem(menu, item, (itemStyle == 0 && style->tsFace == 0) ||
										 (itemStyle & style->tsFace) != 0);
		}
	}
}

/* put check mark by the justification of the current selection */
static void JustMenuAdjust(TextHandle text)
{
	short just = TxJust(text);
	
	MenuCmdEnable(CMD_JUST);
	MenuCmdEnable(CMD_JUSTIFY_LEFT);
	MenuCmdEnable(CMD_JUSTIFY_CENTER);
	MenuCmdEnable(CMD_JUSTIFY_RIGHT);
	MenuCmdCheck(CMD_JUSTIFY_LEFT, just == teForceLeft || just == teJustLeft);
	MenuCmdCheck(CMD_JUSTIFY_CENTER, just == teJustCenter);
	MenuCmdCheck(CMD_JUSTIFY_RIGHT, just == teJustRight);
}

/* enable the text menu */
static void TextMenuAdjust(TextHandle text)
{
	MenuCmdEnable(CMD_TEXT);
	MenuCmdEnable(CMD_MARGIN);
}

/* adjust menus */
void TxTextAdjustMenu(TextScrollHandle scrl)
{	
	TextHandle text;
	TextStyle style;
	short mode;
	
	require(TxScrlValid(scrl));
	text = TxScrlText(scrl);
	if ((TxFlags(text) & TX_READ_ONLY) == 0) {
		mode = doAll;
		TxStyleContinuous(text, &mode, &style);
		FontMenuAdjust(MenuCmdHandle(CMD_FONT), mode, &style, text);
		SizeMenuAdjust(MenuCmdHandle(CMD_SIZE), mode, &style, text);
		StyleMenuAdjust(MenuCmdHandle(CMD_STYLE), mode, &style, text);
		JustMenuAdjust(text);
		TextMenuAdjust(text);
	}
}

/* handle a menu command */
Boolean TxTextMenu(TextScrollHandle scrl, const MenuPickType *pick)
{
	Boolean handled = false;
	
	require(TxScrlValid(scrl));
	switch (MenuPick(pick->id, 0).cmd) {
	case CMD_FONT: handled = FontMenu(scrl, pick); break;
	case CMD_SIZE: handled = SizeMenu(scrl, pick); break;
	case CMD_STYLE:handled = StyleMenu(scrl, pick); break;
	case CMD_JUST: handled = JustMenu(scrl, pick); break;
	case CMD_TEXT: handled = TextMenu(scrl, pick); break;
	}
	return(handled);
}
