/*	Library for creating a scrollable text box. Styled text
	is used, which makes this a fairly useful library. See the text
	document library for an example of how best to use this library.

	94/01/12 aih
	- event tables' within handlers are overriden so we're not sensitive
	to the order in which the handlers are executed
	
	93/12/23 aih
	- added undoable, buffered, textKind, and textState parameters to TxScrlBegin
	
	93/12/19 aih
	- moved clickloop into TextLib.c
	
	93/11/30 aih
	- moved undo code from TextDocumentLib.c

	93/11/25 aih
	- removed support for wrap to view
	
	93/11/17 aih
	- modified resize function to minimize drawing drawn on resizing text

	93/11/03 aih
	- Moderate revisions to separate this library from direct calls
	to TextEdit so that TE32K library can be used
	
	93/03/26 AIH
	- Fixed clipRgn in click loop
	- Added support for page up, page down, home, and end keys
	
	93/03/18 AIH
	- Added special caret hook for read-only text, but removed it after
	I couldn't get it to work
	
	93/03/16 AIH
	- Removed focus flag
	- Added keep inactive flag
	
	93/03/06 AIH
	- Menu commands are disable if memory is low
	
	92/03/04 AIH
	- Added code to erase and invalidate only the area that has changed
	after the text is resize
	
	92/03/03 AIH
	- Added read-only flag
	- The text can't be made longer than its maximum length by typing too
	many characters
	- Strengthened a few assertions to ensure that the scroll text's bounding
	rectangle is not made too small
	- Made the click-loop function an external function so that the help
	window library could use it to scroll the list of help topics
	
	92/02/26 AIH
	- Forgot to dispose of 'textRgn' field of scroll text structure
	when disposing of the structure; it is now disposed of correctly.
	
	92/02/20 AIH
	- Simplifying event handling
	
	91/11/14 AIH
	- Removed checks for system software prior to version 7.0

	91/06/13 AIH
	- Time library is used to delay so that background tasks get time
	
	91/06/01 AIH
	- Exported the common window zoom state code to the window library
	
	91/05/29 AIH
	- Added a few functions to simplify event handling
	
	91/05/20 AIH
	- Added function to determine if the scroll text is being wrapped to the
	view rectangle (as opposed to no word wrap or word wrap to a specific
	margin)
	- The text isn't recalculated after the wrap margin is set to allow
	the application to change the wrap margin more than once without
	accruing a performance penalty; the application must therefore recalculate
	and invalidate the text when it has finished
	- Added two speed scrolling to the click loop
	
	91/05/15 AIH
	- Improved efficiency of entering characters
	
	91/05/09 AIH
	- Strengthened pre and post conditions
	
	91/05/05 AIH
	- Changed functions' prefixes from "U" followed by a letter,
	to "Tx" followed by several characters. For instance, "UT"
	became "TxUndo". Type names were also renamed to be more
	consistent.
	
	91/04/02, and part of April AIH
	- Cleaned up code; ANSI'fied; fixed bugs, made better, etc.
*/

#include <limits.h>
#include <string.h>
#include <Script.h>
#include "ControlLib.h"
#include "DrawLib.h"
#include "EventLib.h"
#include "KeyLib.h"
#include "MathLib.h"
#include "MacLib.h"
#include "MemoryLib.h"
#include "MenuLib.h"
#include "RectangleLib.h"
#include "ScreenLib.h"
#include "TextMenuLib.h"
#include "TextScrollLib.h"
#include "WindowLib.h"

/* true if the scroll text handle is valid */
Boolean TxScrlValid(TextScrollHandle scrl)
{
	if (! HandleValidSize(scrl, sizeof(TextScrollType))) return(false);
	if (! (**scrl).buffered && (**scrl).bufsz > 0) return(false);
	if ((**scrl).bufsz < 0 || (**scrl).bufsz > TXSCRL_BUFSZ) return(false);
	return(true);
}

/* return the window containing the text */
WindowPtr TxScrlWindow(TextScrollHandle scrl)
{
	require(TxScrlValid(scrl));
	return(TxPort((**scrl).text));
}

/* return the text record */
TextHandle TxScrlText(TextScrollHandle scrl)
{
	require(TxScrlValid(scrl));
	return((**scrl).text);
}

/* return text's frame */
FrameHandle TxScrlFrame(TextScrollHandle scrl)
{
	require(TxScrlValid(scrl));
	return((**(**scrl).text).frame);
}

/* true if the text has been modified */
Boolean TxScrlDirty(TextScrollHandle scrl)
{
	require(TxScrlValid(scrl));
	return((**scrl).dirty);
}

/* set the dirty flag */
void TxScrlDirtySet(TextScrollHandle scrl, Boolean dirty)
{
	require(TxScrlValid(scrl));
	(**scrl).dirty = dirty;
	ensure(TxScrlValid(scrl));
}

/* adjust the maximum values of the scroll bars */
static void AdjustScrollBarsMax(TextScrollHandle scrl)
{
	Rect view;
	
	require(TxScrlValid(scrl));
	TxView((**scrl).text, &view);
	SBarMaxSet((**scrl).hsbar, max(0, TxWidthAll((**scrl).text) - RectWidth(&view)));
	SBarMaxSet((**scrl).vsbar, max(0, TxHeightAll((**scrl).text) - RectHeight(&view)));
	ensure(TxScrlValid(scrl));
}

/* adjust the scroll bars to correspond to the current position in the text */
static void AdjustScrollBars(TextScrollHandle scrl)
{
	require(TxScrlValid(scrl));
	SBarValSet((**scrl).hsbar, TxOffsetHorizontal((**scrl).text));
	SBarValSet((**scrl).vsbar, TxOffsetVertical((**scrl).text));
	ensure(TxScrlValid(scrl));
}

/* call this after resizing the text */
void TxScrlResized(TextScrollHandle scrl)
{
	AdjustScrollBarsMax(scrl);
	AdjustScrollBars(scrl);
}

/* call this after the text changes */
void TxScrlChanged(TextScrollHandle scrl)
{
	require(TxScrlValid(scrl));
	(**scrl).dirty = true;
	TxAdjust((**scrl).text, TxSelStart((**scrl).text));
	AdjustScrollBarsMax(scrl);
	AdjustScrollBars(scrl);
	ensure(TxScrlValid(scrl));
}

/* show the selection */
void TxScrlShowSel(TextScrollHandle scrl)
{
	require(TxScrlValid(scrl));
	TxSelectionShow((**scrl).text);
	AdjustScrollBars(scrl);
	ensure(TxScrlValid(scrl));
}

/*	scroll the text by dh pixels horizontally and dv pixels vertically */
void TxScrlScroll(TextScrollHandle scrl, TextIntType dh, TextIntType dv)
{
	require(TxScrlValid(scrl));
	TxScroll((**scrl).text, dh, dv);
	AdjustScrollBars(scrl);
	ensure(TxScrlValid(scrl));
}

/* return amount to scroll horizontally when part of a scroll bar is clicked */
static short ScrollHorizontal(TextScrollHandle scrl, short part)
{
	Rect view;
	short	delta = 0;
	
	require(TxScrlValid(scrl));
	TxView((**scrl).text, &view);
	switch (part) {
	case inUpButton:		delta = TX_DELTAH; break;
	case inDownButton:	delta = -TX_DELTAH; break;
	case inPageUp:			delta = RectWidth(&view) / 2; break;
	case inPageDown:		delta = -RectWidth(&view) / 2; break;
	}
	ensure(TxScrlValid(scrl));
	return(delta);
}

/* return amount to scroll vertically when part of a scroll bar is clicked */
static short ScrollVertical(TextScrollHandle scrl, short part)
{
	Rect view;
	short	delta = 0;
	
	require(TxScrlValid(scrl));
	TxView((**scrl).text, &view);
	check(0 < TX_DELTAV && TX_DELTAV < RectHeight(&view));
	switch (part) {
	case inUpButton:		delta = TX_DELTAV; break;
	case inDownButton:	delta = -TX_DELTAV; break;
	case inPageUp:			delta = RectHeight(&view) - TX_DELTAV; break;
	case inPageDown:		delta = -RectHeight(&view) + TX_DELTAV; break;
	}
	ensure(TxScrlValid(scrl));
	return(delta);
}

/* A general click loop function, usually called indirectly via TxClick.
	If the mouse is dragged outside of the text then the text is scrolled
	in the direction of the mouse. The mouse position in local
	coordinates should be passed in the 'mouse' parameter. */
static Boolean TxScrlClickLoop(TextScrollHandle scrl, Point mouse)
{
	Boolean result;
	
	require(TxScrlValid(scrl));
	result = TxHookDefaultClickLoop((**scrl).text, mouse, NULL);
	AdjustScrollBars(scrl);
	ensure(TxScrlValid(scrl));
	return(result);
}

/* true if point is in text or scroll bars */
Boolean TxScrlWithin(TextScrollHandle scrl, Point where)
{
	return(TxWithin((**scrl).text, where) ||
			 SBarWithin((**scrl).vsbar, where) ||
			 SBarWithin((**scrl).hsbar, where));
}

/* handle a click in the text or scroll bars */
Boolean TxScrlMouseDown(TextScrollHandle scrl, EventRecord *event)
{
	Point where = event->where;
	Boolean result = false;
	
	require(TxScrlValid(scrl));
	GlobalToPort(&where, TxScrlWindow(scrl));
	if (SBarWithin((**scrl).hsbar, where))
		SBarClick((**scrl).hsbar, where);
	else if (SBarWithin((**scrl).vsbar, where))
		SBarClick((**scrl).vsbar, where);
	else if (TxWithin((**scrl).text, where)) {
		if ((**scrl).undo)
			TxUndoClick((**scrl).undo);
		TxClick((**scrl).text, where, (event->modifiers & shiftKey) != 0);
	}
	else
		result = false;
	ensure(TxScrlValid(scrl));
	return(result);
}

/* handle a virtual key code */
static Boolean TxScrlKeyVirtual(TextScrollHandle scrl, char virtual)
{
	switch (virtual) {
	case pageUpVKey:
		TxScrlScroll(scrl, 0, ScrollVertical(scrl, inPageUp));
		return(true);
	case pageDownVKey:
		TxScrlScroll(scrl, 0, ScrollVertical(scrl, inPageDown));
		return(true);
	case homeVKey:
		TxScrlScroll(scrl, SHRT_MAX, SHRT_MAX); /* first char of first line */
		return(true);
	case endVKey:
		TxScrlScroll(scrl, SHRT_MAX, -SHRT_MAX); /* first char of last line */
		return(true);
	}
	return(false);
}

/* enter the key */
void TxScrlKey(TextScrollHandle scrl, char key)
{
	require(TxScrlValid(scrl));
	TxKey((**scrl).text, key);
	if (! KeyIsArrow(key))
		TxScrlChanged(scrl);
	else
		TxScrlShowSel(scrl);
	ensure(TxScrlValid(scrl));
}

/* flush the typing buffer */
void TxScrlFlushBuffer(TextScrollHandle scrl)
{
	short i = 0;

	require(TxScrlValid(scrl));
	require(TXSCRL_BUFSZ > 0);
	TRY {
		if ((**scrl).bufsz > 0) {
			if ((**scrl).undo) {
				for (i = 0; i < (**scrl).bufsz; i++)
					TxUndoKey((**scrl).undo, (**scrl).buf[i]);
			}
			if ((**scrl).bufsz <= 2) {
				for (i = 0; i < (**scrl).bufsz; i++)
					TxScrlKey(scrl, (**scrl).buf[i]);
			}
			else {
				char buf[TXSCRL_BUFSZ];
				BlockMove((**scrl).buf, buf, (**scrl).bufsz);
				TxInsert((**scrl).text, buf, (**scrl).bufsz, NULL);
				TxScrlChanged(scrl);
			}
		}
	} CLEANUP {
		(**scrl).bufsz = 0;
	} CATCH {
		(**scrl).dirty = true;
		if ((**scrl).undo)
			TxUndoReset((**scrl).undo);
	} ENDTRY;
	ensure(TxScrlValid(scrl));
	ensure((**scrl).bufsz == 0);
}

/* add character to typing buffer */
void TxScrlBuffer(TextScrollHandle scrl, char key)
{
	require(TxScrlValid(scrl));
	require(key != backspaceKey);
	require(TXSCRL_BUFSZ > 0);
	if ((**scrl).buffered) {
		if ((**scrl).bufsz >= TXSCRL_BUFSZ)
			TxScrlFlushBuffer(scrl);
		(**scrl).buf[(**scrl).bufsz++] = key;
	}
	ensure(TxScrlValid(scrl));
	ensure((**scrl).bufsz <= TXSCRL_BUFSZ);
}

/* flush typing buffer if any event other than a keydown event */
void TxScrlFilterPre(TextScrollHandle scrl, EventRecord *event)
{
	require(TxScrlValid(scrl));
	if ((**scrl).bufsz > 0) {
		if ((event->what != keyDown && event->what != autoKey) || KeyCmd(event))
			TxScrlFlushBuffer(scrl);
	}
	ensure(TxScrlValid(scrl));
}

/* Handle key down events. To improve performance the keys
	are buffered before inserting them into the text. */
void TxScrlKeyDown(TextScrollHandle scrl, EventRecord *event)
{
	char key = event->message;		/* key to enter */
	char undokey = event->message;/* key to pass to TxUndoKey */
	Boolean dirty = true;			/* true if key modifies document */
	Boolean buffer = true;			/* true if key should be buffered */ 
	Boolean enter = true;			/* true if the key should be entered */

	require(TxScrlValid(scrl));

	/* hide the cursor while the user is typing */
	ObscureCursor();
	
	/* determine if character should be placed in buffer
		and if it will modify the document */
	buffer = (**scrl).buffered;
	if (KeyMovement(event)) {
		dirty = false;
		buffer = false;
		undokey = arrowLeftKey;
	}
	else if (key == backspaceKey) {
		if (TxSelStart((**scrl).text) == 0 && TxSelEnd((**scrl).text) == 0)
			dirty = enter = false;
		buffer = false;
	}
	else {
		/* a graphic character */
		if (TxSelStart((**scrl).text) < TxSelEnd((**scrl).text))
			buffer = false;
		else if (TxLength((**scrl).text) == TxMaxLength((**scrl).text)) {
			/* text would be too long */
			enter = buffer = dirty = false;
			SysBeep(5); 
		}
	}

	if (enter && ! KeyMovement(event) &&
		(TxFlags((**scrl).text) & TX_READ_ONLY) != 0)
	{
		/* can't modify read-only text */
		enter = buffer = dirty = false;
		SysBeep(5);
	}
	
	if (! buffer) {
		/* immediately insert character into document */
		TxScrlFlushBuffer(scrl);
		if (enter) {
			if ((**scrl).undo)
				TxUndoKey((**scrl).undo, undokey);
			if (! TxScrlKeyVirtual(scrl, KeyVirtual(event)))
				TxScrlKey(scrl, key);
			if (dirty)
				(**scrl).dirty = true;
		}
	}
	else if (enter) {
		/* place character in buffer */
		TxScrlBuffer(scrl, key);
	}
	
	ensure(TxScrlValid(scrl));
}

/* undo command */
void TxScrlUndo(TextScrollHandle scrl)
{
	require(TxScrlValid(scrl));
	if ((**scrl).undo) {
		TxUndo((**scrl).undo);
		TxScrlChanged(scrl);
	}
	ensure(TxScrlValid(scrl));
}

/* cut command */
void TxScrlCut(TextScrollHandle scrl)
{
	require(TxScrlValid(scrl));
	if ((**scrl).undo)
		TxUndoCut((**scrl).undo);
	else
		TxCut((**scrl).text);
	TxScrlChanged(scrl);
	ensure(TxScrlValid(scrl));
}

/* copy command */
void TxScrlCopy(TextScrollHandle scrl)
{
	require(TxScrlValid(scrl));
	if ((**scrl).undo)
		TxUndoCopy((**scrl).undo);
	else
		TxCopy((**scrl).text);
	ensure(TxScrlValid(scrl));
}

/* paste command */
void TxScrlPaste(TextScrollHandle scrl)
{
	require(TxScrlValid(scrl));
	if ((**scrl).undo)
		TxUndoPaste((**scrl).undo);
	else
		TxPaste((**scrl).text);
	TxScrlChanged(scrl);
	ensure(TxScrlValid(scrl));
}

/* clear command */
void TxScrlClear(TextScrollHandle scrl)
{
	require(TxScrlValid(scrl));
	if ((**scrl).undo)
		TxUndoClear((**scrl).undo);
	else
		TxDelete((**scrl).text);
	TxScrlChanged(scrl);
	ensure(TxScrlValid(scrl));
}

/* apply the style to the text */
void TxScrlStyleSet(TextScrollHandle scrl, short mode, TextStyle *style)
{
	require(TxScrlValid(scrl));
	if ((**scrl).undo)
		TxUndoStyleSet((**scrl).undo, mode, style);
	else
		TxStyleSet((**scrl).text, mode, style);
	TxScrlChanged(scrl);
	ensure(TxScrlValid(scrl));
}

/* handle a menu command */
Boolean TxScrlMenu(TextScrollHandle scrl, const MenuPickType *pick)
{
	Boolean handled = true;
	
	require(TxScrlValid(scrl));
	switch (pick->cmd) {
	case CMD_UNDO:			TxScrlUndo(scrl); break;
	case CMD_CUT:			TxScrlCut(scrl); break;
	case CMD_COPY:			TxScrlCopy(scrl); break;
	case CMD_PASTE:		TxScrlPaste(scrl); break;
	case CMD_CLEAR:		TxScrlClear(scrl); break;
	case CMD_SELECT_ALL: TxSelectAll((**scrl).text); break;
	default:
		handled = TxTextMenu(scrl, pick);
		break;
	}
	return(handled);
}

/* adjust menus */
void TxScrlAdjustMenu(TextScrollHandle scrl)
{
	require(TxScrlValid(scrl));
	TxTextAdjustMenu(scrl);
	TxAdjustMenu((**scrl).text);
	if ((**scrl).undo && (TxFlags((**scrl).text) & TX_READ_ONLY) == 0)
		TxUndoAdjustMenu((**scrl).undo);
}

/* set sleep period to zero if there's something in the keys buffer */
TicksType TxScrlAdjustSleep(TextScrollHandle scrl)
{
	require(! scrl || TxScrlValid(scrl));
	return(scrl && (**scrl).bufsz > 0 ? 0 : LONG_MAX);
}

/* click loop and scroll bar functions */

static Boolean ClickLoop(TextHandle text, Point where, void *data)
{
	return(TxScrlClickLoop(data, where));
}

static void HSBarAction(void *scrl, short part)
{
	TxScrlScroll(scrl, ScrollHorizontal(scrl, part), 0);
}

static void HSBarThumb(void *scrl, long delta)
{
	TxScrlScroll(scrl, delta, 0);
}

static void VSBarAction(void *scrl, short part)
{
	TxScrlScroll(scrl, 0, ScrollVertical(scrl, part));
}

static void VSBarThumb(void *scrl, long delta)
{
	TxScrlScroll(scrl, 0, delta);
}

/* create a new scroll text */
TextScrollHandle TxScrlBegin(WindowPtr window, const Rect *bounds,
	Boolean vertical, Boolean horizontal, Boolean undoable, Boolean buffered,
	TextKindType textKind, const TextState *textState)
{
	TextScrollHandle scrl = NULL;	/* scroll text */
	TextUndoHandle undo = NULL;	/* undo data */
	ScrollBarHandle sbar = NULL;	/* a scroll bar */
	TextHandle text = NULL;			/* text record */
	GrafPtr port = NULL;				/* saved port */
	TextState state;					/* saved text state */
	Rect objrect;						/* rectangle for objects */
	static EventTableType overrideSBar; /* override scroll bar event table */
	static EventTableType overrideText; /* override text event table */
	
	require(WinValid(window));
	
	TRY {
		GetPort(&port);
		SetPort(window);
		GetTextState(&state);

		/* allocate scroll text handle */
		scrl = HandleBeginClear(sizeof(TextScrollType));
		(**scrl).buffered = buffered;
		
		/* create text */
		objrect = *bounds;
		if (horizontal) objrect.bottom -= SBAR_WIDTH - 1;
		if (vertical) objrect.right -= SBAR_WIDTH - 1;
		if (! textState) {
			TextNormal();
			TextFont(GetAppFont());
			TextSize(GetDefFontSize());
		}
		else
			SetTextState(textState);
		text = TxBegin(window, &objrect, textKind);
		(**scrl).text = text;
		TxAutoViewSet(text, false);
		TxHookDataSet(text, scrl);
		TxHookClickLoopSet(text, ClickLoop);
		
		/* override text's within handler so it isn't called */
		overrideText = *(const EventTableType *) TxEventTable();
		overrideText.focusWindow.within = NULL;
		WinObjectTableSet(window, text, &overrideText);
		
		/* create undo data */
		if (undoable) {
			undo = TxUndoBegin(text);
			(**scrl).undo = undo;
		}

		/* create horizontal scroll bar */
		if (horizontal) {
			objrect = *bounds;
			if (vertical) objrect.right -= SBAR_WIDTH - 1;
			objrect.top = objrect.bottom - SBAR_WIDTH;
		}
		sbar = SBarBegin(window, horizontal ? &objrect : NULL, (**text).frame);
		(**scrl).hsbar = sbar;

		/* create vertical scroll bar */
		if (vertical) {
			objrect = *bounds;
			if (horizontal) objrect.bottom -= SBAR_WIDTH - 1;
			objrect.left = objrect.right - SBAR_WIDTH;
		}
		sbar = SBarBegin(window, vertical ? &objrect : NULL, (**text).frame);
		(**scrl).vsbar = sbar;

		/* setup scroll bars */
		SBarDataSet((**scrl).hsbar, scrl);
		SBarDataSet((**scrl).vsbar, scrl);
		SBarThumbSet((**scrl).hsbar, HSBarThumb);
		SBarThumbSet((**scrl).vsbar, VSBarThumb);
		SBarActionSet((**scrl).hsbar, HSBarAction);
		SBarActionSet((**scrl).vsbar, VSBarAction);
		TxScrlChanged(scrl);
		TxScrlDirtySet(scrl, false);
		
		/* override scroll bars' within handlers so they aren't called */
		overrideSBar = *(const EventTableType *) SBarEventTable();
		overrideSBar.focusWindow.within = NULL;
		WinObjectTableSet(window, (**scrl).hsbar, &overrideSBar);
		WinObjectTableSet(window, (**scrl).vsbar, &overrideSBar);

		/* register with window */
		WinRegister(window, scrl, TxScrlEventTable());
		WinTabUnregister(window, text);
		WinTabRegister(window, scrl);
	} CLEANUP {
		if (port) {
			SetTextState(&state);
			SetPort(port);
		}
	} CATCH {
		TxScrlEnd(scrl);
	} ENDTRY;
	ensure(TxScrlValid(scrl));
	return(scrl);
}

/* dispose of the scroll text */
void TxScrlEnd(TextScrollHandle scrl)
{
	if (scrl) {
		EventPostTaskDelete((**scrl).resizeTask);
		if ((**scrl).text) {
			WinTabUnregister(TxPort((**scrl).text), scrl);
			WinUnregister(TxPort((**scrl).text), scrl);
		}
		SBarEnd((**scrl).vsbar);
		SBarEnd((**scrl).hsbar);
		TxUndoEnd((**scrl).undo);
		TxEnd((**scrl).text);
		HandleEnd(scrl);
		scrl = NULL;
	}
	ensure(! TxScrlValid(scrl));
}
