/*	Functions for handling undo in TextEdit and similar structures (e.g., TE32K),
	Undoable operations include typing, cut, copy, paste, clear, and style
	changes. See the text document library for an example of how best to use
	this library.
	
	Revisions:
	
	94/01/16 aih
	- capitalized constant names for consistency with rest of libraries
	
	93/12/16 aih
	- fixed some bugs left over from disabling copy; as a side effect, removed
	some complexity from the code and eliminated the annoying flashing
	caused by selecting the text and now style changes to an empty selection
	are undoable
	
	93/12/13 aih
	- copy is no longer undoable since the document doesn't change during a
	copy, only the clipboard changes, and saving and restoring the clipboard
	can take a lot of memory and, to be done properly, should save all the 
	data in the clipboard, not just data we put there (e.g., PICT as well
	as TEXT and styl data types, in addition to any other data that may be in
	the clipboard).
	- to avoid flickering the text is deactivated before undo/redo and
	reactivated when finished
	
	93/12/01 aih
	- fixed potential error when svstyle field is null
	
	93/11/30 aih
	- added adjust menu function
	
	93/11/03 aih
	- revisions to allow support of TE32K and similar structures
	
	93/10/25 AIH
	- To avoid failing to complete a cut or copy due to a large clipboard,
	the contents of the clipboard aren't saved if there isn't enough memory
	
	93/10/20 AIH
	- TxUndoValid attempts to repair an invalid undo structure
	
	93/03/19 AIH
	- TESetStyle is called directly, instead of letting caller handle
	it for style changes
	
	91/11/15 AIH
	- Removed unused code and assembly language code
	
	91/11/14 AIH
	- Removed checks for system software version 7.0

	91/06/01 AIH
	- While undo will basically preserve most of the information,
	there are a few known bugs:
program_note:
	- Backspacing over styled text doesn't record the style of the text
	in the undo data. One solution would be to insert the style data
	directly into the style scrap, but this is tricky and prone to compatability
	problems. Another alternative, but one which would require extensive
	revision of this library, is to use an "auxillary" TextEdit record
	which contains the saved text. This is the only known bug which actually
	loses some style information in the process of undoing.
	- The selection range flickers annoyingly when undoing or redoing. The
	solution is probably to analyze precisely when the selection range needs
	to be set and thus minimize the number of times it is set.
	- The selection range is not always returned to its previous value
	after an undo or redo. The solution again involves a detailed analysis
	of the selection range.
	
	91/05/15 AIH
	- Improved undoing key downs (arrow keys don't get stuck in the undo buffer
	anymore)
	
	91/05/12 AIH
	- Improved error checking and reporting
	- Fixed a few bugs with undoing typing
	
	91/05/09 AIH
	- Strengthened pre and post conditions

	91/05/08 AIH
	- Attempting to handle backspacing over multi-byte characters; not sure
	if it will work since I don't know how to test it
	
	91/05/06 AIH
	- Made separate functions for undo/cut/copy/paste/clear instead
	of using a switch statement to choose the function to execute
	- Changed names of undo constants to be more consistent
	
	91/05/05 AIH
	- Changed functions' prefixes from "U" followed by a letter,
	to "Tx" followed by several characters. For instance, "UT"
	became "TxUndo". Type names were also renamed to be more
	consistent.
	
	91/04/25 AIH
	- Added function to reset the undo structure
	
	91/04/03 AIH
	- Brought up to date
	
	89/09/18 Ari Halberstadt (AIH)
	- I've further adapted everything from Kevin's code.	

	This is Kevin Calhoun's original comment for some of the routines used here:

		"Undo routines, adapted from the MacTutor article, the reference
		for which I'll include here as soon as I find it." */

#include <limits.h>
#include <Script.h>
#include "KeyLib.h"
#include "MacLib.h"
#include "MemoryLib.h"
#include "MenuLib.h"
#include "ResourceConstantsLib.h"
#include "TextUndoLib.h"

/* reduce flicker by only selecting if the selection range would change */
static void TextSelect(TextHandle text, TextIntType start, TextIntType end)
{
	if (start != TxSelStart(text) || end != TxSelEnd(text))
		TxSelect(text, start, end);
}

/*----------------------------------------------------------------------------*/
/* data validation */
/*----------------------------------------------------------------------------*/

/* true if the undo structure is valid; repairs are attempted otherwise */
Boolean TxUndoValid(TextUndoHandle undo)
{
	if (! HandleValidSize(undo, sizeof(TextUndoType))) return(false); /* can't recover from this error */
	if ((**undo).svstyle && ! HandleValid((**undo).svstyle)) goto invalid;
	if ((**undo).svtext && ! HandleValid((**undo).svtext)) goto invalid;
	if ((**undo).svstyle && ! *(**undo).svstyle) goto invalid;
	if ((**undo).svtext && ! *(**undo).svtext) goto invalid;
	if ((**undo).undoStart > (**undo).undoEnd) goto invalid;
	if ((**undo).status < TXUNDO_NULL) goto invalid;
	if ((**undo).status >= TXUNDO_LAST) goto invalid;
	return(true);
invalid:
	if (HandleValid((**undo).svstyle)) HandleEnd((**undo).svstyle);
	if (HandleValid((**undo).svtext)) HandleEnd((**undo).svtext);
	(**undo).svstyle = NULL;
	(**undo).svtext = NULL;
	(**undo).undoStart = (**undo).undoEnd = 0;
	(**undo).status = TXUNDO_NULL;
	return(false);
}

/*----------------------------------------------------------------------------*/
/* support functions */
/*----------------------------------------------------------------------------*/

/* save the end points of the current selection */
static void SaveEndPoints(TextUndoHandle undo)
{
	TextIntType start = TxSelStart((**undo).text);
	TextIntType end = TxSelEnd((**undo).text);
 	(**undo).undoStart = start;
	(**undo).undoEnd = end;
}

/* dispose of the text field */
static void DisposeUndoText(TextUndoHandle undo)
{
	HandleEnd((**undo).svtext);
	(**undo).svtext = NULL;
}

/* dispose of the style field */
static void DisposeUndoStyle(TextUndoHandle undo)
{
	HandleEnd((**undo).svstyle);
	(**undo).svstyle = NULL;
}

/* copy the text of the current selection to the undo text */
static void SaveText(TextUndoHandle undo)
{
	Handle text, hText;
	TextIntType start, end;
	
	DisposeUndoText(undo);
	start = TxSelStart((**undo).text);
	end = TxSelEnd((**undo).text);
	hText = TxText((**undo).text);
	text = HandleBegin(end - start);
	(**undo).svtext = text;
	BlockMove(*hText + start, *text, end - start);
}

/* copy the style for the current selection to the undo style */
static void SaveStyle(TextUndoHandle undo)
{
	StScrpHandle style = NULL;
	
	DisposeUndoStyle(undo);
	style = TxStyleScrap((**undo).text);
	(**undo).svstyle = style;
}

/* save the text and style information for the current selection */
static void SaveSelection(TextUndoHandle undo)
{
	SaveText(undo);
	SaveStyle(undo);
}

/* insert the character at the start of the undo text buffer */
static void inschar(TextUndoHandle undo, char c)
{
	Handle text; /* text buffer */
	size_t size; /* size of text buffer */

	text = (**undo).svtext;
	if (! text) {
		text = HandleBegin(0);
		(**undo).svtext = text;
	}
	size = HandleSize(text);
	HandleSizeSet(text, size + 1);
	BlockMove(*text, *text + 1, size);
	**text = c;
}	

static void RestoreText(TextUndoHandle undo)
{
	SignedByte state = HandleLockHi((**undo).svtext);
	TextSelect((**undo).text, (**undo).undoStart, (**undo).undoStart);
	TxInsert((**undo).text, *(**undo).svtext, HandleSize((**undo).svtext), NULL);
	HandleRestore((**undo).svtext, state);
}

static void RestoreStyle(TextUndoHandle undo)
{	
	TextSelect((**undo).text, (**undo).undoStart, (**undo).undoEnd);
	TxStyleScrapSet((**undo).text, (**undo).svstyle);
}

static void RestoreSelection(TextUndoHandle undo)
{
	SignedByte state = HandleLockHi((**undo).svtext);
	TextSelect((**undo).text, (**undo).undoStart, (**undo).undoStart);
	TxInsert((**undo).text, *(**undo).svtext,
		HandleSize((**undo).svtext), (**undo).svstyle);
	HandleRestore((**undo).svtext, state);
}

/* save the selection in a temporary place; restore the selection;
	then put the temporary place into the regular buffer */
static void SwapSelection(TextUndoHandle undo, Boolean text, Boolean style)
{
	void *tmp = NULL;
	Handle svtext = NULL;
	StScrpHandle svstyle = NULL;

	TRY {
		if (! text) DisposeUndoText(undo);
		if (! style) DisposeUndoStyle(undo);
		svtext = (**undo).svtext; (**undo).svtext = NULL;
		svstyle = (**undo).svstyle; (**undo).svstyle = NULL;
		TextSelect((**undo).text, (**undo).undoStart, (**undo).undoEnd);
		if (text) SaveText(undo);
		if (style) SaveStyle(undo);
		if (text) TxDelete((**undo).text);
		tmp = (**undo).svtext; (**undo).svtext = svtext; svtext = tmp;
		tmp = (**undo).svstyle; (**undo).svstyle = svstyle; svstyle = tmp;
		if (text && style)
			RestoreSelection(undo);
		else if (text)
			RestoreText(undo);
		else if (style)
			RestoreStyle(undo);
		DisposeUndoText(undo);
		DisposeUndoStyle(undo);
		(**undo).svtext = svtext; svtext = NULL;
		(**undo).svstyle = svstyle; svstyle = NULL;
		(**undo).undoEnd = TxSelEnd((**undo).text);
	} CATCH {
		HandleEnd(svtext);
		HandleEnd(svstyle);
	} ENDTRY;
}

/*----------------------------------------------------------------------------*/
/* undoing and redoing */
/*----------------------------------------------------------------------------*/

static void SetStyle(TextUndoHandle undo, short mode, TextStyle *style)
{
	if (! (**undo).styleSaved) {
	 	DisposeUndoText(undo);
		SaveEndPoints(undo);
		SaveStyle(undo);
		(**undo).styleSaved = true;
	}
	TxStyleSet((**undo).text, mode, style);
}

static void StyleUndo(TextUndoHandle undo)
{
	SwapSelection(undo, false, true);
}

static void StyleRedo(TextUndoHandle undo)
{
	SwapSelection(undo, false, true);
}

static void Typing(TextUndoHandle undo, unsigned char c)
{
	TextIntType start = TxSelStart((**undo).text);
	TextIntType end = TxSelEnd((**undo).text);

	switch (c) {
	case arrowUpKey:
	case arrowDownKey:
	case arrowLeftKey:
	case arrowRightKey:
		(**undo).typingRun = false;
		return;
	}
	
	/* if we're backspacing over the first character then there's nothing to do */
	if (c == backspaceKey && start == 0 && start == end)
		return;

	/* remember operation */
	if ((**undo).status != TXUNDO_TYPING) {
		DisposeUndoText(undo);
		DisposeUndoStyle(undo);
		(**undo).status = TXUNDO_TYPING;
	}
	
	/* save the current selection, before it's erased by the newly typed
		characters */
	if (start < end || ! (**undo).typingRun) {
	 	(**undo).undoStart = (**undo).undoEnd = start;
	 	SaveSelection(undo);
	}
	
	/* set flag to indicate that we're in a typing run (a sequence of
		entered characters unbroken by mouse clicks or other commands) */
	(**undo).typingRun = true;
		
	/* If we're backspacing, then change the typing start to be one character
		back, and save the character being backspaced over before TxKey deletes
		it. If the character is the second byte of a multi-byte character
		we store both bytes being backspaced over (multi-byte characters
		haven't been tested). CharByte won't move or purge memory, so we
		don't have to lock the text handle. */
	if (c == backspaceKey) {
		if (start == (**undo).undoStart && start == end) {
	 		
	 		/* save character being backspaced over */
	 		inschar(undo, TxCharacter((**undo).text, start - 1));
	 		
	 		/* save first byte of double-byte character */
	 		if (start > 1) {
		 		if (CharByte(*TxText((**undo).text), start - 2) == -1) {
			 		inschar(undo, TxCharacter((**undo).text, start - 2));
					(**undo).undoStart--;
	 			}
	 		}
			(**undo).undoStart--;
		}
		if (start == end)
			(**undo).undoEnd--;
	}
	else
		(**undo).undoEnd++;
}

static void TypingUndo(TextUndoHandle undo)
{
	SwapSelection(undo, true, true);
}

static void TypingRedo(TextUndoHandle undo)
{
	SwapSelection(undo, true, true);
}

static void Paste(TextUndoHandle undo)
{
	SaveEndPoints(undo);
	SaveSelection(undo);
	TxPaste((**undo).text);
	(**undo).undoEnd = TxSelEnd((**undo).text);
}

static void PasteRedo(TextUndoHandle undo)
{
	SwapSelection(undo, true, true);
}

static void PasteUndo(TextUndoHandle undo)
{
	SwapSelection(undo, true, true);
}

static void Cut(TextUndoHandle undo)
{
	SaveEndPoints(undo);
	SaveSelection(undo);
	TxCut((**undo).text);
}

static void CutRedo(TextUndoHandle undo)
{
	TextSelect((**undo).text, (**undo).undoStart, (**undo).undoEnd);
	TxDelete((**undo).text);
}

static void CutUndo(TextUndoHandle undo)
{
	RestoreSelection(undo);
}

static void Clear(TextUndoHandle undo)
{
	SaveEndPoints(undo);
	SaveSelection(undo);
	TxDelete((**undo).text);
}

static void ClearRedo(TextUndoHandle undo)
{
	CutRedo(undo);
}

static void ClearUndo(TextUndoHandle undo)
{
	CutUndo(undo);
}

/*----------------------------------------------------------------------------*/
/* undo status */
/*----------------------------------------------------------------------------*/

static void UndoStatusSet(TextUndoHandle undo, TxUndoStatusType status)
{
	(**undo).status = status; /* set also in Typing() */
}

static void UndoStatusToggle(TextUndoHandle undo)
{
	static TxUndoStatusType toggle[TXUNDO_LAST] = {
		TXUNDO_NULL,
		TXREDO_CUT, TXREDO_COPY, TXREDO_PASTE, TXREDO_CLEAR, TXREDO_STYLE, TXREDO_TYPING,
		TXUNDO_TYPING, TXUNDO_STYLE, TXUNDO_CUT, TXUNDO_COPY, TXUNDO_PASTE, TXUNDO_CLEAR,
	};
	UndoStatusSet(undo, toggle[(**undo).status]);
}

/*----------------------------------------------------------------------------*/
/* external functions */
/*----------------------------------------------------------------------------*/

/* return undo status */
TxUndoStatusType TxUndoStatus(TextUndoHandle undo)
{
	require(TxUndoValid(undo));
	return((**undo).status);
}

/* reset the undo structure */
void TxUndoReset(TextUndoHandle undo)
{
	require(TxUndoValid(undo));
	DisposeUndoText(undo);
	DisposeUndoStyle(undo);
	UndoStatusSet(undo, TXUNDO_NULL);	
	ensure(TxUndoValid(undo));
}

/* undo (or redo) the last command */
void TxUndo(TextUndoHandle undo)
{
	Boolean active = true;
	
	require(TxUndoValid(undo));
	TRY {
		(**undo).styleSaved = false;
		(**undo).typingRun = false;
		active = TxActive((**undo).text);
		TxActivate((**undo).text, false);
		switch ((**undo).status) {
		case TXUNDO_CUT:		CutUndo(undo); break;
		case TXUNDO_PASTE:	PasteUndo(undo); break;
		case TXUNDO_CLEAR:	ClearUndo(undo); break;
		case TXUNDO_STYLE:	StyleUndo(undo); break;
		case TXUNDO_TYPING:	TypingUndo(undo); break;
		case TXREDO_TYPING:	TypingRedo(undo); break;
		case TXREDO_STYLE:	StyleRedo(undo); break;
		case TXREDO_CUT:		CutRedo(undo); break;
		case TXREDO_PASTE:	PasteRedo(undo); break;
		case TXREDO_CLEAR:	ClearRedo(undo); break;
		}
		UndoStatusToggle(undo);
		TxActivate((**undo).text, active);
	} CATCH {
		TxUndoReset(undo);
		TxActivate((**undo).text, active);
	} ENDTRY;
	ensure(TxUndoValid(undo));
}

/* paste command */
void TxUndoCut(TextUndoHandle undo)
{
	require(TxUndoValid(undo));
	TRY {
		(**undo).styleSaved = false;
		(**undo).typingRun = false;
		Cut(undo);
		UndoStatusSet(undo, TXUNDO_CUT);
	} CATCH {
		TxUndoReset(undo);
	} ENDTRY;
	ensure(TxUndoValid(undo));
}

/* copy command */
void TxUndoCopy(TextUndoHandle undo)
{
	require(TxUndoValid(undo));
	TRY {
		(**undo).styleSaved = false;
		(**undo).typingRun = false;
		TxCopy((**undo).text);
	} CATCH {
		TxUndoReset(undo);
	} ENDTRY;
	ensure(TxUndoValid(undo));
}

/* paste command */
void TxUndoPaste(TextUndoHandle undo)
{
	require(TxUndoValid(undo));
	TRY {
		(**undo).styleSaved = false;
		(**undo).typingRun = false;
		Paste(undo);
		UndoStatusSet(undo, TXUNDO_PASTE);
	} CATCH {
		TxUndoReset(undo);
	} ENDTRY;
	ensure(TxUndoValid(undo));
}

/* clear command */
void TxUndoClear(TextUndoHandle undo)
{
	require(TxUndoValid(undo));
	TRY {
		(**undo).styleSaved = false;
		(**undo).typingRun = false;
		Clear(undo);
		UndoStatusSet(undo, TXUNDO_CLEAR);
	} CATCH {
		TxUndoReset(undo);
	} ENDTRY;
	ensure(TxUndoValid(undo));
}

/* apply the style to the text */
void TxUndoStyleSet(TextUndoHandle undo, short mode, TextStyle *style)
{
	require(TxUndoValid(undo));
	TRY {
		(**undo).typingRun = false;
		SetStyle(undo, mode, style);
		UndoStatusSet(undo, TXUNDO_STYLE);
	} CATCH {
		TxUndoReset(undo);
	} ENDTRY;
	ensure(TxUndoValid(undo));
}

/* enter the character */
void TxUndoKey(TextUndoHandle undo, char key)
{
	require(TxUndoValid(undo));
	TRY {
		(**undo).styleSaved = false;
		Typing(undo, key);
	} CATCH {
		TxUndoReset(undo);
	} ENDTRY;
	ensure(TxUndoValid(undo));
}

/* the text is being clicked */
void TxUndoClick(TextUndoHandle undo)
{
	require(TxUndoValid(undo));
	(**undo).typingRun = false;
	(**undo).styleSaved = false;
	ensure(TxUndoValid(undo));
}

/* adjust undo menu item */
void TxUndoAdjustMenu(TextUndoHandle undo)
{
	Str255 title;
	
	require(TxUndoValid(undo));
	GetIndString(title, RLS_UNDO, TxUndoStatus(undo) + 2);
	SetItem(MenuCmdHandle(CMD_UNDO), MenuCmdItem(CMD_UNDO), title);
	if (TxUndoStatus(undo) != TXUNDO_NULL) MenuCmdEnable(CMD_UNDO);
	ensure(TxUndoValid(undo));
}

/* allocate a new undo structure */
TextUndoHandle TxUndoBegin(TextHandle text)
{
	TextUndoHandle undo = NULL;
	
	undo = HandleBeginClear(sizeof(TextUndoType));
	(**undo).text = text;
	ensure(TxUndoValid(undo));
	return(undo);
}

/* dispose of an undo structure */
void TxUndoEnd(TextUndoHandle undo)
{
	if (undo) {
		DisposeUndoText(undo);
		DisposeUndoStyle(undo);
		HandleEnd(undo);
		undo = NULL;
	}
	ensure(! TxUndoValid(undo));
}
